/*
 * Enhydra Java Application Server Project
 *
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site (http://www.enhydra.org/).
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific terms governing rights and limitations
 * under the License.
 *
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 * $Id: Builder.java,v 1.1 2001/09/28 16:01:06 cryd0221 Exp $
 */

/*
 *   Builder.java
 *
 * @author    Jay Gunter
 * @since     LBS1.8
 * @version   $Revision: 1.1 $
 *
 */

package org.enhydra.dods.builder.engine;

import org.enhydra.dods.store.*;

import org.enhydra.dods.store.doa.*;
import org.enhydra.dods.gui.tree.*;
import org.enhydra.dods.builder.generator.base.SourceGenerator;
import org.enhydra.dods.builder.engine.GeneratorInfo;

import  com.lutris.appserver.server.session.*;
import  com.lutris.util.Config;
import  com.lutris.util.ConfigFile;
import  com.lutris.util.ConfigException;
import  java.io.*;
import  java.util.*;
import  java.math.BigDecimal;
import  java.net.URLEncoder;

import  javax.swing.JTextArea;

/**
 *
 *   For each DataObject in the current DODSProject,
 *   the run() method invokes the source code generators
 *   described in the dods.conf file.
 *   After generating the tree of source code for for the DataObjects,
 *   a makefile is created in each directory in the tree.
 *   Finally, 'make' is run at the top directory.
 *
 */
public class Builder {
    /**
     * DODS_HOME is the name of a property that is set with -D
     * when DODS is run.  The value for the property should be
     * the directory where DODS can find its config file, dods.conf.
     */
    private static final String DODS_HOME = "DODS_HOME";

    /**
     * dodsDir is the value of the DODS_HOME property.
     */
    private static String dodsDir = null;

    /**
     * ENHYDRA_DIR is the name of a property that is set with -D
     * when DODS is run.  The value for the property should be
     * the directory where DODS can find the Lutris stdlib.mk file.
     */
    private static final String ENHYDRA_DIR = "ENHYDRA_DIR";

    /**
     * lutrisDir is the value of the ENHYDRA_DIR property.
     */
    private static String lutrisDir = null;

    /**
     * CLASS_PATH is an optional key in dods.conf that specifies where
     * the SourceGenerator classes are located.
     */
    private static final String CLASS_PATH = "ClassPath";

    /**
     * SOURCE_GEN is a required key in dods.conf that informs the builder
     * which SourceGenerator classes are to be loaded and run.
     * @see RUN_SOURCE_GEN for example line in dods.conf file.
     */
    private static final String SOURCE_GEN = "SourceGenerator";

    /**
     * GEN_RUN is a required key in dods.conf that tells the builder
     * if a given SourceGenerator class is to be run.
     *   e.g.
     *      SourceGenerator.DO.RunMe= true
     */
    private static final String GEN_RUN = "RunMe";

    /**
     * GEN_RUN_FOR_ABSTRACT is an optional key in dods.conf
     * that tells the builder if a given SourceGenerator class is to be run
     * for Data Objects marked in DODS as "abstract".
     * If not defined, default value is "false".
     *   e.g.
     *      SourceGenerator.DO.RunForAbstract= true
     */
    private static final String GEN_RUN_FOR_ABSTRACT = "RunForAbstract";

    /**
     * GEN_HANDLES_ARRAYS is an optional key in dods.conf
     * that tells the builder if a given SourceGenerator class should
     * perform any actions for Attributes with isArray()==true.
     * If not defined, default value is "false".
     *   e.g.
     *      SourceGenerator.DO.HandleArrays= true
     */
    private static final String GEN_HANDLES_ARRAYS = "HandleArrays";

    /**
     * GEN_CLASS_NAME is a required key in dods.conf.
     * It tells the builder the name of a given SourceGenerator class.
     *   e.g.
     *      SourceGenerator.DO.GeneratorClass= SourceGenerator_DO
     */
    private static final String GEN_CLASS_NAME = "GeneratorClass";

    /**
     * GEN_OUTPUT_SUFFIX is a required key in dods.conf.
     * It tells the builder the suffix of the file
     * that a given SourceGenerator class will create.
     *   e.g.
     *      SourceGenerator.DO.OutputFileSuffix= java
     */
    private static final String GEN_OUTPUT_SUFFIX = "OutputFileSuffix";

    /**
     * GEN_OBJECT_SUFFIX is an optional key in dods.conf.
     * It tells the builder the SourceGenerator what string to append
     * to the object name.
     *   e.g.
     *      SourceGenerator.HomeInterface.ObjectSuffix= Home
     */
    private static final String GEN_OBJECT_SUFFIX = "ObjectSuffix";

    /**
     * GEN_OUTPUT_CLASS_PARENT is an optional key in dods.conf.
     * It tells the builder what (if any) class is extended by the
     * classes that will be generated.
     *   e.g.
     *      SourceGenerator.DO.OutputClassParent= com.lutris.blah.CoreDO
     */
    private static final String GEN_OUTPUT_CLASS_PARENT = "OutputClassParent";

    /**
     * GEN_EXTRA_CLASSPATH is an optional key in dods.conf.
     * It tells the builder where the GEN_EXTRA_CLASSPATH can be found.
     *   e.g.
     *      SourceGenerator.DO.ExtraClasspath= /users/fred/LBS/Harmony
     */
    private static final String GEN_EXTRA_CLASSPATH = "ExtraClasspath";

    /**
     * GEN_TEMPLATE_DIR is a required key in dods.conf that tells the builder
     * the location of the templates used by a given SourceGenerator class.
     *   e.g.
     *      SourceGenerator.DO.TemplateDir= .../templates
     */
    private static final String GEN_TEMPLATE_DIR = "TemplateDir";

    /**
     * sourceGenerators is a Vector of GeneratorInfo objects.
     * GeneratorInfo objects are used to initialized SourceGenerator objects.
     */
    private static Vector sourceGenerators = null;

    /**
     * We perform some special post-processing on the .sql files
     * produced by the SourceGenerator_SQL class,
     * but only if that class was run.
     */
    private static boolean generating_sql_files = false;

    /**
     * OUTPUT_DIR is the name of a property that is set with -P
     * when DODS is run.  The value for the property is the top of the
     * directory tree where DODS will write the source files it generates.
     */
    private static final String OUTPUT_DIR = "DODS_OUTPUT_DIR";

    /**
     * outputDir is the value of the OUTPUT_DIR property.
     */
    private static String outputDir = null;

    /**
     * ROOT_DIR is the name of a property that is set with -P
     * when DODS is run.  The value for the property is the value of the
     * ROOT_DIR macro that is set in a top-level Lutris makefile.
     */
    private static final String ROOT_DIR = "ROOT_DIR";

    /**
     * rootDir is the value of the ROOT_DIR property.
     */
    private static String rootDir = null;

    /**
     * MAKEFILE_SCRIPT is a required key in dods.conf that tells the builder
     * facts about the script that generates the Makefiles for the
     * generated source code.
     */
    private static final String MAKEFILE_SCRIPT = "MakefileGeneratorScript";

    /**
     * MAKEFILE_SCRIPT_PATH is a required key in dods.conf that tells
     * the builder the pathname of the script.
     */
    private static final String MAKEFILE_SCRIPT_PATH = "Path";

    /**
     * makefileScriptPath is the value of the MAKEFILE_SCRIPT_PATH property.
     */
    private static String makefileScriptPath = null;

    /**
     * MAKEFILE_SCRIPT_RUN is a required key in dods.conf that tells
     * the builder whether to run the script.
     */
    private static final String MAKEFILE_SCRIPT_RUN = "RunAutomatically";

    /**
     * makefileScriptRun is the value of the MAKEFILE_SCRIPT_RUN property.
     */
    private static String makefileScriptRun = null;

    /**
     * RUN_MAKE is a required key in dods.conf that tells
     * the builder whether to run 'make' after the makefiles have
     * been generated.
     */
    private static final String RUN_MAKE = "RunMake";

    /**
     * makefileScriptRun is the value of the RUN_MAKE property.
     */
    private static String runMake = null;

    /**
     * MAKEFILE_SCRIPT_CLEANUP is a required key in dods.conf that tells the
     * script whether to delete the generated source code after compilation.
     */
    private static final String MAKEFILE_SCRIPT_CLEANUP = "DeleteSourceCode";

    /**
     * makefileScriptCleanup is the value of the
     * MAKEFILE_SCRIPT_CLEANUP property.
     */
    private static String makefileScriptCleanup = null;

    /**
     * MAKEFILE_EXTRA_CLASSPATH is an optional key in dods.conf that tells the
     * script of any value for the EXTRA_CLASSPATH macro
     * in the generated Makefiles.
     */
    private static final String MAKEFILE_EXTRA_CLASSPATH = "ExtraClasspath";

    /**
     * makefileExtraClasspath is the value of the
     * MAKEFILE_EXTRA_CLASSPATH property.
     */
    private static String makefileExtraClasspath = null;

    /**
     * dirNames is the list of directories that were created by mkdir()
     * during this run of the Builder.
     */
    //private static Hashtable dirNames;
    private static Vector dirNames;

    /**
     * numDirs is the number of elements in dirNames.
     */
    private static int numDirs;

    /*
     * dodsProject is the contents of the DODS project file saved to disk
     */
    private static DODSProject dodsProject;

    /*
     * treeRoot is the top of the DataObject hierarchy for the DODS project
     */
    private static DOTreeNode treeRoot;

    /*
     * createdFiles is the list of files generated by Builder.
     */
    private static Vector createdFileNames = new Vector();

    /*
     * Write error and halt JVM.
     */
    static private void fatalErr(String msg) {
        System.err.println(msg);
        Runtime vm = Runtime.getRuntime();
    // FIX
//        vm.exit(1);
    }

    /*
     * Here are the meaningful combinations of options:
     *
     *    -enhydra3
     *    -enhydra4    (implies -jdbc1)
     *    -ejb        (implies -enhydra4)
     *    -jdbc2        (implies -enhydra4)
     *    -ejb,-jdbc2    (implies -enhydra4)
     *
     *    -lazy        can be added to any combination of options
     *
     * Note:  In Enhydra 4, EJB generators are enabled by default in dods.conf.
     */

    /*-------------------------------------------------------------------------
         -help    Explains the various options accepted by Builder.main.
    */
    public static final String OPT_HELP        = "-help";

    /*-------------------------------------------------------------------------
         -lazy (or "-DLAZY=true")

               Causes all DO classes to use lazy loading.
        In a .doml file, the isLazyLoading=true attribute
        can be added to individual tables.

        "Lazy loading" means that if XxxDO has a reference to YyyDO,
        the YyyDO object is not actually queried from the database
        until the application calls XxxDO.getYyy().

        This is almost always the desired behavior.
    */
    public static final String PROP_LAZY    = "LAZY";
    public static final String OPT_LAZY        = "-lazy";

    public static boolean VERBOSE = false;
    public static final String OPT_VERBOSE        = "-verbose";

    /*-------------------------------------------------------------------------
         -enhydra3 (or "-DENHYDRA3=true")

               EJB generators are off.
        Generate Enhydra 3 style DO and Query classes.
        Connection usage sequence:
                       XxxDO.save()
            calls   Enhydra.getDatabaseManager().createTransaction()
            calls   StandardDBConnectionAllocator.allocateConnection()
            calls   DriverManager.getConnection()

               These options are ignored:
            -enhydra4, -ejb, -jdbc1, -jdbc2
    */
    public static final String PROP_ENHYDRA3    = "ENHYDRA3";
    public static final String OPT_ENHYDRA3    = "-enhydra3";

    /*-------------------------------------------------------------------------
         -enhydra4 (or no options)

        The default if no -enhydra3 option.
               EJB generators are on.
              Generate Enhydra-independent DO and Query classes.
        Connection usage sequence for -jdbc (default):
                       XxxDO.save()
            calls    XxxDO.getConnection()
                   calls    DriverManager.getConnection()
        Connection usage sequence for -jdbc2:
                       XxxDO.save()
            calls    XxxDO.getConnection()
            calls   Context.lookup()  // JNDI
            to call    DataSource.getConnection()
    */
    public static final String PROP_ENHYDRA4    = "ENHYDRA4";
    public static final String OPT_ENHYDRA4    = "-enhydra4";

    /*-------------------------------------------------------------------------
         -ejb or (or "-D_EJB=true")

        Option -enhydra4 is implied, and EJB generators are on.
    */
    public static final String PROP_EJB    = "EJB";
    public static final String OPT_EJB    = "-ejb";

    /*-------------------------------------------------------------------------
         -jdbc1    (or "-DJDBC1=true")

        Option -enhydra4 is implied.
    */
    public static final String PROP_JDBC1    = "JDBC1";
    public static final String OPT_JDBC1    = "-jdbc1";

    /*-------------------------------------------------------------------------
         -DJDBC1=true
         -jdbc2

        Option -enhydra4 is implied.
    */
    public static final String PROP_JDBC2    = "JDBC2";
    public static final String OPT_JDBC2    = "-jdbc2";

    //   -DDB_VENDOR=vendorname
    //   -vendor=vendorname    Override database= setting in .doml file.
    //
    public static final String OPT_VENDOR    = "-vendor";

    public static boolean runEjbGenerators = true;
    private static String EjbGeneratorTypes = ":PK:Home:Remote:Bean:";

    private static void usage() {
        System.err.println(
            "DODS Builder.main() options: \n" +
            "    -help         Shows this message.\n" +
            "    -enhydra3     Generate code compatible with Enhydra3\n" +
            "                  (the EAF in Enhydra4).\n" +
            "    -enhydra4     Generate code that uses Connections or DataSources\n" +
            "                  (instead of the DBTransactions used in Enhydra3).\n" +
            "                  This is the default.  Implies -jdbc1.\n" +
            "    -jdbc1        Generate code that accepts Connections as parameters.\n" +
            "                  Can be used with either -enhydra3 or -enhydra4.\n" +
            "    -jdbc2        Generate code that uses JNDI to lookup DataSources\n" +
            "                  to obtain Connections.  Implies -enhydra4.\n" +
            "    -ejb          Enable EJB code generators.  Implies -enhydra4.\n" +
            "    -lazy         Set lazy-loading=true for all data objects.\n" +
            "                  Referenced objects will be fetched only when needed.\n" +
            "    -vendor=      Specify database vendor:\n" +
            "                               Standard, Oracle, Msql, InstantDB,\n" +
            "                               Informix, Sybase, PostgreSQL.\n" +
            "                  This is handy for testing against various vendors;\n" +
            "                  you don't need to change the database=\n" +
            "                  line in your .doml file.\n" +
            "\n" +
            "These options can be concatenated, e.g.\n" +
            "    -lazy-jdbc2-ejb\n" +
            "which is handy when the Ant taskdef is used to invoke DODS, e.g.:\n" +
            "    make ANT_ARGS=-DDODS_OPTIONS='-enhydra3,-vendor=Oracle'\n"
        );
    }

//This way of getting paramters is really bad, and should be re-written...csc

    public static void main(String argv[]) throws Exception {
        String dir  = null;
        String doml = null;

        // Process command line arguments and options.
        // Convert options to values in System Properties.
        // This allows DODS options to be set using java -D property settings.
        //
        // Note on running options together:
        // The DODS regression tests use Makefiles to invoke Ant
        // which in turn runs DODS.  An example command line is
        //     make ANT_ARGS=-DDODS_OPTIONS="-enhydra3,-vendor=InstantDB"
        // When invoking DODS in this way, the DODS options are
        // collected into a single string that is the value of
        // the DODS_OPTIONS property.
        // There can be no spaces within this string; use commas instead.
        //
        // -enhydra4,-jdbc1 is the default.
        setBoolProp(          PROP_ENHYDRA4,    OPT_ENHYDRA4,    OPT_ENHYDRA4);
        setBoolProp(        PROP_JDBC1,    OPT_JDBC1,    OPT_JDBC1);
        for (int i = 0; i < argv.length; i++) {
//System.err.println("argv[i]="+argv[i]);
            goodOption = false;
            if (0==argv[i].indexOf("-")) {
                if (argv[i].equals(OPT_HELP)) {
                    usage();
                    System.exit(1);
                }
                setBoolProp(    PROP_LAZY,    argv[i],    OPT_LAZY);
                setBoolProp(    PROP_ENHYDRA3,    argv[i],    OPT_ENHYDRA3);
                setBoolProp(    PROP_ENHYDRA4,    argv[i],    OPT_ENHYDRA4);
                setBoolProp(    PROP_EJB,    argv[i],    OPT_EJB);
                setBoolProp(    PROP_JDBC1,    argv[i],    OPT_JDBC1);
                setBoolProp(    PROP_JDBC2,    argv[i],    OPT_JDBC2);
                if (argv[i].equals(OPT_VERBOSE)) {
                    VERBOSE = true;
                    goodOption = true;
                }
                setVendorProp(argv[i]);
                if (! goodOption) {
                    System.err.println(
                        "DODS: Invalid option: " + argv[i] +
                    ".  Use -help to see valid options.");
                    System.exit(1);
                }
            } else if (     argv[i].equals("regen")) {
                // Redundant "regen" option, Builder.main does regen.
            } else if (-1 != argv[i].indexOf(".doml")) {
                doml = argv[i];
//System.err.println("doml ="+doml);
            } else {
                dir  = argv[i];
//System.err.println("dir  ="+dir );
            }
        }
        if (getBoolProp(PROP_ENHYDRA3)) {
//         setBoolProp(PROP_NO_EJB,    OPT_NO_EJB,    OPT_NO_EJB);
            System.getProperties().put(PROP_EJB, "false");
            System.getProperties().put(PROP_ENHYDRA4, "false");
        }
        if (getBoolProp(PROP_JDBC2)) System.getProperties().put(PROP_JDBC1, "false");
        if (getBoolProp(PROP_JDBC1)) System.getProperties().put(PROP_JDBC2, "false");
        if (getBoolProp(PROP_EJB)) {
//        runEjbGenerators = false;
            System.getProperties().put(PROP_ENHYDRA4, "true");
            runEjbGenerators = true;
        } else {
            runEjbGenerators = false;
        }

        StringBuffer sb = new StringBuffer(500);
        sb.append("Running DODS: ");
        sb.append(OPT_LAZY+" = "+getBoolProp(PROP_LAZY));
        sb.append(", "+OPT_ENHYDRA3+"="+getBoolProp(PROP_ENHYDRA3));
        sb.append(", "+OPT_ENHYDRA4+"="+getBoolProp(PROP_ENHYDRA4));
        sb.append(", "+OPT_EJB+"="+getBoolProp(PROP_EJB));
        sb.append(", "+OPT_JDBC1+"="+getBoolProp(PROP_JDBC1));
        sb.append(", "+OPT_JDBC2+"="+getBoolProp(PROP_JDBC2));
        sb.append(", "+OPT_VENDOR+"="+System.getProperty(OPT_VENDOR, "[default]"));
        Builder.print(sb.toString());
        
        run(dir, doml);
    }

    private static boolean goodOption = false;

    private static void setVendorProp(String arg) {
        int i = arg.indexOf(OPT_VENDOR);
        if (-1 != i) {
/*
To me this looks like a pretty stupid way to parse the parameter...
//        String afterEquals = arg.substring(i + OPT_VENDOR.length());
            StringTokenizer parser = new StringTokenizer(afterEquals, ",-");
            String vendor = null;
            if (parser.hasMoreTokens())
            vendor = parser.nextToken();
*/
            String vendor = arg.substring(i + OPT_VENDOR.length()+1);
            if (!(vendor.equals("InstantDB")) &&
                !(vendor.equals("Oracle")) &&
                !(vendor.equals("Informix")) &&
                !(vendor.equals("Msql")) &&
                !(vendor.equals("Sybase")) &&
                !(vendor.equals("PostgreSQL")) &&
                !(vendor.equals("Standard"))) {
                System.err.println ("Unknown vendor:"+vendor);
                return;
            }

//This does not seem to be the right place to be putting this value...
//        System.getProperties().put(SourceGenerator.DB_VENDOR, vendor);
            System.getProperties().put(OPT_VENDOR, vendor);
            goodOption = true;
        }
    }

    private static void setBoolProp(String propName, String arg, String option) {
        if (-1 != arg.indexOf(option)) {
            System.getProperties().put(propName, "true");
            goodOption = true;
        }
    }

    public static boolean getBoolProp(String propName) {
//System.err.println(propName + " = "+ System.getProperty(propName, "false"));
        return System.getProperty(propName, "false").equals("true");
    }

    /**
     * Initializes the DODS Builder, which generates source code
     * for each DataObject defined in a DODS project file.
     *
     * @exception Exception
     *   If an unrecoverable error occurs.
     *
     * 1) Get the location of the top of the output tree
     * 2) Get the location of the dods.conf file
     * 3) Load the dods.conf file.
     * 4) Get the classpath from the config file. It is an optional field
     * 5) Load the SourceGenerator classes
     * 6) Get info for running the MakefileGeneratorScript.
     */
    private static void init() throws Exception {

        //1) Get the location of the top of the output tree.
        if (null==outputDir) {
            outputDir = System.getProperty(OUTPUT_DIR);
            if (null==outputDir)
            fatalErr("Cannot find property:  " + OUTPUT_DIR);
        }

        //The property ROOT_DIR contains the value of the ROOT_DIR macro
        //that is set at the top of Lutris makefiles.
        rootDir = System.getProperty(ROOT_DIR);
        if (null==rootDir) rootDir = outputDir;

//csc_092401_start
/*        
        //2) Get the location of the dods.conf file.
        dodsDir = System.getProperty(DODS_HOME);
        if (null==dodsDir) fatalErr("Cannot find property:  " + DODS_HOME);
        String confFilename = dodsDir + File.separator + "dods.conf";
        File confFile = new File(confFilename);
        if (!confFile.exists()) fatalErr("No conf file " + confFilename);
        if (!confFile.isFile()) fatalErr("Not a file: " + confFilename);

        //3) Load the dods.conf file.
        Config cfg = null;
        try {
            FileInputStream cfgFIS = new FileInputStream(confFilename);
            ConfigFile cfgF = new ConfigFile(cfgFIS);
            cfgFIS.close();
            cfg = cfgF.getConfig();
        } catch (IOException ioe) {
            fatalErr("IO error reading " + confFilename + ":\n" + ioe.getMessage());
        } catch (ConfigException ce) {
            fatalErr("Bad file format reading " + confFilename + ":\n" + ce.getMessage());
        }
*/        
        //2) Get the dods.conf file (look first in DODS_HOME, but if we can't
        //   find it there, try and read it from the classpath (note that dods.conf
        //   is now in the same directory as Builder.java!)
        dodsDir = System.getProperty(DODS_HOME);
        File confFile = null;
        FileInputStream cfgFIS = null;
        ConfigFile cfgF = null;
        Config cfg = null;
        String confFilename = null;
        try {
            if (dodsDir!=null) {
                confFilename = dodsDir + File.separator + "dods.conf";
                confFile = new File(confFilename);
                if (confFile.exists() && confFile.isFile()) {
                    cfgFIS = new FileInputStream(confFilename);
                    cfgF = new ConfigFile(cfgFIS);
                } else {
                    confFile = null;
                }
            } 
            if (confFile==null) {
                confFilename = "dods.conf";
                InputStream is = Builder.class.getResourceAsStream(confFilename);
                cfgF = new ConfigFile(is);
            }
            cfg = cfgF.getConfig();
            if (cfg==null) fatalErr("Unable to read dods.conf!");
            if (cfgFIS!=null) cfgFIS.close();
        } catch (IOException ioe) {
            fatalErr("IO error reading " + confFilename + ":\n" + ioe.getMessage());
        } catch (ConfigException ce) {
            fatalErr("Bad file format reading " + confFilename + ":\n" + ce.getMessage());
        }
//csc_092401_end


        //4) Get the classpath from the config file. It is an optional field.
        String[] strArray = new String[1];
        String classPath = null;
        if (cfg.containsKey(CLASS_PATH)) {
            try {
                strArray = cfg.getStrings(CLASS_PATH);
                classPath = strArray[0];
            } catch (ConfigException ce) {
                fatalErr("Invalid config file " + confFilename + ". No key " + CLASS_PATH);
            }
        }

        //5) Load the SourceGenerator classes.
        SourceGenerator.reset();
        Config srcGen = null;
        try {
            srcGen = (Config) cfg.getSection(SOURCE_GEN);
        } catch (ConfigException e) {
            // Do nothing.
        } catch (Exception e) {
            fatalErr("Internal error parsing config file (section " + SOURCE_GEN + ").");
        }
        if (null==srcGen) fatalErr("Illegal config file. Section " + SOURCE_GEN + " not found.");
        sourceGenerators = new Vector();
        String[] genTypes = srcGen.keys();
        String runIt, templateDir, className, suffix, objSuffix;
        String parentClass, extraCP, runForAbstract;
        boolean genAbstract, handlesArrays;
        makefileExtraClasspath = "";
        Hashtable extraClasspaths = new Hashtable();
        for (int k = 0; k < genTypes.length; k++) {
            Config props = null;
            try {
                props = (Config) srcGen.getSection(genTypes[k]);
            } catch (Exception e) {}
            if (null==props) fatalErr("Illegal config file. Section " + SOURCE_GEN + "." + genTypes[k] + " not found.");
            runIt = props.getString(GEN_RUN);
            if (!runEjbGenerators && -1!=EjbGeneratorTypes.indexOf(genTypes[k])) runIt = null;
            if (null!=runIt && runIt.equals("true")) {
                // This SourceGenerator is enabled; prepare for using it.
//                templateDir    = fixDodsPath(props,  GEN_TEMPLATE_DIR);
                templateDir    = getDodsKey(props, GEN_TEMPLATE_DIR);
//System.out.println("templateDir: "+templateDir);                
                className      = props.getString(    GEN_CLASS_NAME);
//System.out.println("className: "+className);                
                try {
                    objSuffix  = props.getString(    GEN_OBJECT_SUFFIX);
                } catch (Exception e) {
                    objSuffix  = null;
                }
                suffix         = props.getString(    GEN_OUTPUT_SUFFIX);
//                parentClass    = fixIfMissing(props, GEN_OUTPUT_CLASS_PARENT);
                parentClass    = getDodsKey(props, GEN_OUTPUT_CLASS_PARENT);
//                extraCP        = fixDodsPath(fixIfMissing(props, GEN_EXTRA_CLASSPATH));
                extraCP        = getDodsKey(props, GEN_EXTRA_CLASSPATH);
                genAbstract    = getBoolean(  props, GEN_RUN_FOR_ABSTRACT);
                handlesArrays  = getBoolean(  props, GEN_HANDLES_ARRAYS);
                extraClasspaths.put(extraCP, "x");
                Class generator = Class.forName(className);
                GeneratorInfo info = new GeneratorInfo(
                    cfg, genAbstract, handlesArrays,
                    outputDir, genTypes[k],
                    suffix, parentClass, extraCP,
                    generator, templateDir, objSuffix);
                sourceGenerators.addElement(info);
                if (suffix.equals("sql")) generating_sql_files = true;
                /*
                if (genTypes[k].equals("PK")) {
                    // One of the EJB generators is turned on.
                    // But we allow the EJB property (if set) to override.
                    // On the DODS command line, either
                    //          -DEJB=true|false
                    // or
                    //        -noejb
                    //        -ejb
                    String ejb = sysProps.getProperty(EJB);
                    if (null != ejb) {
                    if (ejb.equals("false")
                    Properties sysProps = System.getProperties();
                    sysProps.put("EJB", "true");
                    }
                */
            }
        }

//csc_092401_start - we're not using make any more, so we don't need this
/*
        //6) Get info for running the MakefileGeneratorScript.
        lutrisDir = System.getProperty(ENHYDRA_DIR);
        if (null==lutrisDir) fatalErr("Cannot find property:  " + ENHYDRA_DIR);

        File dir = new File(lutrisDir);
        if (!dir.exists()) fatalErr("ENHYDRA_DIR " + lutrisDir + " does not exist.");
        if (dir.isFile()) fatalErr("ENHYDRA_DIR " + lutrisDir + " is not a directory.");

        Config mkGen = null;
        try {
            mkGen = (Config) cfg.getSection(MAKEFILE_SCRIPT);
        } catch (ConfigException e) {
            // Do nothing.
        } catch (Exception e) {
            fatalErr("Internal error parsing config file (section " + MAKEFILE_SCRIPT + ").");
        }
        if (null==mkGen) fatalErr("Illegal config file. Section " + MAKEFILE_SCRIPT + " not found.");
        makefileScriptPath    = fixDodsPath(mkGen, MAKEFILE_SCRIPT_PATH);
        runMake               = mkGen.getString(   RUN_MAKE);
        makefileScriptRun     = mkGen.getString(   MAKEFILE_SCRIPT_RUN);
        makefileScriptCleanup = mkGen.getString(   MAKEFILE_SCRIPT_CLEANUP);
        String cp             = mkGen.getString(   MAKEFILE_EXTRA_CLASSPATH);
        if (null != cp) makefileExtraClasspath += cp + ":";
*/        
//csc_092401_end
        //dirNames = new Hashtable();
        dirNames = new Vector();
        numDirs = 0;
    }

/*
    private static String fixDodsPath(Config props, String key) throws Exception {
        String dir = props.getString(key);
        if (null==dir || 0==dir.length()) throw new Exception("Bad value for key '" + key + "'");
        return fixDodsPath(dir);
    }

    private static String fixDodsPath(String dir) throws Exception {
        if (null==dir || 0==dir.length()) return "";
        if (dir.startsWith(DODS_HOME)) dir = dodsDir + dir.substring(DODS_HOME.length());
        return dir;
    }

    private static String fixIfMissing(Config props, String key) throws Exception {
        String str = "";
        try {
            str = props.getString(key);
            if (null==str) str = "";
        } catch (Exception e) {
            str = "";
        }
        return str;
    }
*/

    private static String getDodsKey(Config props, String key) {
        String str = null;
        try {str = props.getString(key);}
        catch (Exception e) {}
        if (str==null) str = "";
        return str;
    }

    private static boolean getBoolean(Config props, String key) {
        boolean ret = false;
        String value = null;
        try {
            value = props.getString(key);
            if (null != value && value.equals("true"))
            ret = true;
        } catch (Exception e) {}
        return ret;
    }


   /**
     * Runs the DODS Builder, creating source code for each DataObject.
     *
     * This version of the run() method can be used for invoking
     * the DODS Builder from the command-line.
     *
     * @param dir
     *   The directory containing the DODS project file.
     * @param dodsFilename
     *   The name of the DODS project file.
     * @exception Exception
     *   If an unrecoverable error occurs.
     *
     * open DODS project file to get DataObject hierarchy;
     * invoke processDO on the top node in the hierarchy;
     * run the makefileGenerator script;
     * invoke make on the generated makefile at the top level directory;
     */
    public static void run(String dir, String dodsFilename) throws Exception {
        try {
            //File file = new File(dir, dodsFilename);
            File file = new File(dodsFilename);
            /*
            FileInputStream finput = new FileInputStream(file);
            ObjectInputStream oInput = new ObjectInputStream(finput);
            dodsProject = (DODSProject) oInput.readObject();
            dodsProject.setFile(file);
            //run(dodsProject);
            */
            Store store = new Store();
            store.setInputFile(file);
            store.read();
            dodsProject = store.getProject();
    /*         dodsProject.setFile(file); */
            run(dir, dodsProject);
        } catch (Exception e) {
            System.err.println("ERROR run: " + e.getMessage());
            throw e;
        }
    }

   /**
     * Runs the DODS Builder, creating source code for each DataObject.
     *
     * This version of the run() method is used by the Build button on the GUI.
     *
     * @param targetDir
     *   The directory where the source code will be written.
     * @param project
     *   The project object containing the DataObject hierarchy.
     * @exception Exception
     *   If an unrecoverable error occurs.
     *
     * Record the targetDir value.
     * Invoke the run(DODSProject) method.
     */
    public static void run(String targetDir, DODSProject project) throws Exception {
        outputDir = targetDir;
        run(project);
    }

   /**
     * Runs the DODS Builder, creating source code for each DataObject.
     *
     * This version of the run() method supports invoking
     * the DODS Builder both from the DODS GUI ('Build' button)
     * and from the the command-line (via the other run() method).
     *
     * @param dodsProject
     *   The project object containing the DataObject hierarchy.
     * @exception Exception
     *   If an unrecoverable error occurs.
     *
     * open DODS project file to get DataObject hierarchy;
     * invoke processDO on the top node in the hierarchy;
     * run the makefileGenerator script;
     * invoke make on the generated makefile at the top level directory;
     */
    public static void run(DODSProject project) throws Exception {
        DOgraphNode.reset();
        init();
        dodsProject = project;
        clearDOarray();
        DataObject[] a = getAllDataObjects(project);
        for (int k = 0; k < a.length; k++) {
            processDO(a[k]);
        }
        if (generating_sql_files) generateSQLcommands(dodsProject.getName());

//csc_092401        if ("true".equals(makefileScriptRun)) generateMakefiles();
    }

    /**
     * Ideally, the Builder class would not be concerned with the
     * generation of any particular type of file.
     * But the SourceGenerator classes are designed to generate
     * one file per DataObject; they are not well suited to handle
     * final roundup actions like we're doing here.
     * Uglier contortions would be necessary to place this method in the
     * SourceGenerator_SQL class.
     */
    private static void generateSQLcommands(String projectName) {
        String objectid_tablename = "objectid";
//        String create_sql = outputDir + "/create_tables.sql";
        String create_sql = outputDir + File.separator + projectName + "_create_tables.sql";
        try {
            FileWriter create_tables = new FileWriter(create_sql);
            char[] buf = new char[ 512 ];
            int n;
            String[] sql_files = DOgraphNode.getOrderedFilenames();
            for (int i = 0; i < sql_files.length; i++) {
            try {
                FileReader in = new FileReader(sql_files[ i ]);
                while (0 <= (n = in.read(buf)))
                create_tables.write(buf, 0, n);
                in.close();
            } catch (FileNotFoundException e) {
                // assume DO was abstract, thus having no sql file.
            }
            }
            String s = "create table " + objectid_tablename +
                "(\n next " + SourceGenerator.oidDbType +
                " NOT NULL \n);\n";
            create_tables.write(s.toCharArray());
            create_tables.flush();
            create_tables.close();
        } catch (Exception e) {
            print("ERROR IN GENERATING FILE " + create_sql + ": " + e.getMessage());
        }

//        String drop_sql = outputDir + "/drop_tables.sql";
        String drop_sql = outputDir + File.separator + projectName + "_drop_tables.sql";
        try {
            FileWriter drop_tables = new FileWriter(drop_sql);
            drop_tables.write(DOgraphNode.getDropCommands().toCharArray());
            String s = "drop table " + objectid_tablename + ";\n";
            drop_tables.write(s.toCharArray());
            drop_tables.flush();
            drop_tables.close();
        } catch (Exception e) {
            print("ERROR IN GENERATING FILE " + drop_sql);
        }
    }

    /** 
     * Walk the project tree to build an array of all DataObjects.
     */
    private static DataObject[] DOarray;
    private static void clearDOarray() {
        DOarray = null;
    }
    public static DataObject[] getAllDataObjects(DODSProject project) {
        if (null != DOarray) return DOarray;
        Vector all = new Vector();
        DOTreeNode root = new DOTreeNode(project, true);
        root.updateChildrenOrder();
        addChildrenOf(all, root);
        DOarray = new DataObject[ all.size() ];
        all.copyInto(DOarray);
        return DOarray;
    }
    private static void addChildrenOf(Vector all, DOTreeNode node) {
        DataObject DO = node.getDataObject();
        if (! (DO instanceof DODSProject || DO instanceof DOPackage)) {
            all.addElement(DO);
        }
        Vector children = node.getChildren();
        if (null != children) for (int i = 0; i < children.size(); i++) {
            addChildrenOf(all, (DOTreeNode) children.elementAt(i));
        }
    }

    /**
     * Processes a DataObject.
     *
     *    create the directory for the source files to be generated;
     *    for each SourceGenerator {
     *       initialize the SourceGenerator;
     *       invoke the writeFileHeader method, passing the DataObject;
     *       for each Attribute belonging to the DataObject {
     *          invoke the writeMemberStuff method, passing the Attribute;
     *       }
     *       invoke the writeFileFooter method, passing the DataObject;
     *       invoke the adjustOutputFile method;
     *    }
     *
     * @param DO
     *   The DataObject to be processed
     * @exception Exception
     *   If an unrecoverable error occurs.
     *
     */
    private static void processDO(DataObject DO) throws Exception {

        if (!DO.isStoredProc()) DOgraphNode.addNode(DO);

        Vector members = DO.getAttributes();

        // Each array Attribute is made into a separate db table
        // with separate classes to access it.
        for (int j = 0; j < members.size(); j++) {
            Attribute attr = (Attribute) members.elementAt(j);
            if (attr.isArray()) processDO(SourceGenerator.handleArray(DO, attr));
        }

        for (int i = 0; i < sourceGenerators.size(); i++) {
            GeneratorInfo info  = (GeneratorInfo) sourceGenerators.elementAt(i);

            if (DO.isAbstract() && ! info.runForAbstract)
            continue;

            String flavor = info.outputFileFlavor;
            if (DO.isStoredProc() && 
               (flavor.equals("SQL") || -1 != EjbGeneratorTypes.indexOf(flavor))) continue;

//System.out.println ("DO: "+DO);
            String outDir = fixOutputDir(DO);
//System.out.println ("outDir: "+outDir);
            String subDir = mkdir(outDir, DO.getAdjustedPackageName());
//System.out.println ("subDir: "+subDir);
            info.outputDir = outDir + subDir;
//System.out.println ("info.outputDir: "+info.outputDir);
            SourceGenerator gen = (SourceGenerator) info.processor.newInstance();
            gen.objectPrep(info, dodsProject, DO);
            gen.writeFileHeader(DO);

            for (int j = 0; j < members.size(); j++) {
                Attribute attr = (Attribute) members.elementAt(j);
                if (! (attr.isArray() && ! info.handlesArrays)) {
                    gen.memberPrep(attr);
                    gen.writeMemberStuff(attr);
                }
            }
            gen.writeFileFooter(DO);
            gen.adjustOutputFile();
            createdFileNames.addElement(gen.getOutputFilename());
        }
    }

    /**
     * Return the list of filenames created.
     *
     * @return the list of filenames created by all generators.
     * @author Jay Gunter
     */
    public static String[] getcreatedFileNames() {
        String[] a = new String[ createdFileNames.size() ];
        createdFileNames.copyInto(a);
        return a;
    }


   /**
     * Creates a directory corresponding to the passed package path.
     *
     * @param pkgName
     *   The package for the DataObject currently being processed.
     * @exception Exception
     *   If an unrecoverable error occurs.
     *
     */
    private static String mkdir(String outputDir, String pkgName) throws Exception {
        // FIX replace the all this exec stuff with
        //     File dir = new File(outputDir + File.separator + dirName);
        //     dir.mkdir();
        // Will it create missing intermediate dirs like mkdir -p?
        String dirName = pkgName;
        dirName = dirName.replace('.', File.separatorChar);
        File dir = new File(outputDir + File.separator + dirName);
        dir.mkdirs();
        //dirNames.put(dirName, "");
        boolean found = false;
        for (int i = 0; ! found && i < dirNames.size(); i++) {
            String s = (String) dirNames.elementAt(i);
            if (null != s && s.equals(dirName))
            found = true;
        }
        if (! found) dirNames.addElement(dirName);
        numDirs++;
        return dirName;
    }

   /**
     * Invokes a shell script that creates a makefile in each output directory.
     * The script then runs 'make' in the top output directory.
     *
     * @exception Exception
     *   If an unrecoverable error occurs.
     */
//csc_092401_start
/*
    private static void generateMakefiles() throws Exception {
        if (numDirs==0) return;

        //Enumeration dirs = dirNames.keys();
        //String[] argv = new String[ numDirs + 7 ];
        //String[] argv = new String[ dirNames.size() + 7 ];
        Vector args = new Vector();

        String os = System.getProperty("os.name");
        if (-1 != os.indexOf("Windows")) args.addElement("bash");
        args.addElement(fixDodsPath("DODS_HOME/makefileGenerator.sh"));

        //if (null==argv[0] || 0==argv[0].length())
        if (null==makefileScriptPath || 0==makefileScriptPath.length()) throw new Exception ("Bad value for MAKEFILE_GENERATOR_SCRIPT.");

        args.addElement(fixScriptOption(runMake));
        args.addElement(fixScriptOption(makefileScriptCleanup));
        args.addElement(fixPath(outputDir));
        args.addElement(fixPath(rootDir));
        args.addElement(fixPath(makefileExtraClasspath));
        args.addElement(fixPath(lutrisDir));

        for (int i = 0; i < dirNames.size(); i++) {
            args.addElement(dirNames.elementAt(i));
        }
        String[] argv = new String[ args.size() ];
        args.copyInto(argv);

        Process p = Runtime.getRuntime().exec(argv);

        InputStream bins = p.getInputStream();
        InputStream bers = p.getErrorStream();

        int ex = 777;
        while(true) {
            try {
               // Sleep for 1 second waiting the completing of the process
               Thread.currentThread().sleep(1000);
                for (int ch = bins.read(); 0 <= ch; ch = bins.read()) {
                    print((char) ch);
                }
                for (int ch = bers.read(); 0 <= ch; ch = bers.read()) {
                    print((char) ch);
                }
                // Every call to Process.exitValue() will throw an
                // IllegalThreadStateException if the process is running.
                // We catch and ignore it to continuing waiting.
                // When the process finishes ex will contain the return value.
                ex = p.exitValue();
                print("\nDODS BUILD COMPLETE (exit code " + ex + ")\n");
                break;
            } catch(IllegalThreadStateException  e) {
                //System.err.println("ERROR run: " + e.getMessage());
            }
        }
        bins.close();
        bers.close();

        if (ex != 0 && ex != 777) {
            throw new Exception ("Cannot run " + argv[0]);
        }
    }
*/
//csc_092401_end

    static private JTextArea textArea = null;
    static public void setTextArea(JTextArea t) {
        textArea = t;
    }

    static private final int bufSize = 200;
    static private StringBuffer buf = new StringBuffer(bufSize);
    static private int size = 0;
    public static void print(char ch) {
        if (!VERBOSE) return;
        buf.append(ch);
        size++;
        if ('\n'==ch /* || size >= bufSize */) {
            print(buf.toString());
            size = 0;
            buf = new StringBuffer(bufSize);
        }
    }

    public static void print(String str) {
        if (!VERBOSE) return;
        fireMessageEvent(str);        // OMEAD
        if (! str.endsWith("\n"))
            str += "\n";
        if (null != textArea) {
            textArea.append(str);
        } else {
            System.out.print(str);
        }
    }

    private static String fixScriptOption(String optionValue) throws Exception {
        if (null==optionValue || 0==optionValue.length()) {
            return "-unset";
        } else { 
            return optionValue;
        }
    }

    private static String fixPath(String path) {
        return path.replace('\\', '/');
    }

    /**
     * fixOutputDir
     *
     * The Enhydra utility 'newapp' creates a stub source tree
     * for a new Enhydra application.  Some directories in that tree are:
     *     app/app/presentation
     *     app/app/business
     *     app/app/data
     * When the DODS user clicks the Build button, DODS asks for
     * the directory into which the generated source code will be written.
     * It is often desirable to enter the path to the data/ directory
     * of an Enhydra application being developed:
     *     /home/jay/Projects/App/app/app/data/
     *
     * When the package hierarchy in the DODS project is
     *                                app.data[...]
     * then fixOutputDir silently truncates the output directory to
     *     /home/jay/Projects/App/app/
     * so that the source files are written to
     *     /home/jay/Projects/App/app/app/data/*.java
     *
     * But if the package hierarchy in the DODS project is
     *                                notapp.data[...]
     *                                       or
     *                                app.notdata[...]
     * then fixOutputDir does not alter the output directory,
     * and the source files are written to
     *     /home/jay/Projects/App/app/app/data/notapp/data/*.java
     *        or
     *     /home/jay/Projects/App/app/app/data/app/notdata/*.java
     */
    public static String fixOutputDir(String pkg) {
    //System.err.println("pkg ="+pkg);
        /*        outputDir    pkg    return        pkg goes in
         *        ---------    ---    ------        -----------
         * case 1:    x/a/d        a.d    x/        x/a/d
         * case 2:    x/a        a.d    x/        x/a/d
         * case 3:    x        a.d    x/        x/a/d
         * case 4:    x/x/a        a.d    x/x/        x/x/a/d
         * case 5:    x/a/a        a.d    x/a/        x/a/a/d
         * case 6:    x/a/x/a        a.d    x/a/x/        x/a/x/a/d
         * case 7:    x/a/d/q        a.d    x/a/d/q/    x/a/d/q/a/d
         * case 8:    x/a/d        a    x/a/d/        x/a/d/a
         * case 9:    x/a        b.d    x/a/        x/a/b/d
         * case 10:    x/a/d        a.i    x/a/d       x/a/d/a/d/i
         * case 11:    x/a/d/z/q    a.d.i    x/a/d/z/q/  x/a/d/z/q/a/d/i
         *
         * Cases 1 thru 6 are "normal".
         *
         * In other cases, the output directory hierarchy does not match
         * the package hierarchy, and so we cannot assume the user
         * wants the package hierarchy integrated into the output tree;
         * so, we simply place the package hierarchy under the output directory.
         */
        String    dir    = outputDir.replace('\\', '/').trim();
//System.err.println("dir = "+dir    );
        String    pkgList    = pkg.replace('.', '/').trim();
        String    pl    = pkgList;
//System.err.println("pl = "+pl    );
        int    plen    = pl.length();
//System.err.println("plen = "+plen    );
        int    x    = -1;
        int    sp    = -1;
        while (-1==(x = dir.lastIndexOf(pl))) {
            // pkg hierarchy does not match the ending of the outputDir
            if (-1==(sp = pl.lastIndexOf('/'))) break; // we've run out of pkg components
//System.err.println("sp = "+sp);
            // truncate last pkg component
            pl = pl.substring(0, pl.lastIndexOf('/'));
//System.err.println("pl = "+pl);
        }
//System.err.println("pkgList="+pkgList);


        if (-1 != x && pkgList.startsWith(dir.substring(x))) {
            // all or leading part of pkg hierarchy matches ending of dir
            dir = dir.substring(0, x);  // truncate dir at point of match
        } else { 
            dir += '/';
        }
/*
I do not understand what this message is for or why it is generated. I get it all 
the time, and it seems to have no adverse effects whatsoever. Therefore, I'm removing
it... csc
        if (-1==x) {
            print( "WARNING!!!  Package " + pkg +
                "\ndoes not fit into target directory " + outputDir + ".");
        }
*/        
        return dir;
    }
    public static String fixOutputDir(DataObject DO) {
        String pp = DO.getAdjustedPackageName();
        return fixOutputDir(pp);
    }
    public static void testFix(String[] argv) {
System.err.println("x argv[0]="+argv[0]);
System.err.println("x argv[1]="+argv[1]);
            outputDir = argv[0];
            String s = fixOutputDir(argv[1]);
System.err.println("s ="+s+"\n");
    }


    // START OMEAD'S MODS
    private static MessageListener listener = null;
    public static void setMessageListener(MessageListener l){
        listener = l;
    }

    public static void fireMessageEvent(String message) {
        if (listener==null) {
            return;
        } else {
            listener.messageSent(message);
        }
    }
    // END OMEAD'S MODS

}
