/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: DOgraphNode.java,v 1.1 2001/09/28 16:01:06 cryd0221 Exp $
 */

/*
 *   DOgraphNode.java
 *
 * @author    Jay Gunter
 * @since     LBS1.8
 * @version   $Revision: 1.1 $
 *
 */

package org.enhydra.dods.builder.engine;

import org.enhydra.dods.store.doa.*;

import  java.io.File;
import  java.util.Vector;
import  java.util.Hashtable;
import  java.util.Enumeration;


/**
 * This class is used by the Builder to generate two lists:
 * 1) a list of generated .sql files in a valid execution order.
 * 2) a list of "drop table" SQL commands in a valid execution order.
 * The 2nd list is in the reverse order of the 1st list.
 * The order is determined by the inter-tables references (foreign keys).
 * The tables with no references are created first and dropped last.
 */
class DOgraphNode {
    static private Hashtable set;
    static private Vector sqls;
    static private Vector tables;

    static public void reset() {
	set = new Hashtable();
	sqls = null;
	tables = null;
    }
    static {
	reset();
    }
    static public void addNode( DataObject DO ) {
	DOgraphNode node = new DOgraphNode( DO );
	set.put( DO, node );
    }
    static public String getCreateOrder() {
	String ret = "";
	process();
	for ( int i = 0; i < sqls.size(); i++ ) {
	    ret += (String) sqls.elementAt( i ) + "\n";
	}
	return ret;
    }
    static public String[] getOrderedFilenames() {
	process();
	String[] ret = new String[ sqls.size() ];
	for ( int i = 0; i < sqls.size(); i++ ) {
	    ret[ i ] = (String) sqls.elementAt( i );
	}
	return ret;
    }
    static public String getDropCommands() {
	String ret = "";
	process();
	for ( int i =  tables.size() - 1; 0 <= i; i-- ) {
	    String table = (String) tables.elementAt( i );
	    ret += "drop table " + table + ";\n";
	}
	return ret;
    }

    static private void process() {
	if ( null != sqls )
	    return;
	sqls = new Vector();
	tables = new Vector();
	fillRefs();
	Enumeration DOs;
	int prev_count = -1;
	boolean circular_refs = false;
	for (	DOs = set.keys();
		! set.isEmpty() && ! circular_refs;
		DOs = set.keys() ) {
	    int this_count = 0;
//System.err.println("========================================");
	    while ( DOs.hasMoreElements() ) {
		this_count++;
		DataObject DO = (DataObject) DOs.nextElement();
		DOgraphNode node = (DOgraphNode) set.get( DO );
//System.err.println("Looking at node ="+node );
		if ( node.noRefs() ) {
		    String outDir = Builder.fixOutputDir( DO );
		    String pkg = DO.getAdjustedPackageName();
		    String subDir = pkg.replace( '.', File.separatorChar );
		    sqls.addElement( outDir + subDir + File.separatorChar +
						DO.getName() + "SQL.sql" );
		    tables.addElement( DO.getDBTableName() );
		    zapRef( node );
		    set.remove( DO );
		}
	    }
	    if ( prev_count == this_count ) {
	    	circular_refs = true;
	    }
	    prev_count = this_count;
	}
	if ( circular_refs ) {
	    Builder.print(
		"\n!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
	    Builder.print( "WARNING!!!" );
	    Builder.print(
	    	"A circular reference (e.g. A -> B -> C -> A)" );
	    Builder.print(
	    	"has been detected among the following objects/tables:" );
	    DOs = set.keys();
	    while ( DOs.hasMoreElements() ) {
		DataObject DO = (DataObject) DOs.nextElement();
		DOgraphNode node = (DOgraphNode) set.get( DO );
		tables.addElement( DO.getDBTableName() );
		Builder.print( "        " + node );
	    }
	    Builder.print(
	    	"Some of these objects/tables may not be part of the circle." );
	    Builder.print(
	    	"A circular reference makes it impossible to establish" );
	    Builder.print(
	    	"a correct order for creating/dropping database tables," );
	    Builder.print(
	    	"and so some of the generated SQL commands in the files" );
	    Builder.print(
	    	"create_tables.sql and drop_tables.sql may fail." );
	    Builder.print(
	    	"A circular reference can also cause methods in the generated");
	    Builder.print(
	    	"Java classes to call each other in an infinite loop." );
	    Builder.print(
	    "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n\n");
	    return;
	}


	for ( DOs = set.keys(); ! set.isEmpty(); DOs = set.keys() ) {
//System.err.println("========================================");
	    while ( DOs.hasMoreElements() ) {
		DataObject DO = (DataObject) DOs.nextElement();
		DOgraphNode node = (DOgraphNode) set.get( DO );
//System.err.println("Looking at node ="+node );
		if ( node.noRefs() ) {
		    String outDir = Builder.fixOutputDir( DO );
		    String pkg = DO.getAdjustedPackageName();
		    String subDir = pkg.replace( '.', File.separatorChar );
		    sqls.addElement( outDir + subDir + File.separatorChar +
						DO.getName() + "SQL.sql" );
		    tables.addElement( DO.getDBTableName() );
		    zapRef( node );
		    set.remove( DO );
		}
	    }
	}
    }
    static private void zapRef( DOgraphNode ref ) {
//System.err.println("zapping ref ="+ref );
	Enumeration nodes = set.elements();
	while ( nodes.hasMoreElements() ) {
	    DOgraphNode node = (DOgraphNode) nodes.nextElement();
	    node.removeRef( ref );
	}
    }
    static private void fillRefs() {
	Enumeration DOs = set.keys();
	while ( DOs.hasMoreElements() ) {
	    DataObject DO = (DataObject) DOs.nextElement();
	    DOgraphNode node = (DOgraphNode) set.get( DO );
	    Vector attrs = DO.getAttributes();
	    for ( int i = 0; i < attrs.size(); i++ ) {
		Attribute attr = (Attribute) attrs.elementAt( i );
		if ( attr.isObjectRef() ) {
		    DataObject refDO    = attr.getObjectRefDO();
		    DOgraphNode refNode = (DOgraphNode) set.get( refDO );
		    node.addRef( refNode );
		}
	    }
	}
    }

    DataObject DO;
    Hashtable references;
    private DOgraphNode( DataObject o ) {
	DO = o;
	references = new Hashtable();
    }
    private boolean noRefs() {
	return references.isEmpty();
    }
    private void addRef( DOgraphNode ref ) {
//System.err.println("Node " + toString() + ": adding ref to " + ref.toString() );
	if ( this != ref )
	    references.put( ref, "" );
    }
    private void removeRef( DOgraphNode ref ) {
//if ( references.containsKey( ref ) )
//System.err.println("Node " + toString() + ": removing ref to " + ref.toString() );
	references.remove( ref );
    }
    public String toString() {
	return DO.getName() + "/" + DO.getDBTableName();
    }
}
