/*
 * Enhydra Java Application Server Project
 *
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific terms governing rights and limitations
 * under the License.
 *
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 * $Id: QueryBuilder.java,v 1.1 2001/09/28 16:01:06 cryd0221 Exp $
 */




package org.enhydra.dods.builder.generator.query;

import com.lutris.appserver.server.*;
import com.lutris.appserver.server.sql.*;
import java.io.*;
import java.sql.*;
import java.util.*;
import java.io.Serializable;
import java.math.BigDecimal;
import java.text.SimpleDateFormat;

import org.enhydra.dods.builder.generator.dataobject.GenericDO;


/**
 * <PRE>
 *	A QueryBuilder is a helper-object used by the xxxQuery classes.
 *
 *	Stand-alone usage:
 *
 *		QueryBuilder qb;
 *		qb = new QueryBuilder( "cats" );
 *			or
 *		qb = new QueryBuilder( "cats", "name, age" );
 *			...
 *		qb.addWhereClause( "owner.name", "Smith",
 *					QueryBuilder.NOT_NULL,
 *					QueryBuilder.EXACT_MATCH );
 *		qb.addWhereClause( "paws", 3, QueryBuilder.GREATER_THAN );
 *              qb.addTwoColumnWhereClause(
 *              	"cat.ownerId", "owner.Id", QueryBuilder.EQUAL );
 *		ResultSet rs = qb.executeQuery( conn );
 *			...
 *		qb.reset();
 *
 * </PRE>
 *
 * @author Jay Gunter
 * @version $Revision: 1.1 $
 */
//public class QueryBuilder implements Query {
public class QueryBuilder implements Query, Serializable {	//bug 11817 - csc

// Special values for whereClauses:
private static final String	NOT			= " NOT ";
private static final String	OR			= " OR ";
private static final String	OPEN_PAREN		= " ( ";
private static final String	CLOSE_PAREN		= " ) ";

// Comparison operator passed as 3rd arg to addWhereClause( int|float|double )
//
public static final String	EQUAL			= "=";
public static final String	NOT_EQUAL		= "!=";
public static final String	LESS_THAN		= "<";
public static final String	LESS_THAN_OR_EQUAL	= "<=";
public static final String	GREATER_THAN		= ">";
public static final String	GREATER_THAN_OR_EQUAL	= ">=";
public static final String	IS_NULL			= " IS NULL ";
public static final String	IS_NOT_NULL		= " IS NOT NULL ";

public static final String	CASE_SENSITIVE_CONTAINS		= "%X%";
public static final String	CASE_INSENSITIVE_CONTAINS	= "%x%";
public static final String	CASE_SENSITIVE_STARTS_WITH	= "X%";
public static final String	CASE_INSENSITIVE_STARTS_WITH	= "x%";
public static final String	CASE_SENSITIVE_ENDS_WITH	= "%X";
public static final String	CASE_INSENSITIVE_ENDS_WITH	= "%x";
public static final String	CASE_INSENSITIVE_EQUAL		= "%xxxx";

/**
 * Null-acceptability indicator passed as 3rd arg to addWhereClause( String )
 * @deprecated Use addWhere() methods instead of addWhereClause() methods.
 */
public static final boolean	NULL_OK			= true;
/**
 * Null-acceptability indicator passed as 3rd arg to addWhereClause( String )
 * @deprecated Use addWhere() methods instead of addWhereClause() methods.
 */
public static final boolean	NOT_NULL		= false;

/**
 * Wild-card          indicator passed as 4th arg to addWhereClause( String )
 * @deprecated Use addWhere() methods instead of addWhereClause() methods.
 */
public static final boolean	EXACT_MATCH		= true;
/**
 * Wild-card          indicator passed as 4th arg to addWhereClause( String )
 * @deprecated Use addWhere() methods instead of addWhereClause() methods.
 */
public static final boolean	NOT_EXACT		= false;

// Default inexact string match keyword (e.g. can be MATCHES or LIKE)
public static final String DEFAULT_MATCHES_KEYWORD = "MATCHES";
private static String likeKeyword = DEFAULT_MATCHES_KEYWORD;

// Default inexact string wildcard (e.g. can be * or %)
public static final String DEFAULT_WILDCARD = "*";
private static String wildcard = DEFAULT_WILDCARD;

// See debug() method.
static private boolean debugAllSQL = false;
private boolean debugSQL = false;

/*
The DODS (Data Object Design Studio) development tool
generates data-layer source code which uses QueryBuilder objects
to perform database queries.
This is the original (1st, old) purpose of QueryBuilder.

QueryBuilder can also be used by application developers
to perform straight SQL queries.  Certain static final data members
of the DO classes generated by DODS are used to specify the tables and columns
for the query.  This allows developers to write straight SQL queries
but still enjoy the compile-time checking.
This is the 2nd (new) purpose of QueryBuilder.

================= Usage 1: DODS generated code =================================

NOTE!  Application developers, please see "Usage 2" section below.

1st, old, String-based style (used by DODS-generated Query classes):

    Query creation:

	The ctor takes a String tablename
	(and optional String list of comma-separated column names).
	E.g.:

	    // In PersonQuery class:
	    QueryBuilder qb = new QueryBuilder( "person", "person.LastName" );

    addWhereClause() methods:

	A where-clause is added by calling one of the addWhereClause() methods
	which takes a String column name a String value for comparison.
	Because table and column names are Strings, the application code
	can become out-of-sync with a changing database schema,
	and the developer gets no warning from the compiler.
	This scheme is ok for use by DODS since the application code
	(DO, Query classes) are (re)generated every time the
	schema (the DODS project file) is changed.
	However, this scheme is non-ideal for manual coding.
	E.g.:

	    // In PersonQuery class:
	    qb.addWhereClause( "FirstName", "Bob", QueryBuilder.EQUAL );

    Obtaining results:

	The results of the query are returned as a JDBC ResultSet.
	Column values are retrieved by their postion within each returned row,
	or by the String name of the column in the returned row.
	Again, no hope for compile-time checking.
	E.g.:

	    // PersonQuery class creates a DBQuery object to run the query:
	    DBQuery dbQuery = Enhydra.getDatabaseManager().createQuery();
	    dbQuery.query( this );  // invokes PersonQuery.executeQuery();
	    // PersonQuery class extends Query and implements executeQuery():
	    ResultSet rs = qb.executeQuery( connection );
	    while ( rs.next() ) {
		String lastname = rs.getString( 1 );
	    }

================= Usage 2: Application Developers =============================

2nd, new, Object Oriented style (used by application developers):

    Query creation:

	The ctor takes a Vector of RDBColumn objects.
	Each RDBColumn object knows which RDBTable it belongs to.
	In each DO class generated by DODS, there is a static RDBColumn object
	for each column in database table for the DO.
	E.g.:

	    Vector fields = new Vector();
	    fields.addElement( PersonDO.LastName );
	    QueryBuilder qb = new QueryBuilder( fields );

	Or:
	    QueryBuilder qb = new QueryBuilder();
	    qb.select( PersonDO.LastName );


    addWhere() methods:

	A where-clause is added by passing an RDBColumn object and a value.
	Because objects (not Strings) are used to specify columns in the query,
        errors in hand-written application code can be caught by the compiler.
	E.g.:

	    qb.addWhere( PersonDO.FirstName, "Bob" );

    Obtaining results:

	The results of the query are returned as a series of RDBRow objects.
	Each RDBRow contains an RDBColumnValue object for each returned column.
	An RDBColumnValue object is retrieved from the RDBRow using
	a one of the RDBColumn objects passed to the QueryBuilder ctor.
	The value is extracted from an RDBColumnValue object via its
	getString or getInteger (...) method.
	These methods throw an exception if the application attempts to
	extract an incorrect type for the RDBColumnValue.
	E.g.:

	    RDBRow row;
	    while ( null != ( row = qb.getNextRow() ) ) {
		String lastname = row.get( PersonDO.LastName ).getString();
	    }
	    qb.close();  // always release the resources

    Note:
	The templates used by the DODS sourceGenerator_Query class
	could be converted to use this 2nd style.
*/

/**
 * Construct a QueryBuilder object that will return specified fields of a table.
 * Style 2 (see above.)
 *
 * @param fields Vector of RDBColumn objects,
 * one for each column to be returned.  Fields can refer to different tables.
 */
public QueryBuilder( Vector fields ) {
    reset();
    if ( null != fields )
	selectedFields = (Vector) fields.clone();
}

/**
 * Construct a QueryBuilder object that will return specified fields of a table.
 * Style 2 (see above.)
 *
 * @param fields array of RDBColumn objects,
 * one for each column to be returned.  Fields can refer to different tables.
 */
public QueryBuilder( RDBColumn[] fields ) {
    reset();
    for ( int i = 0; i < fields.length; i++ ) {
        select( fields[i] );
    }
}

/**
 * Construct a QueryBuilder object.
 * When no fields are specified in the constructor,
 * you must use calls to select() to specify the RDBColumns
 * that will be returned.
 * This can be more convenient than constructing the Vector before
 * calling the QueryBuilder(Vector) constructor.
 * Style 2 (see above.)
 */
public QueryBuilder() {
    reset();
}

/**
 * Add a field to be returned by this QueryBuilder.
 * Style 2 (see above.)
 *
 * @param field RDBColumn object to be returned.
 * Fields can refer to different tables.
 */
public void select( RDBColumn field ) {
    if ( null == field )
	return;
    for ( int i = 0; i < selectedFields.size(); i++ ) {
	if ( field.equals( (RDBColumn) selectedFields.elementAt(i) ) )
	    return;
    }
    selectedFields.addElement( field );
}


/**
 * Returns an array of RDBRow objects.
 *
 * @return array of RDBRow objects.
 * @exception ATException If the data is not accessible
 * @exception BusinessRuleException If x is invalid.
 * @see getNextRow
 * @author Jay Gunter
 */
public RDBRow[] getRows()
throws QueryException {
    Vector v = new Vector();
    RDBRow row;
    while ( null != ( row = getNextRow() ) )
	v.addElement( row );
    RDBRow[] ret = new RDBRow[ v.size() ];
    v.copyInto( ret );
    return ret;
}


/**
 * Returns an RDBRow object containing the fields specified in the ctor.
 * On first call, executes the query.
 * NOTE:  to use this method, the QueryBuilder object must be created
 * using the
 *	QueryBuilder( RDBColumn[] )
 * constructor.
 *
 * @return row the RDBRow representing 1 row from the ResultSet.
 * @exception ATException If the data is not accessible
 * @exception BusinessRuleException If x is invalid.
 * @author Jay Gunter
 */
private DBQuery dbQuery = null;
private ResultSet rs = null;
private boolean done = false;
public RDBRow getNextRow()
throws QueryException {
    if ( done )
	return null;
    if ( null == rs ) {
	try {
	    if ( null == dbQuery ) {
		// May not be null if I later implement
		// a ctor that takes a DBQuery.
		dbQuery = Enhydra.getDatabaseManager().createQuery();
	    }
	} catch ( Exception e ) {
	    throw new QueryException(
		"SQL=[" + sql + "]: Unable to create query", e );
	}
        try {
	    // invokes executeQuery which sets ResultSet rs
	    dbQuery.query( this );
        } catch ( Exception e ) {
	    throw new QueryException(
		"SQL=[" + sql + "]: Unable to run query", e );
        }
	if ( null == rs )
	    throw new QueryException( "No ResultSet for Query." );
    }
    try {
	if ( ! rs.next() ) {
	    // ResultSet.next() only returns false once.
	    // After that, drivers typically throw "result set empty" exception.
	    // That isn't very friendly or useful to an application programmer.
	    // Setting done=true allows us to just keep returning nulls.
	    done = true;
	    dbQuery.release();
	    return null;
	}
    } catch ( Exception e ) {
	throw new QueryException(
	    "SQL=[" + sql + "]: Unable to get query results", e );
    }
    RDBColumnValue[] vals = new RDBColumnValue[ selectedFields.size() ];
    RDBColumn field = null;
    for ( int i = 0; i < selectedFields.size(); i++ ) {
        try {
	    field = (RDBColumn) selectedFields.elementAt(i);
	    vals[i] = new RDBColumnValue( field, rs.getObject(i+1) );
        } catch ( Exception e ) {
	    throw new QueryException(
		"SQL=[" + sql + "]: Unable to get query result for column "
		+ field, e );
        }
    }
    return new RDBRow( vals );
}


/**
 * WARNING!  This method is disabled and should never be called.
 * It's implementation is forced by the Query interface.
 * Use getNextRow instead().
 */
public Object next(ResultSet rs)
throws SQLException, ObjectIdException {
    return null;
}

///////////////////////////////////////////////////////////////////////////////
/// Constructors used by DODS-generated Query classes. ////////////////////////
///////////////////////////////////////////////////////////////////////////////

/**
 * Construct a QueryBuilder object that will return all fields of a table.
 * Style 1 (see above.)
 * WARNING: QueryBuilder objects created with this constructor
 * cannot be used with the getNextRow() method.
 * @see getNextRow
 *
 * @param table Name of the table on which to perform the SQL 'select'.
 */
public QueryBuilder( String tableName ) {
    this( tableName, tableName + ".*" );
}

/**
 * Construct a QueryBuilder object that will return specified fields of a table.
 * Style 1 (see above.)
 *
 * @param table     Name of the table on which to perform the SQL 'select'.
 * @param fieldList Comma-separated string of field names to retrieve.
 */
public QueryBuilder( String tableName, String fieldList  ) {
    reset();
/*     mainTableName = tableName.toLowerCase(); */
    mainTableName = tableName;
    storeTableName( mainTableName );
    selectClause = fieldList;
}

/**
 * Determine table name of column.
 * For example, we return 'tableX' from the following column specification:
 *     UPPER(tableX.columnY)
 * WARNING: trickier column specifications will cause this method to fail.
 *
 * @param column     A column (field) name, possibly with a leading table name.
 */
private String getTableName( String column ) {
    int offset = 0;
    int end = 0;
    String str = column;
    int index = str.lastIndexOf( "(" );
    if ( -1 != index )
//	str = str.substring( index + 1, 1000 );
	str = str.substring( index + 1);
    index = str.indexOf( ")" );
    if ( -1 != index )
	str = str.substring( 0, index );
    index = str.indexOf( "." );
    if ( -1 == index )
	return null;
    return str.substring( 0, index );
}

/**
 * Store all table names that occur as prefix of column specifications.
 * All table names must appear in the "from" clause.
 *
 * @param column     A column (field) name, possibly with a leading table name.
 */
private void storeTableNameForColumn( String column ) {
    String table = getTableName( column );
    storeTableName( table );
}

/**
 * Store all table names that occur as prefix of column specifications.
 * All table names must appear in the "from" clause.
 *
 * @param column     RDBColumn containing table name.
 */
private void storeTableNameForColumn( RDBColumn column ) {
    storeTableName( column.getTableName() );
}

/**
 * Store a table name.
 * All table names must appear in the "from" clause.
 *
 * @param table name.
 */
private void storeTableName( String table ) {
    if ( null != table )
        tableNames.put( table, "" );
//        tableNames.put( table.toLowerCase(), "" );
}

/**
 * <PRE>
 * Specify that the next conditional expression (where-clause)
 * should be preceded by OR instead of AND.  (AND is the default).
 * For example:
 *          addWhereClause( "person.fname", "Bob", EQUAL );
 *          addWhereClause( "person.lname", "Smith", EQUAL );
 *          	produces
 *          person.fname = 'Bob' AND person.lname = 'Smith'
 * But:
 *          addWhereClause( "person.fname", "Bob", EQUAL );
 *          addWhereOr();
 *          addWhereClause( "person.lname", "Smith", EQUAL );
 *          	produces
 *          person.fname = 'Bob' OR person.lname = 'Smith'

 * Disallows 'OR OR'.
 * Trailing ORs are removed by prepareStatement().
 *
 * </PRE>
 *
 * @author Jay Gunter
 */
public void addWhereOr() {
    int n = whereClauses.size();
    if ( 0 == n )
	return;
    String prev = (String) whereClauses.lastElement();
    if ( ! OR.equals( prev ) )
	whereClauses.addElement( OR );
}

public void addWhereNot() {
	whereClauses.addElement( NOT );
}

/**
 * Specify that the next conditional expression (where-clause)
 * should be preceded by '('.
 * Trailing OPEN_PARENs are removed by prepareStatement().
 *
 * @author Jay Gunter
 */
public void addWhereOpenParen() {
    whereClauses.addElement( OPEN_PAREN );
/*
for ( int i = 0; i < whereClauses.size(); i++ ) {
System.err.println("whereClauses.elementAt("+i+")="+whereClauses.elementAt(i));
}
*/
}

/**
 * Close a previous addWhereOpenParen.
 * Automatically disallows bad sequences like 'OR )' and '( )'.
 *
 * @author Jay Gunter
 */
public void addWhereCloseParen() {
    int n = whereClauses.size();
    if ( 0 == n )
	return;
    String prev = (String) whereClauses.lastElement();
    if ( ! ( OR.equals( prev ) || OPEN_PAREN.equals( prev ) ) )
	whereClauses.addElement( CLOSE_PAREN );
}

/**
 * Add a where-clause that compares a column against a String.
 * @param column  A column (field) name to use as a search constraint.
 * @param value   int value to compare against.
 * @param cmp_op   The comparison operator to use:  QueryBuilder.EQUAL, ...
 *
 * @author Jay Gunter
 */
public void addWhere( RDBColumn column, GenericDO value, String cmp_op ) {
    BigDecimal id = null;
    if ( null != value )
        id = value.getOId().toBigDecimal();
    _addWhereClause( column, id, cmp_op );
}

/**
 * Add a where-clause that compares a column against a byte array.
 * @param column  A column (field) name to use as a search constraint.
 * @param value   byte array to compare against.
 * @param cmp_op   The comparison operator to use:  QueryBuilder.EQUAL, ...
 *
 * @author Jay Gunter
 */
public void addWhere( RDBColumn column, byte[] value, String cmp_op ) {
    String v = null;
    if ( null != value )
	v = new String( value );
    _addWhereClause( column, v, cmp_op );
}

/**
 * Add a where-clause that compares a column against a String.
 * @param column  A column (field) name to use as a search constraint.
 * @param value   int value to compare against.
 * @param cmp_op   The comparison operator to use:  QueryBuilder.EQUAL, ...
 *
 * @author Jay Gunter
 */
public void addWhere( RDBColumn column, java.sql.Time value, String cmp_op ) {
    _addWhereClause( column, value, cmp_op );
}

/**
 * Add a where-clause that compares a column against a String.
 * @param column  A column (field) name to use as a search constraint.
 * @param value   int value to compare against.
 * @param cmp_op   The comparison operator to use:  QueryBuilder.EQUAL, ...
 *
 * @author Jay Gunter
 */
public void addWhere( RDBColumn column, java.sql.Timestamp value, String cmp_op ) {
    _addWhereClause( column, value, cmp_op );
}

/**
 * Add a where-clause that compares a column against a String.
 * @param column  A column (field) name to use as a search constraint.
 * @param value   long value to compare against.
 * @param cmp_op   The comparison operator to use:  QueryBuilder.EQUAL, ...
 *
 * @author Jay Gunter
 */
public void addWhere( RDBColumn column, long value, String cmp_op ) {
    _addWhereClause( column, new Long( value ), cmp_op );
}

/**
 * Add a where-clause that compares a column against a String.
 * @param column  A column (field) name to use as a search constraint.
 * @param value   double value to compare against.
 * @param cmp_op   The comparison operator to use:  QueryBuilder.EQUAL, ...
 *
 * @author Jay Gunter
 */
public void addWhere( RDBColumn column, double value, String cmp_op ) {
    _addWhereClause( column, new Double( value ), cmp_op );
}

/**
 * Add a where-clause that compares a column against a String.
 * @param column  A column (field) name to use as a search constraint.
 * @param value   float value to compare against.
 * @param cmp_op   The comparison operator to use:  QueryBuilder.EQUAL, ...
 *
 * @author Jay Gunter
 */
public void addWhere( RDBColumn column, float value, String cmp_op ) {
    _addWhereClause( column, new Float( value ), cmp_op );
}

/**
 * Add a where-clause that compares a column against a String.
 * @param column  A column (field) name to use as a search constraint.
 * @param value   BigDecimal value to compare against.
 * @param cmp_op   The comparison operator to use:  QueryBuilder.EQUAL, ...
 *
 * @author Jay Gunter
 */
public void addWhere( RDBColumn column, BigDecimal value, String cmp_op ) {
    _addWhereClause( column, value, cmp_op );
}

/**
 * Add a where-clause that compares a column against a String.
 * @param column  A column (field) name to use as a search constraint.
 * @param value   int value to compare against.
 * @param cmp_op   The comparison operator to use:  QueryBuilder.EQUAL, ...
 *
 * @author Jay Gunter
 */
public void addWhere( RDBColumn column, int value, String cmp_op ) {
    _addWhereClause( column, new Integer( value ), cmp_op );
}

/**
 * Add a where-clause that compares a column against a String.
 * @param column  A column (field) name to use as a search constraint.
 * @param value   boolean value to compare against.
 * @param cmp_op   The comparison operator to use:  QueryBuilder.EQUAL, ...
 *
 * @author Jay Gunter
 */
public void addWhere( RDBColumn column, boolean value, String cmp_op ) {
    _addWhereClause( column, new Boolean( value ), cmp_op );
}

/**
 * Add a where-clause that tests if a column is equal to a given String.
 * @param column  A column (field) name to use as a search constraint.
 * @param value   int value to compare against.
 *
 * @author Jay Gunter
 */
public void addWhere( RDBColumn column, int value ) {
    addWhere( column, value, EQUAL );
}

/**
 * Add a where-clause that compares a column against a String.
 * @param column  A column (field) name to use as a search constraint.
 * @param value   String value to compare against.
 * @param cmp_op   The comparison operator to use:  QueryBuilder.EQUAL, ...
 *
 * @author Jay Gunter
 */
public void addWhereLower( RDBColumn column, String value, String cmp_op ) {
    _addWhereClause( "LOWER(" + column.getFullColumnName() + ")", value, cmp_op );
}

/**
 * Add a where-clause that compares a column against a String.
 * @param column  A column (field) name to use as a search constraint.
 * @param value   String value to compare against.
 * @param cmp_op   The comparison operator to use:  QueryBuilder.EQUAL, ...
 *
 * @author Jay Gunter
 */
public void addWhere( RDBColumn column, String value, String cmp_op ) {
    _addWhereClause( column, value, cmp_op );
}

/**
 * Add a where-clause that tests if a column is equal to a given String.
 * @param column  A column (field) name to use as a search constraint.
 * @param value   String value to compare against.
 *
 * @author Jay Gunter
 */
public void addWhere( RDBColumn column, String value ) {
    addWhere( column, value, EQUAL );
}

/**
 * Add a where-clause that compares a column against a Date.
 * @param column  A column (field) name to use as a search constraint.
 * @param value   Date value to compare against.
 * @param cmp_op   The comparison operator to use:  QueryBuilder.EQUAL, ...
 *
 * @author Jay Gunter
 */
public void addWhere( RDBColumn column, java.sql.Date value, String cmp_op ) {
    _addWhereClause( column, value, cmp_op );
}

/**
 * Add a where-clause that compares a column against another column.
 * @param column1  A column (field) name to use as a search constraint.
 * @param column2  A column (field) name to use as a search constraint.
 * @param cmp_op   The comparison operator to use:  QueryBuilder.EQUAL, ...
 *
 * @author Jay Gunter
 */
public void addWhere( RDBColumn column1, RDBColumn column2, String cmp_op ) {
    /*
    storeTableNameForColumn( column1 );
    storeTableNameForColumn( column2 );

    addTwoColumnWhereClause(	column1.getFullColumnName(),
				column2.getFullColumnName(),
				cmp_op );
    */
    String cn1 = column1.getFullColumnName();
    String cn2 = column2.getFullColumnName();
    storeTableNameForColumn( cn1 );
    storeTableNameForColumn( cn2 );

    whereClauses.addElement( cn1 + " " + cmp_op + " " + cn2 );
}

/**
 * Add a where-clause that tests if a column is equal to another column.
 * @param column1  A column (field) name to use as a search constraint.
 * @param column2  A column (field) name to use as a search constraint.
 *
 * @author Jay Gunter
 */
public void addWhere( RDBColumn column1, RDBColumn column2 ) {
    addWhere( column1, column2, EQUAL );
}

/**
 * Add a where-clause that imposes a RIGHT OUTER JOIN to
 * test if a column is equal to another column.
 * @param column1  A column (field) name to use as a search constraint.
 * @param column2  A column (field) name to use as a search constraint.
 *
 * @author Jay Gunter
 */
public void addWhereRightOuter( RDBColumn column1, RDBColumn column2 ) {

    String cn1 = column1.getFullColumnName() + "(+)";
    String cn2 = column2.getFullColumnName();
    storeTableNameForColumn( cn1 );
    storeTableNameForColumn( cn2 );

    whereClauses.addElement( cn1 + " = " + cn2 );
    /*
    addTwoColumnWhereClause(
	    column1.getFullColumnName() + "(+)",
	    column2.getFullColumnName(), QueryBuilder.EQUAL );
	    */
}

/**
 * Add a where-clause that imposes a LEFT OUTER JOIN to
 * test if a column is equal to another column.
 * @param column1  A column (field) name to use as a search constraint.
 * @param column2  A column (field) name to use as a search constraint.
 *
 * @author Jay Gunter
 */
public void addWhereLeftOuter( RDBColumn column1, RDBColumn column2 ) {
    String cn1 = column1.getFullColumnName();
    String cn2 = column2.getFullColumnName() + "(+)";
    storeTableNameForColumn( cn1 );
    storeTableNameForColumn( cn2 );

    whereClauses.addElement( cn1 + " = " + cn2 );
}

/**
 * <PRE>
 * Add a where-clause to the SQL command.
 * The where-clause compare the values of two columns
 * using the specified comparison operator.
 * E.g., we want to find cats with too many paws:
 *      column1 = "cat.paws"
 *      column2 = "animalNorms.catPaws"
 *      cmp_op  = QueryBuilder.GREATER_THAN
 * generates the where-clause:
 *      cat.paws > animalNorms.catPaws
 * and remembers the table names (cat, animalNorms)
 * for inclusion in the from-clause.
 * </PRE>
 *
 * @param column1  A column (field) name to use as a search constraint.
 * @param column2  Another column to compare against.
 * @param cmp_op   The comparison operator to use:  QueryBuilder.EQUAL, ...
 * @deprecated  Use addWhere(RDBColumn c1, RDBColumn c1, String cmp_op) instead.
 */
public void addTwoColumnWhereClause(
			       String column1, String column2, String cmp_op )
{
    storeTableNameForColumn( column1 );
    storeTableNameForColumn( column2 );

    whereClauses.addElement( column1 + " " + cmp_op + " " + column2 );
}

/**
 * <PRE>
 * Add a where-clause to the SQL command.
 * The where-clause compare the values of two columns
 * using the EQUAL operator.
 * E.g., we want to find cats with the normal number of paws:
 *      column1 = "cat.paws"
 *      column2 = "animalNorms.catPaws"
 * generates the where-clause:
 *      cat.paws = animalNorms.catPaws
 * and remembers the table names (cat, animalNorms)
 * for inclusion in the from-clause.
 *
 * </PRE>
 *
 * @param column1  A column (field) name to use as a search constraint.
 * @param column2  Another column to compare against.
 * @deprecated  Use addWhere(RDBColumn c1, RDBColumn c1) instead.
 */
public void addTwoColumnWhereClause( String column1, String column2 )
{
    addTwoColumnWhereClause( column1, column2, EQUAL );
}

/**
 * Replace a substring with another string.
 *
 * @return the resulting string after replacement.
 * Returns null if any params are null.
 * Returns original string if 'find' not found.
 * @param s the original string.
 * @param find the substring to replace.
 * @param replace the replacement string.
 * @author Jay Gunter
 */
public static String stringReplace( String s, String find, String replace ) {
    if ( null == s || null == find || null == replace )
	return s;
    int x;
    if ( -1 == ( x = s.indexOf( find ) ) )
	return s;
    String ret = s.substring( 0, x );
    ret += replace;
    ret += s.substring( x + find.length() );
    return ret;
}


private void _addWhereClause( RDBColumn column, Object value, String cmp_op )
{
    storeTableNameForColumn( column );

    if ( null == value && column.notNull() ) {
	// column is NOT NULL but we are searching for null.
	// so, this new WHERE clause is obviously false.
	// but, there is no harm in that.
	cmp_op = IS_NULL;
    }
    buildWhereClause( column.getFullColumnName(), value, cmp_op );
}

private void _addWhereClause( String column, Object value, String cmp_op )
{
    storeTableNameForColumn( column );
    buildWhereClause( column, value, cmp_op );
}

private void buildWhereClause( String column, Object value, String cmp_op )
{
    if ( null == value )
	if ( EQUAL.equals( cmp_op ) )
	    cmp_op = IS_NULL;
	else
	    cmp_op = IS_NOT_NULL;


    if ( IS_NULL.equals( cmp_op ) || IS_NOT_NULL.equals( cmp_op ) ) {
	whereClauses.addElement( column + cmp_op );
    } else {
	if ( -1 != cmp_op.indexOf( '%' ) ) {  // CASE_...
	    // There are 7 wildcarded string comparison operators:
	    //	CASE_INSENSITIVE_EQUAL		= "%xxxx";
	    //	CASE_SENSITIVE_CONTAINS		= "%X%";
	    //	CASE_INSENSITIVE_CONTAINS	= "%x%";
	    //	CASE_SENSITIVE_STARTS_WITH	= "X%";
	    //	CASE_INSENSITIVE_STARTS_WITH	= "x%";
	    //	CASE_SENSITIVE_ENDS_WITH	= "%X";
	    //	CASE_INSENSITIVE_ENDS_WITH	= "%x";
	    // A lower-case 'x' means a case-insensitive comparison is desired.
	    String s = (String) value;
	    if ( -1 != cmp_op.indexOf( 'x' ) ) {
		// for case-insensitive comparisons,
		// push both column and value to lower case.
		s = s.toLowerCase();
		// Or, could do
		//     s = "LOWER(" + s + ")";
		// David Corbin thought that might be better somehow.
		// I'd rather not count on the database optimizer
		// any more than necessary.
		column = "LOWER( " + column + ")";
	    }
	    // now we transform the CASE_ operator into the value
	    // and set the comparison operator
	    if ( cmp_op.equals( CASE_INSENSITIVE_EQUAL ) ) {
		value = s;
		cmp_op = EQUAL;
	    } else {
		value = stringReplace( cmp_op.toLowerCase(), "x", s );
		cmp_op = likeKeyword;
	    }
	}
	whereClauses.addElement( column + " " + cmp_op + " ?" );
	parms.addElement( value );
    }
}

/**
 * Add a where-clause that compares a column against a string value
 * to the SQL command.
 *
 * @param column   The column (field) name to use as a search constraint.
 * @param value    The value to search for.
 * @param cmp_op   The comparison operator to use:  QueryBuilder.EQUAL, ...
 * @deprecated  Use addWhere(RDBColumn column, String value, String cmp_op ) instead.
 */
public void addWhereClause(	String column, String value, String cmp_op )
{
    _addWhereClause( column, value, cmp_op );
}

/**
 * Add a where-clause that compares a column against a string value
 * to the SQL command.  The 3rd and 4th params allow checking to be done
 * on the string value before the where-clause is generated.
 *
 * @param column   The column (field) name to use as a search constraint.
 * @param value    The value to search for.
 * @param nullOk   Whether a value of NULL is acceptable.
 * @param exact    Whether the value must match exactly (no wildcards).
 * @return Whether a where-clause was generated.
 * false if value==null and nullOk==false.
 * @deprecated  Use addWhere(RDBColumn c1, String value, String cmp_op) instead.
 */
/*
public boolean addWhereClause(	String column, String value,
				boolean nullOk, boolean exact )
{

    String append = "";
    boolean added = false;

    if ( value == null ) {
	if ( nullOk ) {
	    append = " " + column + " is null";
	    added = true;
	}
    } else {
        String pattern = value.trim();
        //if ( pattern.length() > 0 && ! pattern.equals( wildcard ) ) {
        if ( ! pattern.equals( wildcard ) ) {
            if ( exact )
                append = " " + column + " = ?";
            else {
                append = " " + column + " " + matchKeyword + " ?";
                fuzzies.set( parms.size() );
	    }
            added = true;
	    parms.addElement( pattern );
        }
    }
    if ( added ) {
	storeTableNameForColumn( column );
        whereClauses.addElement( append );
    }

    return added;
}
*/

/**
 * Add a where-clause to the SQL command.
 *
 * @param column   The column (field) name to use as a search constraint.
 * @param value    The value to search for.
 * @param nullOk   Whether a value of NULL is acceptable.
 * @return Whether a where-clause was generated.
 * false if value==null and nullOk==false.
 * @deprecated  Use addWhere(RDBColumn c1, BigDecimal value, String cmp_op) instead.
 */
public boolean addWhereClause(	String column, BigDecimal value,
				boolean nullOk )
{
    storeTableNameForColumn( column );

    String append = "";
    boolean added = false;

    if ( value == null ) {
	if ( nullOk ) {
	    append = " " + column + " is null";
	    added = true;
	}
    } else {
        append = " " + column + " = ?";
        added = true;
	parms.addElement( value );
    }
    if ( added )
        whereClauses.addElement( append );

    return added;
}

/**
 * Add a where-clause to the SQL command.
 *
 * @param column   The column (field) name to use as a search constraint.
 * @param value    The int value to search for.
 * @deprecated  Use addWhere(RDBColumn column, int value) instead.
 */
public void addWhereClause(	String column, int value )
{
    _addWhereClause( column, new Integer( value ), EQUAL );
}

/**
 * Add a where-clause to the SQL command.
 *
 * @param column   The column (field) name to use as a search constraint.
 * @param value    The int value to search for.
 * @param cmp_op   The comparison operator to use:  QueryBuilder.EQUAL, ...
 * @deprecated  Use addWhere(RDBColumn column, int value, String cmp_op ) instead.
 */
public void addWhereClause(	String column, int value, String cmp_op )
{
    _addWhereClause( column, new Integer( value ), cmp_op );
}

/**
 * Add a where-clause to the SQL command.
 *
 * @param column   The column (field) name to use as a search constraint.
 * @param value    The long value to search for.
 * @deprecated  Use addWhere(RDBColumn column, long value) instead.
 */
public void addWhereClause(	String column, long value )
{
    _addWhereClause( column, new Long( value ), EQUAL );
}

/**
 * Add a where-clause to the SQL command.
 *
 * @param column   The column (field) name to use as a search constraint.
 * @param value    The long value to search for.
 * @param cmp_op   The comparison operator to use:  QueryBuilder.EQUAL, ...
 * @deprecated  Use addWhere(RDBColumn column, long value, String cmp_op ) instead.
 */
public void addWhereClause(	String column, long value, String cmp_op )
{
    _addWhereClause( column, new Long( value ), cmp_op );
}

/**
 * Add a where-clause to the SQL command.
 *
 * @param column   The column (field) name to use as a search constraint.
 * @param value    The value to search for.
 * @deprecated  Use addWhere(RDBColumn column, float value) instead.
 */
public void addWhereClause(	String column, float value )
{
    _addWhereClause( column, new Float( value ), EQUAL );
}

/**
 * Add a where-clause to the SQL command.
 *
 * @param column   The column (field) name to use as a search constraint.
 * @param value    The value to search for.
 * @param cmp_op   The comparison operator to use:  QueryBuilder.EQUAL, ...
 * @deprecated  Use addWhere(RDBColumn column, float value, String cmp_op ) instead.
 */
public void addWhereClause(	String column, float value, String cmp_op )
{
    _addWhereClause( column, new Float( value ), cmp_op );
}

/**
 * Add a where-clause to the SQL command.
 *
 * @param column   The column (field) name to use as a search constraint.
 * @param value    The value to search for.
 * @deprecated  Use addWhere(RDBColumn column, double value) instead.
 */
public void addWhereClause(	String column, double value )
{
    _addWhereClause( column, new Double( value ), EQUAL );
}

/**
 * Add a where-clause to the SQL command.
 *
 * @param column   The column (field) name to use as a search constraint.
 * @param value    The value to search for.
 * @param cmp_op   The comparison operator to use:  QueryBuilder.EQUAL, ...
 * @deprecated  Use addWhere(RDBColumn column, double value, String cmp_op ) instead.
 */
public void addWhereClause(	String column, double value, String cmp_op )
{
    _addWhereClause( column, new Double( value ), cmp_op );
}

/**
 * Add a where-clause to the SQL command.
 *
 * @param column   The column (field) name to use as a search constraint.
 * @param value    The value to search for.
 * @deprecated  Use addWhere(RDBColumn column, boolean value) instead.
 */
public void addWhereClause(	String column, boolean value )
{
    _addWhereClause( column, new Integer( value ? 1: 0 ), EQUAL );
}

/**
 * Add a where-clause to the SQL command.
 *
 * @param column   The column (field) name to use as a search constraint.
 * @param value    The value to search for.
 * @param cmp_op   The comparison operator to use:  QueryBuilder.EQUAL, ...
 * @deprecated  Use addWhere(RDBColumn column, boolean value, String cmp_op ) instead.
 */
public void addWhereClause(	String column, boolean value, String cmp_op )
{
    _addWhereClause( column, new Integer( value ? 1: 0 ), cmp_op );
}

/**
 * Add a trailing clause (order by, group by, etc.) to the SQL command.
 *
 * @param clause     The clause to append to the SQL command.
 */
public void addEndClause( String clause ) {
    endClauses.put( clause, "" );
}

/**
 * Append to the 'select' portion of the SQL command.
 *
 * @param str     The string to append to the SQL command.
 * @deprecated  Use addEndClause() instead.
 */
public void add( String str ) {
    addEndClause( str );
}

/**
 * Sets debug flag to print the SQL statement just before its execution.
 * Affects just this instance of QueryBuilder.
 */
public void debug() {
    debugSQL = true;
}

/**
 * Sets debug flag to print the SQL statement just before its execution.
 * Affects all instances of QueryBuilder.
 */
static public void debugAll() {
    debugAllSQL = true;
}

/**
 * Add DISTINCT keyword to query.
 */
private String distinct = "";
public void distinct() {
    distinct = "distinct ";
}

public void addWhere( String fullClause ) {
    whereClauses.addElement( fullClause );
}

/**
 * QueryBuilder.addWhereIn( RDBColumn, QueryBuilder )
 * allows queries like
 *   select Person.Name from Person where Person.City
 *   in ( select City.oid from City where City.Size > 1000 )
 * with code like
 *   QueryBuilder mainQuery = new QueryBuilder();
 *   mainQuery.select( PersonDO.Name );
 *   QueryBuilder subQuery = new QueryBuilder();
 *   subQuery.select( CityDO.PrimaryKey );
 *   subQuery.addWhere( CityDO.Size, 1000, QueryBuilder.GREATER_THAN );
 *   mainQuery.addWhereIn( PersonDO.City, subQuery );
 *
 * @param field T1DO.T1Y in the example above
 * @param qb subQuery in the example above
 * @author Jay Gunter
 */
private Vector subQueries = new Vector();
public void addWhereIn( RDBColumn field, QueryBuilder qb )
throws QueryException {
    addWhereIn( field, qb, true );
}

/**
 * Note: immediately after creating this method (and making
 * the old addWhereIn call this new method),
 * I implemented the addWhereNot() method.  See 'NOT-note' below.
 *
 * QueryBuilder.addWhereIn( RDBColumn, QueryBuilder, boolean )
 * allows queries like
 *   select Person.Name from Person where Person.City
 *   not in ( select City.oid from City where City.Size > 1000 )
 * with code like
 *   QueryBuilder mainQuery = new QueryBuilder();
 *   mainQuery.select( PersonDO.Name );
 *   QueryBuilder subQuery = new QueryBuilder();
 *   subQuery.select( CityDO.PrimaryKey );
 *   subQuery.addWhere( CityDO.Size, 1000, QueryBuilder.GREATER_THAN );
 *   mainQuery.addWhereIn( PersonDO.City, subQuery, false );
 *
 * NOT-note:  The same thing can be accomplished like so:
 *   QueryBuilder mainQuery = new QueryBuilder();
 *   mainQuery.select( PersonDO.Name );
 *   QueryBuilder subQuery = new QueryBuilder();
 *   subQuery.select( CityDO.PrimaryKey );
 *   subQuery.addWhere( CityDO.Size, 1000, QueryBuilder.GREATER_THAN );
 *   mainQuery.addWhereNot();
 *   mainQuery.addWhereIn( PersonDO.City, subQuery );
 *
 * @param field T1DO.T1Y in the example above
 * @param qb subQuery in the example above
 * @author Jay Gunter
 */
public void addWhereIn( RDBColumn field, QueryBuilder qb, boolean in )
throws QueryException {
    if ( this == qb )
        throw new QueryException(
	    "Recursion detection: subquery cannot be " +
	    "the same object as the main query." );
    storeTableName( field.getTableName() );
    addWhere( field.getFullColumnName() + ( in ? "" : " NOT " ) +
    					" IN ( " + qb.getSQL() + ")" );
    subQueries.addElement( qb );
}

/**
 * QueryBuilder.addWhereIn( RDBColumn, Object[] )
 * allows queries like
 *   select Person.Age from Person
 *   where Person.Name in ( "Bob", "Joe" )
 * with code like
 *   QueryBuilder qb = new QueryBuilder();
 *   qb.select( PersonDO.Age );
 *   String[] values = { "Bob", "Joe" };
 *   qb.addWhereIn( PersonDO.Name, values );
 *
 * @param field PersonDO.Name in the example above
 * @param values array of Objects containing values for IN clause
 * @exception QueryException if values is null or empty
 * @author Jay Gunter
 */
public void addWhereIn( RDBColumn field, Object[] values )
throws QueryException {
    if ( null == values || 0 == values.length )
        throw new QueryException(
	    "IN clause must have at least one value." );
    storeTableName( field.getTableName() );
    String comma = "";
    String qmarks = "";
    for ( int i = 0; i < values.length; i++ ) {
	qmarks += comma + "?";
	comma = ", ";
	parms.addElement( values[i] );
    }
    addWhere( field.getFullColumnName() + " IN ( " + qmarks + ")" );
}

public String getSQL() {
    // Query begins with SELECT <fields> FROM <tablelist>
    String glue = 0 == selectClause.length() ? "" : ", ";
    for ( int i = 0; i < selectedFields.size(); i++ ) {
	RDBColumn field = (RDBColumn) selectedFields.elementAt(i);
	selectClause += glue + field.getFullColumnName();
	storeTableName( mainTableName = field.getTableName() );
	glue = ", ";
    }
    sql = "select " + distinct + selectClause + " from ";

    sql += getTableList();

    sql += getWhereClauses();

    /*
     * ORDER BY code added by Chris Ryan (cryan@plugged.net.au)
     * The following code adds the ORDER BY 'xxx[, xxx]' to the
     * statement for querying.
     */
    int nr_OrderFields = orderFields.size();

    // ORDER BY required?
    if (nr_OrderFields > 0) {

      // yes, fill out the SQL...
      sql += " ORDER BY ";
      for ( int i = 0; i< nr_OrderFields; i++ ) {
            if ( i > 0 )
              sql += ", ";
            sql += orderFields.elementAt(i)+" "+orderDirections.elementAt(i);
            //sql += orderFields.elementAt(i);
      }

    }

    Enumeration endings = endClauses.keys();
    while ( endings.hasMoreElements() )
	sql += endings.nextElement();

    return sql;
}

public String getTableList() {
    String sql = "";
    Enumeration tables = tableNames.keys();
    for ( int i = 0; tables.hasMoreElements(); i++ ) {
	if ( i > 0 )
	    sql += ", ";
	sql += tables.nextElement();
    }
    return sql;
}

public String getWhereClauses() {
    String sql = "";

    // Cleanup WHERE clauses.
    // Discard trailing ORs and OPEN_PARENs.
    String chunk = null;
    int numWhereClauses = whereClauses.size();
    for (; numWhereClauses > 0; numWhereClauses-- ) {
	chunk = (String) whereClauses.elementAt( numWhereClauses - 1 );
	if ( ! ( chunk.equals( OR ) || chunk.equals( OPEN_PAREN ) ) )
	    break;
    }

    // Add WHERE clauses to query.
    int parenDepth = 0;
    boolean prevWas_OpenParen_or_OR = true;
    for ( int i = 0; i < numWhereClauses; i++ ) {
	chunk = (String) whereClauses.elementAt( i );
/* System.err.println("chunk ="+chunk ); */
	if ( 0 == i )
	    sql += " WHERE ";
	if (		chunk.equals( OR ) || chunk.equals( NOT ) ) {
	    prevWas_OpenParen_or_OR = true;
	} else if (	chunk.equals( OPEN_PAREN ) ) {
            if ( ! prevWas_OpenParen_or_OR )
	        sql += " AND ";
	    prevWas_OpenParen_or_OR = true;
	    parenDepth++;
	} else if (	chunk.equals( CLOSE_PAREN ) ) {
	    if ( parenDepth > 0 ) {
		prevWas_OpenParen_or_OR = false;
		parenDepth--;
	    } else {
		chunk = "";
	    }
	} else { // chunk is actual expression for where-clause
            if ( ! prevWas_OpenParen_or_OR )
	        sql += " AND ";
	    prevWas_OpenParen_or_OR = false;
	}
	sql += chunk;
    }

    // Add any needed trailing CLOSE_PARENs.
    while ( parenDepth-- > 0 )
	sql += CLOSE_PAREN;

    return sql;
}

public int executeDelete( Connection conn ) throws SQLException {
    String sql = "delete from " + getTableList() + getWhereClauses();
    if ( debugSQL || debugAllSQL )
	System.out.println("\nQueryBuilder.prepareStatement sql=\n"+
		prettySQL( sql ) +"\n" );
    try {
	stmt = conn.prepareStatement( sql );
	applyParameters( stmt, 0 );
	return stmt.executeUpdate();
    } catch ( SQLException se ) {
	System.err.println( "\n\nSQLException: "+se.getMessage() +
			    ","+ se.getErrorCode()+","+se.getSQLState() );
        throw se;
    }
}

public void applyParameters( PreparedStatement ps, int paramCount )
throws SQLException {
    applyParameters( ps, parms, paramCount );
}

public void applyParameters(	PreparedStatement	ps,
				Vector			parms,
				int			paramCount )
throws SQLException {
    for ( int i = 0; i < parms.size(); i++ ) {
	int n = ++paramCount;
        Object    o = parms.elementAt( i );
	if (debugSQL || debugAllSQL) {
    	    System.out.print("Param: " + i + ", paramCount=" + paramCount);
    	    System.out.println(", Value: " + o);
	}
	if ( o instanceof String ) {
	    /*
	    if ( fuzzies.get( i ) ) {
		if (likeKeyword.equals(DEFAULT_MATCHES_KEYWORD)) {
		    ps.setString(     n, getRegExpMatch( (String) o ) );
		} else {
		    ps.setString(     n, (String)                 o );
		}
	    } else {
		    ps.setString(     n, (String)                 o );
	    }
	    */
		    ps.setString(     n, (String)                 o );
	} else if ( o instanceof java.sql.Date      )  { /*VB 2/2/2000 */
		    ps.setDate(       n, ((java.sql.Date)         o) );
	} else if ( o instanceof java.sql.Timestamp )  { /*VB 2/9/2000 */
		    ps.setTimestamp(  n, ((java.sql.Timestamp)    o) );
        } else if ( o instanceof BigDecimal	    ) {
		    ps.setBigDecimal( n,  (BigDecimal)            o );
	} else if ( o instanceof Integer	    ) {
		    ps.setInt(        n, ((Integer)    o).intValue() );
	} else if ( o instanceof Long	    	    ) {
		    ps.setLong(       n, ((Long)       o).longValue() );
	} else if ( o instanceof Float		    ) {
		    ps.setFloat(      n, ((Float)      o).floatValue() );
	} else if ( o instanceof Double		    ) {
		    ps.setDouble(     n, ((Double)     o).doubleValue());
	} else if ( o instanceof Boolean	    ) {
		    ps.setBoolean(    n, ((Boolean)    o).booleanValue() );
    	} else if ( o instanceof GenericDO ) {
		    ps.setBigDecimal( n, ((GenericDO)  o).getOId().toBigDecimal());
	}
    }
    for ( int i = 0; i < subQueries.size(); i++ ) {
        QueryBuilder subQuery = (QueryBuilder) subQueries.elementAt(i);
	subQuery.applyParameters( ps, paramCount );
    }
}

/**
 * Generate a JDBC PreparedStatement using the values passed
 * for the where-clauses.
 *
 * @param conn   The DBConnection object to use for the search.
 */
private String sql = null;
private void prepareStatement( DBConnection conn )
throws SQLException {
    prepareStatement( conn.getConnection() );
}
private void prepareStatement( Connection conn )
throws SQLException {
    getSQL();
    if ( debugSQL || debugAllSQL )
	System.out.println("\nQueryBuilder.prepareStatement sql=\n"+
		prettySQL( sql ) +"\n" );

    try {
	stmt = conn.prepareStatement( sql );
//	if ( maxRows > 0 )
	    stmt.setMaxRows( maxRows );
    } catch ( SQLException se ) {
	System.err.println("\n\nSQLException: "+se.getMessage() +","+ se.getErrorCode()+","+se.getSQLState() );
        throw se;
    }

    applyParameters( stmt, 0 );
}

private int maxRows = -1;

public void setMaxRows( int x ) {
    if ( 0 > x ) x = 0;
    maxRows = x;
}

private String[]    breakWords = {
	"(", ")", "FROM", "from", "WHERE", "where", "AND", "and", "OR" };

private String indent( int level ) {
    String ret = "";
    for ( int i = 0; i < level; i++ )
	ret += "    ";
    return ret;
}

private String prettySQL( String sql ) {
    int next   = 0;
    int sp     = 0;
    int level  = 0;
    int word   = 0;
    String out = "";
    int pastend = 9999999;
    sql = sql + " ";
    for ( int offset = 0; offset < sql.length(); ) {
	int nearest= pastend;
	int i = 0;
	for ( ; i < breakWords.length; i++ ) {
	    if ( -1 != ( next = sql.indexOf( breakWords[i], offset ) ) ) {
		if ( next < nearest ) {
		    nearest = next;
		    word = i;
		}
	    }
	}
	next = nearest;
	i = word;

	if ( i < breakWords.length && pastend != nearest ) {
		int wordlen = breakWords[i].length();
		if ( 0 == i )
		    level++;
		out += sql.substring( offset, next );
		out += "\n";
		out += indent(level);
		out += sql.substring( next, next + wordlen ) + " ";
		next += wordlen;
		if ( 1 == i )
		    level--;
		offset = next;
	} else {
	    out += indent(level) + sql.substring( offset );
	    break;
	}
    }
    return out;
}


/**
 * Return the JDBC PreparedStatement generated by prepareStatement().
 * @return The JDBC PreparedStatement.
 */
public PreparedStatement getStatement() {
    // Would anyone ever want this?
    return stmt;
}

/**
 * Execute the PreparedStatement and return a JDBC ResultSet.
 * @param conn   The DBConnection object to use for the search.
 * @return The ResultSet object generated by JDBC for the search.
 */
public ResultSet executeQuery( DBConnection conn ) throws SQLException {
    return executeQuery( conn.getConnection() );
}
public ResultSet executeQuery( Connection conn ) throws SQLException {
    prepareStatement( conn );
    rs = stmt.executeQuery();
    return rs;
}

/**
 * Reset the where-clause information.
 * The same fields specified in the QueryBuilder constructor,
 * can be retrieved for a different set of where-clauses.
 */
public void reset() {
    setMaxRows( 0 );

    whereClauses = new Vector(); // keep the select, forget the where-clauses
    parms   = new Vector();	// discard parms for previous where-clause
    fuzzies = new BitSet();	// forget special string handling

    // Clear the ORDER BY Vectors
    orderFields = new Vector();
    orderDirections = new Vector();

    try {
	if ( null != rs )
	    rs.close();
    } catch ( SQLException e ) {
	System.err.println(
	    "QueryBuilder unable to close ResultSet: SQLException=" +
	    e.getMessage() );
    }

    /*
    if ( null != conn )	// if this class created the connection, close it.
	conn.close();
    conn    = null;
    */
}

public void close() {
//System.err.println("IN CLOSE..");
    try {
//System.err.println("dbQuery ="+dbQuery );
	if ( null != dbQuery )
//{
//System.err.println("#######");
//try {
	    dbQuery.release();
//} catch ( Exception e ) {
//System.err.println("@@@@@@@@@@");
//}
//}
//System.err.println("rs ="+rs );
	if ( null != rs )
	    rs.close();
//System.err.println("DONE");
    } catch ( SQLException e ) {
	System.err.println(
	    "QueryBuilder unable to close ResultSet: SQLException=" +
	    e.getMessage() );
    }
}

/**
 * Returns a regular expression, as supported by "MATCHES"
 * that will will perform a case insensitive match of a substring.
 * @param base   The base String to convert to a regular expression.
 * @return       The generated regular expression.
 */
private String getRegExpMatch( String base )
{
    int len = base.length();
    char [] baseChars = new char[len];
    base.getChars(0, len, baseChars, 0);

    // The result will never exceed 5 * the length of
    // the original string.
    char [] result = new char[len * 5];
    int resultIdx = 0;
    for (int idx = 0; idx < baseChars.length; idx++) {
        if (Character.isSpaceChar(baseChars[idx])) {
            // Convert spaces to ' *'
            if ((resultIdx > 0) && (result[resultIdx-1] != '*')) {
                result[resultIdx++] = ' ';
                //result[resultIdx++] = '*';
                result[resultIdx++] = '%';
            }
        }
        else if (Character.isLetter(baseChars[idx])) {
            // Convert letters to [Aa]
            result[resultIdx++] = '[';
            result[resultIdx++] = Character.toUpperCase(baseChars[idx]);
            result[resultIdx++] = Character.toLowerCase(baseChars[idx]);
            result[resultIdx++] = ']';
        }
        else {
            // Leave everything else as is.
            result[resultIdx++] = baseChars[idx];
        }
    }
    // Match anything at end of line
    result[resultIdx++] = '*';
    return new String(result);
}

private String mainTableName;			// prepended to column names
private Hashtable tableNames = new Hashtable();	// table names in FROM clause
private Hashtable endClauses = new Hashtable();	// clauses appended to SQL
private Vector selectedFields = new Vector();	// RDBColumns to return
private String selectClause = "";		// comma separated fields
private Vector whereClauses = new Vector();	// WHERE clauses
private Vector orderFields = new Vector();	// fields for ORDER BY clauses
private Vector orderDirections = new Vector();	// ascending/descending
private PreparedStatement stmt;			// what we execute
private Vector parms;				// parameters for stmt
private BitSet fuzzies;				// which parms use regex match
private int param = 0;				// parameter index
//private DBConnection conn = null;		// for execution of stmt



/**
 * Set the string matching details for inexact string matches
 * See dods.conf for details.
 * Code added by Chris Ryan (cryan@plugged.net.au)
 *
 * @param match_keyword  Whether the DB can use "MATCHES" or "LIKE"
 * @param wildcard  The wildcard character for string searches
 */
public void setStringMatchDetails( String match_keyword, String wildcard ) {
  this.likeKeyword = match_keyword;
  this.wildcard = wildcard;

}




/**
 * Add a where-clause that compares a column against a date value
 * to the SQL command.
 * Contributed by: Victor Brilon
 *
 * @param column   The column (field) name to use as a search constraint.
 * @param value    The value to compare against.
 * @param cmp_op   The comparison operator to use:  QueryBuilder.EQUAL, ...
 * @deprecated  Use addWhere(RDBColumn c1, Date value, String cmp_op) instead.
 */
public void addWhereClause(	String column, java.sql.Date date,
				String cmp_op )
{
    _addWhereClause( column, date, cmp_op );
}

public void addWhereClause(	String column, java.sql.Timestamp date,
				String cmp_op )
{
    _addWhereClause( column, date, cmp_op );
}

public void addWhereClause(	String column, java.sql.Time date,
				String cmp_op )
{
    _addWhereClause( column, date, cmp_op );
}

/**
 * Add another field to the ORDER BY list.
 * NOTE: is used by SourceGenerator_Query and writeMemberStuff.template.
 *
 * This method added by Chris Ryan (cryan@plugged.net.au) to
 * allow easier access to the ORDER BY part of the SQL query.
 *
 * @param column     The column to order by
 * @param direction  ASCENDING or DESCENDING order
 */
public void addOrderByColumn(String column, String direction) {
  orderFields.addElement(column);
  orderDirections.addElement(direction);
}

/**
 * Add another field to the ORDER BY list.
 * NOTE: is used by SourceGenerator_Query and writeMemberStuff.template.
 *
 * This method added by Chris Ryan (cryan@plugged.net.au) to
 * allow easier access to the ORDER BY part of the SQL query.
 *
 * @param column     The column to order by
 * @param direction  ASCENDING or DESCENDING order
 */
public void addOrderByColumn( RDBColumn column, String direction ) {
  orderFields.addElement( column.getFullColumnName() );
  orderDirections.addElement( direction );
}

public void setDatabaseVendor( String vendor ) {
    if (      vendor.toLowerCase().equals( "informix" ) ) {
	likeKeyword = "MATCHES";
	wildcard = "*";
    } else if ( vendor.toLowerCase().equals( "sybase" ) ) {
	likeKeyword = "LIKE";
	wildcard = "%";
    } else if ( vendor.toLowerCase().equals( "oracle" ) ) {
	likeKeyword = "LIKE";
	wildcard = "%";
    } else if ( vendor.toLowerCase().equals( "msql" ) ) {
	likeKeyword = "LIKE";
	wildcard = "%";
    } else if ( vendor.toLowerCase().equals( "standard" ) ) {
	likeKeyword = "LIKE";
	wildcard = "%";
    } else if ( vendor.toLowerCase().equals( "postgresql" ) ) {
	likeKeyword = "LIKE";
	wildcard = "%";
    } else if ( vendor.toLowerCase().equals( "instantdb" ) ) {
	likeKeyword = "LIKE";
	wildcard = "%";
    } else {
        System.out.println(
                "ERROR in QueryBuilder!  Unimplemented Database Vendor: " +
                vendor );
    }
}

////////////////////////////////////////////////////////////////////////////////
static private final int op_EQUAL = 1;
static private final int op_NOT_EQUAL = 2;
static private final int op_LESS_THAN = 3;
static private final int op_LESS_THAN_OR_EQUAL = 4;
static private final int op_GREATER_THAN = 5;
static private final int op_GREATER_THAN_OR_EQUAL = 6;
static private final int op_IS_NULL = 7;
static private final int op_IS_NOT_NULL = 8;
static private final int op_CASE_SENSITIVE_CONTAINS = 9;
static private final int op_CASE_INSENSITIVE_CONTAINS = 10;
static private final int op_CASE_SENSITIVE_STARTS_WITH = 11;
static private final int op_CASE_INSENSITIVE_STARTS_WITH = 12;
static private final int op_CASE_SENSITIVE_ENDS_WITH = 13;
static private final int op_CASE_INSENSITIVE_ENDS_WITH = 14;
static private final int op_CASE_INSENSITIVE_EQUAL = 15;

static Hashtable cmp_ops = new Hashtable();
static {
    cmp_ops.put( EQUAL, new Integer( op_EQUAL ) );
    cmp_ops.put( NOT_EQUAL, new Integer( op_NOT_EQUAL ) );
    cmp_ops.put( LESS_THAN, new Integer( op_LESS_THAN ) );
    cmp_ops.put( LESS_THAN_OR_EQUAL, new Integer( op_LESS_THAN_OR_EQUAL ) );
    cmp_ops.put( GREATER_THAN, new Integer( op_GREATER_THAN ) );
    cmp_ops.put( GREATER_THAN_OR_EQUAL, new Integer( op_GREATER_THAN_OR_EQUAL ) );
    cmp_ops.put( IS_NULL, new Integer( op_IS_NULL ) );
    cmp_ops.put( IS_NOT_NULL, new Integer( op_IS_NOT_NULL ) );
    cmp_ops.put( CASE_SENSITIVE_CONTAINS, new Integer( op_CASE_SENSITIVE_CONTAINS ) );
    cmp_ops.put( CASE_INSENSITIVE_CONTAINS, new Integer( op_CASE_INSENSITIVE_CONTAINS ) );
    cmp_ops.put( CASE_SENSITIVE_STARTS_WITH, new Integer( op_CASE_SENSITIVE_STARTS_WITH ) );
    cmp_ops.put( CASE_INSENSITIVE_STARTS_WITH, new Integer( op_CASE_INSENSITIVE_STARTS_WITH ) );
    cmp_ops.put( CASE_SENSITIVE_ENDS_WITH, new Integer( op_CASE_SENSITIVE_ENDS_WITH ) );
    cmp_ops.put( CASE_INSENSITIVE_ENDS_WITH, new Integer( op_CASE_INSENSITIVE_ENDS_WITH ) );
    cmp_ops.put( CASE_INSENSITIVE_EQUAL, new Integer( op_CASE_INSENSITIVE_EQUAL ) );
}

/**
 * Wrapper for compare() method that takes Objects for arguments.
 * This method takes doubles, and so will handle
 * shorts, ints, longs, floats, doubles, but not booleans.
 *
 * @return true if a compares to b according to cmp_op
 * @param a the DO member value
 * @param b the value to compare against
 * @exception QueryException if cmp_op is not a valid comparison operator.
 * @see compare( Object a, Object b, String cmp_op )
 * @author Jay Gunter
 */
static public boolean compare( double a, double b, String cmp_op )
throws QueryException {
    return compare( new Double( a ), new Double( b ), cmp_op );
}

/**
 * Wrapper for compare() method that takes Objects for arguments.
 * This method takes booleans.
 *
 * @return true if a compares to b according to cmp_op
 * @param a the DO member value
 * @param b the value to compare against
 * @exception QueryException if cmp_op is not a valid comparison operator.
 * @see compare( Object a, Object b, String cmp_op )
 * @author Jay Gunter
 */
static public boolean compare( boolean a, boolean b, String cmp_op )
throws QueryException {
    return compare( new Double( a ? 1 : 0 ), new Double( b ? 1 : 0 ), cmp_op );
}

/**
 * Compare two values according to a QueryBuilder comparison operator.
 * This method has some general applicability.
 * It is handy for writing sorting algorithms, etc.
 *
 * The Query classes generated by DODS use this method
 * to prune undesired values from the cache held by a DO class.
 *
 * @return true if a compares to b according to cmp_op
 * @param a the DO member value
 * @param b the value to compare against
 * @exception QueryException if cmp_op is not a valid comparison operator.
 * @author Jay Gunter
 */
static public boolean compare( Object a, Object b, String cmp_op )
throws QueryException {
    Integer cmp = (Integer) cmp_ops.get( cmp_op );
    if ( null == cmp )
	throw new QueryException(
	    "Unrecognized comparison operator: " + cmp_op );

    int op = cmp.intValue();
    switch ( op ) {
	case op_IS_NULL:
	    return null == b;
	case op_IS_NOT_NULL:
	    return null != b;

	case op_EQUAL:
	    if ( null == a && null == b )
		return true;
	    if ( null == a || null == b )
		return false;
	    byte[] z = {};
	    if ( a.getClass() == z.getClass() ) {
		byte[] ab = (byte[]) a;
		byte[] bb = (byte[]) b;
		if ( ab.length != bb.length )
		    return false;
		for ( int i = 0; i < ab.length; i++ ) {
		    if ( ab[i] != bb[i] )
			return false;
		}
		return true;
	    } else if ( a instanceof GenericDO ) {
		String aoid = ( (GenericDO) a ).getOId().toString();
		String boid = ( (GenericDO) b ).getOId().toString();
		boolean equals = aoid.equals( boid );
		if ( equals && a != b )
		    System.err.println(
			"\n\nDATA OBJECT WARNING: Two DOs with the same OID: "
			+ aoid );
		return equals;
	    } else
		return a.equals( b );

	case op_NOT_EQUAL:
	    return ! compare( a, b, EQUAL );

	case op_GREATER_THAN:
	    return ! compare( a, b, LESS_THAN_OR_EQUAL );
	case op_GREATER_THAN_OR_EQUAL:
	    return ! compare( a, b, LESS_THAN );

	case op_LESS_THAN:
	case op_LESS_THAN_OR_EQUAL:
	    if ( null == a || null == b )
		return false;
	    if ( a instanceof String ) {
		String aStr = (String) a;
		String bStr = null;
		if ( ! ( b instanceof String ) )
		    bStr = b.toString();
		int x = aStr.compareTo( bStr );
		return ( -1 == x || ( 0 == x && op_LESS_THAN_OR_EQUAL==op ) );
	    } else if (  a instanceof java.util.Date ) {
		if ( ! ( b instanceof java.util.Date ) )
		    throw new QueryException(
			"Second arg for " + cmp_op + " is not a Date" );
		java.util.Date ad = (java.util.Date) a;
		java.util.Date bd = (java.util.Date) b;
		return (   ad.before( bd ) ||
			 ( ad.equals( bd ) && op_LESS_THAN_OR_EQUAL==op ) );
	    } else if (  a instanceof BigDecimal ) {
		if ( ! ( b instanceof BigDecimal ) )
		    throw new QueryException(
			"First arg for " + cmp_op + " is BigDecimal, " +
			"but second arg is not." );
		int x = ( (BigDecimal) a ).compareTo( b );
		return ( -1 == x || ( 0 == x && op_LESS_THAN_OR_EQUAL==op ) );
	    } else if (  a instanceof Number ) {
		if ( ! ( b instanceof Number ) )
		    throw new QueryException(
			"First arg for " + cmp_op + " is Number, " +
			"but second arg is not." );
		double ad = ( (Number) a ).doubleValue();
		double bd = ( (Number) b ).doubleValue();
		return ( ad < bd || ( ad == bd && op_LESS_THAN_OR_EQUAL==op ) );
	    } else {
		throw new QueryException(
		    "Comparison operator " + cmp_op +
		    " not supported for objects of type " +
		    a.getClass().getName() );
	    }

	case op_CASE_INSENSITIVE_EQUAL:
	    if ( null == a && null == b )
		return true;
	    if ( null == a || null == b )
		return false;
	    String aStr = (String) a;
	    String bStr = (String) b;
	    return aStr.equalsIgnoreCase( bStr );

	case op_CASE_SENSITIVE_CONTAINS:
	case op_CASE_INSENSITIVE_CONTAINS:
	case op_CASE_SENSITIVE_STARTS_WITH:
	case op_CASE_INSENSITIVE_STARTS_WITH:
	case op_CASE_SENSITIVE_ENDS_WITH:
	case op_CASE_INSENSITIVE_ENDS_WITH:
	    if ( ! ( a instanceof String ) )
		throw new QueryException(
		    "First arg for " + cmp_op + " is not a string." );
	    if ( ! ( b instanceof String ) )
		throw new QueryException(
		    "Second arg for " + cmp_op + " is not a string." );
	    String as = (String) a;
	    String bs = (String) b;
	    if ( -1 != cmp_op.indexOf( 'x' ) ) {
		as = as.toLowerCase();
		bs = bs.toLowerCase();
	    }
	    switch ( op ) {
		case op_CASE_SENSITIVE_CONTAINS:
		case op_CASE_INSENSITIVE_CONTAINS:
		    return -1 != as.indexOf( bs );
		case op_CASE_SENSITIVE_STARTS_WITH:
		case op_CASE_INSENSITIVE_STARTS_WITH:
		    return as.startsWith( bs );
		case op_CASE_SENSITIVE_ENDS_WITH:
		case op_CASE_INSENSITIVE_ENDS_WITH:
		    return as.endsWith( bs );
	    }
	    throw new QueryException(
		"QueryBuilder.compare bug in CASE_ processing" );

	default:
	    throw new QueryException(
		"Unchecked comparison operator: " + cmp_op );
    }
//    return false; // should never get here
}













//===================

private String		updateComma		= "";
private String		updateSet		= "";
private Vector		updateParms		= new Vector();
private String	updateTableName	= null;
private RDBColumn	lastUpdateColumn	= null;

private void _addUpdate( RDBColumn column, Object value )
throws QueryException {
    if ( null != updateTableName
    	 &&   !  updateTableName.equals( column.getTableName()))
	throw new QueryException(
	    "Usage error: QueryBuilder.addUpdate():" +
	    " column " + column.getFullColumnName() +
	    " not in table " + updateTableName + "." );
    if ( null == value && column.notNull() )
	throw new QueryException(
	    "Usage error: QueryBuilder.addUpdate():" +
	    " column " + column.getFullColumnName() +
	    " does not allow NULLs." );

    updateTableName	= column.getTableName();
    updateSet		+= updateComma + column.getColumnName() + " = ?";
    updateComma		= ", ";
    updateParms.addElement( value );
}

public int executeUpdate( Connection conn )
throws QueryException, SQLException {
    if ( null == updateTableName ) {
	/*
    	throw new QueryException(
	    "Usage error:  QueryBuilder.executeUpdate(): " +
	    "must call addUpdate to set values to change." );
    	*/
	return 0;
    }
    String sql = "UPDATE "	+ updateTableName +
    		 " SET "	+ updateSet + getWhereClauses();
    if ( debugSQL || debugAllSQL )
	System.out.println("\nQueryBuilder.prepareStatement sql=\n"+
		prettySQL( sql ) +"\n" );
    try {
	stmt = conn.prepareStatement( sql );
	int paramCount = 0;
	applyParameters( stmt,	updateParms,	0 );
	applyParameters( stmt,	parms,		updateParms.size() );
	return stmt.executeUpdate();
    } catch ( SQLException se ) {
	System.err.println( "\n\nSQLException: "+se.getMessage() +
			    ","+ se.getErrorCode()+","+se.getSQLState() );
        throw se;
    }
}






/**
 * Add a where-clause that compares a column against a String.
 * @param column  A column (field) name to use as a search constraint.
 * @param value   int value to compare against.
 *
 * @author Jay Gunter
 */
public void addUpdate( RDBColumn column, GenericDO value )
throws QueryException {
    BigDecimal id = null;
    if ( null != value )
        id = value.getOId().toBigDecimal();
    _addUpdate( column, id );
}

/**
 * Add a where-clause that compares a column against a byte array.
 * @param column  A column (field) name to use as a search constraint.
 * @param value   byte array to compare against.
 *
 * @author Jay Gunter
 */
public void addUpdate( RDBColumn column, byte[] value )
throws QueryException {
    String v = null;
    if ( null != value )
	v = new String( value );
    _addUpdate( column, v );
}

/**
 * Add a where-clause that compares a column against a String.
 * @param column  A column (field) name to use as a search constraint.
 * @param value   int value to compare against.
 *
 * @author Jay Gunter
 */
public void addUpdate( RDBColumn column, java.sql.Time value )
throws QueryException {
    _addUpdate( column, value );
}

/**
 * Add a where-clause that compares a column against a String.
 * @param column  A column (field) name to use as a search constraint.
 * @param value   int value to compare against.
 *
 * @author Jay Gunter
 */
public void addUpdate( RDBColumn column, java.sql.Timestamp value )
throws QueryException {
    _addUpdate( column, value );
}

/**
 * Add a where-clause that compares a column against a String.
 * @param column  A column (field) name to use as a search constraint.
 * @param value   long value to compare against.
 *
 * @author Jay Gunter
 */
public void addUpdate( RDBColumn column, long value )
throws QueryException {
    _addUpdate( column, new Long( value ) );
}

/**
 * Add a where-clause that compares a column against a String.
 * @param column  A column (field) name to use as a search constraint.
 * @param value   double value to compare against.
 *
 * @author Jay Gunter
 */
public void addUpdate( RDBColumn column, double value )
throws QueryException {
    _addUpdate( column, new Double( value ) );
}

/**
 * Add a where-clause that compares a column against a String.
 * @param column  A column (field) name to use as a search constraint.
 * @param value   float value to compare against.
 *
 * @author Jay Gunter
 */
public void addUpdate( RDBColumn column, float value )
throws QueryException {
    _addUpdate( column, new Float( value ) );
}

/**
 * Add a where-clause that compares a column against a String.
 * @param column  A column (field) name to use as a search constraint.
 * @param value   BigDecimal value to compare against.
 *
 * @author Jay Gunter
 */
public void addUpdate( RDBColumn column, BigDecimal value )
throws QueryException {
    _addUpdate( column, value );
}

/**
 * Add a where-clause that compares a column against a String.
 * @param column  A column (field) name to use as a search constraint.
 * @param value   int value to compare against.
 *
 * @author Jay Gunter
 */
public void addUpdate( RDBColumn column, int value )
throws QueryException {
    _addUpdate( column, new Integer( value ) );
}

/**
 * Add a where-clause that compares a column against a String.
 * @param column  A column (field) name to use as a search constraint.
 * @param value   boolean value to compare against.
 *
 * @author Jay Gunter
 */
public void addUpdate( RDBColumn column, boolean value )
throws QueryException {
    _addUpdate( column, new Boolean( value ) );
}

/**
 * Add a where-clause that compares a column against a String.
 * @param column  A column (field) name to use as a search constraint.
 * @param value   String value to compare against.
 *
 * @author Jay Gunter
 */
public void addUpdate( RDBColumn column, String value )
throws QueryException {
    _addUpdate( column, value );
}

/**
 * Add a where-clause that compares a column against a Date.
 * @param column  A column (field) name to use as a search constraint.
 * @param value   Date value to compare against.
 *
 * @author Jay Gunter
 */
public void addUpdate( RDBColumn column, java.sql.Date value )
throws QueryException {
    _addUpdate( column, value );
}

} // end QueryBuilder class
