/*
 * Enhydra Java Application Server Project
 *
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific terms governing rights and limitations
 * under the License.
 *
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 * $Id: SourceGenerator_SQL.java,v 1.1 2001/09/28 16:01:06 cryd0221 Exp $
 */

package org.enhydra.dods.builder.generator.sql;

import org.enhydra.dods.store.doa.*;
import org.enhydra.dods.gui.tree.*;

import org.enhydra.dods.builder.engine.*;
import org.enhydra.dods.builder.generator.base.*;

import java.io.File;
import java.io.StringWriter;
import java.util.*;

public class SourceGenerator_SQL extends SourceGenerator {

	final String REFERENCES	= "REFERENCES";	// FOREIGN KEY constraint
	final String KEYNAME		= "KEYNAME";		// PRIMARY KEY
	final String DEF_VALUE	= "DEF_VALUE";
	final String NOT_NULL	= "NOT_NULL";
	final String INDEXES	= "INDEXES";
	final String REFERENCE_TYPE	= "REFERENCE_TYPE";
	final String ON_DELETE_CASCADE	= "ON_DELETE_CASCADE";
	final String PRIMARY_KEY_CLAUSE = "PRIMARY_KEY_CLAUSE";
	final String FOREIGN_KEY_REFERENCES_CLAUSE = "FOREIGN_KEY_REFERENCES_CLAUSE";

	final String OID_DB_TYPE	= "OIDtype";
	final String COMMENT_DB_VENDOR = "COMMENT_DB_VENDOR";

    // Variables used to build the value of the INDEXES tag.
	private String tablename;
	private String indexes;
	private int    index_num;
	private boolean comments = true; // add comments to .sql files

	private DODSProject project = null;

	public SourceGenerator_SQL() throws Exception {
	}

    /**
     * Adds tag values to the <CODE>TagSet</CODE>.
     * This method invokes
     *    <CODE>TemplateFilter.process( tagSet )</CODE>
     * which will create the output source file.
     *
     * @param o
     *   DataObject for which source code will be generated by
     *    <CODE>writeFileHeader</CODE> and
     *    <CODE>writeFileFooter</CODE>
     * @exception Exception
     *   If an unrecoverable error occurs.
     */
	public void objectPrep(	GeneratorInfo sgi, DODSProject proj, DataObject o )	throws Exception {
		super.objectPrep( sgi, proj, o );
		project = proj;
		indexes = "";
		index_num = 1;
		tablename = o.getDBTableName();
		tagSet.setTagValue( REFERENCE_TYPE, oidDbType );
	}

	public void memberPrep(	Attribute a ) throws Exception {
		super.memberPrep( a );

		tagSet.setTagValue( KEYNAME, "" );
		// In this implementation, oid is the primary key of every table.
		//tagSet.setTagValue( KEYNAME,	a.isIndex()   ? "PRIMARY KEY" : "" );

		tagSet.setTagValue(NOT_NULL, db_vendor.equals("Sybase") ? "NULL" : "");
		if (! a.canBeNull()) tagSet.setTagValue( NOT_NULL, "NOT NULL" );
		String defValue = a.getInitialValue();
		if ( a.getJavaType().equals( "String" ) ) {
			/*
			 * The StringQuoteCharacter was added by
			 * Chris Ryan (cryan@plugged.net.au).
			 * There does not appear to be a standard string quote character
			 * amongst * the DB vendors.
			 * Most appear to support ["] but Sybase really objects
			 * to this.  It requires ['].
			 */
			// Use the string quote character for this db type.
			defValue =	val( STRING_QUOTE_CHARACTER ) + defValue + val( STRING_QUOTE_CHARACTER );
		}

	  	/*
		 * The Java type boolean needs a specific mapping to the
		 * DB BIT type.  False is 0, True is 1.
		 * Code added by Chris Ryan (cryan@plugged.net.au).
		 */
		if (a.getJavaType().equals("boolean")) {
			if ( ! db_vendor.equals( "PostgreSQL" ) ) {
				// convert the default value
				if (defValue.equalsIgnoreCase("true")) {
					defValue = "1";
				} else if (defValue.equalsIgnoreCase("false")) {
					defValue = "0";
				} else {
					// assume false and notify
					defValue = "0";
			 		Builder.print("Invalid initial value for "+a.getName()+", set to 0 (false)");
				}
			}
		}

		//default of "" caused problems in some databases
		//Open Source contribution by Joel
		if ( null != defValue && 0 != defValue.length() && !"".equals(defValue))
		defValue = "DEFAULT " + defValue;
		tagSet.setTagValue( DEF_VALUE,  defValue );

		// If necessary, append (size) to database type.
		String size  = a.getSize();
		String jtype = a.getJavaType();
		if ( jtype.equals( "java.math.BigDecimal" ) ) {
			/* FIX? See SourceGenerator_DO */
		} else if ( jtype.equals( "String" ) ) {
			String dbtype = tagSet.getTagValue( MEMBER_DB_TYPE );
		 	if ( ! columnIsText( a ) ) {
				if ( null == size || 0 == size.length() ) size = "32";
				dbtype += "(" + size + ")";
			}
			tagSet.setTagValue( MEMBER_DB_TYPE, dbtype );
//System.out.println ("Setting MEMBER_DB_TYPE @pt3:"+dbtype);
		}

		if ( a.isIndex() ) {
			//using ON conforms to standard SQL, fixes problems with some db's
			//forgot in increment counter - Open Source contribution by Joel Boehland
			indexes += "CREATE INDEX I" + index_num + "_" + tablename +	" ON " + tablename + " ( " + a.getName() + " );\n";
			index_num++;
		}

		tagSet.setTagValue( ON_DELETE_CASCADE, "" );
//System.err.println("a.getName()="+a.getName());
//System.err.println("a.isForeignKey()="+a.isForeignKey());
//System.err.println("a.getForeignKeyColumnName()="+a.getForeignKeyColumnName());
		// "foreign keys" refer to user-defined primary keys;
		// they are not handled as ordinary refConstraints which
		// are known to use OIDs.
		// hack around InstantDB lack of foreign key support.
		if ( a.isObjectRef() && a.isRefConstraint() && ! a.isForeignKey() && ! db_vendor.equals( "InstantDB" )) {
			 /*
			  * When the user defined Attribute 'A'
			  * of Data Object 'D1' (table 'd1'),
			  * he marked the Attribute (class member, table column)
			  * as a referential constraint.
			  * The Attribute is an reference to
			  * some other Data Object 'D2' (table 'd2').
			  * This means that when column 'xx' of table 'd1' is not NULL,
			  * the value must be a valid index into table 'd2'.
			  * Within the 'CREATE TABLE' SQL command for table 'd1',
			  * this referential constraint is specified as part of
			  * the definition of column 'xx'.  E.g.:
			  *    CREATE TABLE d1 (
			  *	  xx DECIMAL(19,0) REFERENCES d2(oid)
			  *
			  * We must learn the name of the table ('d2') that
			  * corresponds to the class 'D2'.
			  * So, 'D2' must be a Data Object that is defined in
			  * the DODSProject that we are processing.
			  */
			tagSet.setTagValue( MEMBER_DB_TYPE, oidDbType );

			tagSet.setTagValue( REFERENCES, "REFERENCES " +
			getTableName( a ) + " ( " + val( OID_DB_COL_NAME ) + " )" );

			String constraintDef = "";
			if ( namedConstraint ) {
				// database allows constraint notation
				String cn = constraintName( tablename, getTableName( a ) );
				constraintDef = "CONSTRAINT " + "\"" + cn + "\" ";
			}
			tagSet.setTagValue( REFERENCES, constraintDef + "REFERENCES " + getTableName( a ) + " ( " + val( OID_DB_COL_NAME ) + " )" );

		 	if ( cascadeDelete ) tagSet.setTagValue( ON_DELETE_CASCADE, "ON DELETE CASCADE" );
		} else {
			tagSet.setTagValue( REFERENCES, "" );
		}

	}

	static private int maxDbIdLen = 30;
	// FIX get from dods.conf
	// FIX issue a warning if any table or column name exceeds maxDbIdLen

	private int constraintNumber = 0;

	private String constraintName( String table1, String table2 ) {
	  // e.g.  CONSTRAINT "table1_REFS_table2"
	  // Oracle limits identifiers
	  // (table, column and constraint names) to 30 chars.
	  String constraintName = null;
	  /*
	  constraintName = table1 + "_REFS_" + table2;
	  if ( maxDbIdLen < constraintName.length() )
		constraintName = shortName( table1 ) + "_REFS_" +
					  shortName( table2 );
	  if ( maxDbIdLen < constraintName.length() )
		constraintName = shortName( table1 ) + "_" +
					shortName( table2 );
	  if ( maxDbIdLen < constraintName.length() )
	*/
//		constraintName = "constraint_" + constraintNumber++;
		constraintName = "C" + (constraintNumber++) + "_" + table1;
	  return constraintName;
	}

	private String shortName( String n ) {
	  String ret = n.substring( 0, 1 );
		 for ( int i = 1; i < n.length(); i++ ) {
		char ch = n.charAt( i );
		// discard lowercase vowels
		if ( Character.isUpperCase( ch ) || -1 == "aeiou_".indexOf( ch ) )
			  ret += ch;
		 }
	  return ret;
	}


	private String getTableName( Attribute a ) throws Exception {
	// FIX?  we are searching the entire project because
	// Attribute does not have a method to return the
	// DataObject to which it belongs.
	// Rather inelegant.
		String table = null;
		String classname = (String) a.getJavaType();
	// classname is something like pkgX.AaaDO
		int    len = classname.length();
		String last2 = classname.substring( len - 2, len );
/*		 System.err.println("getTableName: a.getName()="+a.getName()); */
	/* System.err.println("last2 ="+last2 ); */
		if ( ! last2.equals( "DO" ) ) {
		// The Attribute is not a Data Object created by DODS,
		// so we don't know what table it references.
		throw new Exception(
			"ERROR:  Attribute '" + a.getName() +
			"':  type '" + classname + "' was not created by DODS.\n" +
			"	  The corresponding table is not known." );
		}
		classname = classname.substring( 0, len - 2 );
	// classname is something like pkgX.Aaa
		int lastDot = classname.lastIndexOf( '.' );
		classname = classname.substring( lastDot + 1, classname.length() );
	// classname is something like Aaa.
	// Now we can go looking for the Data Object named 'Aaa'.
	// When we find it, we return the table name for that Data Object.
	/* System.err.println("classname ="+classname ); */
		DOTreeNode treeRoot = new DOTreeNode( project, true );
		Vector children = treeRoot.getChildren();
		for ( int k = 0; k < children.size(); k++ ) {
		DOTreeNode node = (DOTreeNode) children.elementAt(k);
		if ( null != ( table = searchChildrenForTable( node, classname ) ) )
			return table;
		}
		return table;
	}

	private String searchChildrenForTable( DOTreeNode node, String classname )
	throws Exception {
		String table = null;
		DataObject DO = node.getDataObject();
	/* System.err.println("DO.getName() ="+DO .getName()); */
		if ( ! ( DO instanceof DODSProject || DO instanceof DOPackage ) ) {
		if ( DO.getName().equals( classname ) )
			return DO.getDBTableName();
		}
		Vector children = node.getChildren();
		if ( null != children ) {
		for ( int k = 0; k < children.size(); k++ ) {
			table = searchChildrenForTable(
			(DOTreeNode) children.elementAt( k ),
			classname );
			if ( null != table )
			return table;
		}
		}
		return table;
	}

	public void writeFileHeader(  DataObject o ) throws Exception {
	  if ( "NONE".equals( tagSet.getTagValue( BEGIN_COMMENT ) ) )
		comments = false;
	  tagSet.setTagValue ( COMMENT_DB_VENDOR,
		  comments ?  getBeginCommentMark() +
				"This SQL was generated for a " +
				tagSet.getTagValue( DB_VENDOR ) + " database." +
				getEndCommentMark()
					: "" );
		defaultWriteFileHeader( o );

		Vector ancestors = o.getAncestors();
	/* System.err.println("ancestors.size()="+ancestors.size()); */
		if ( null != ancestors ) {
		for ( int i = 0; i < ancestors.size(); i++ ) {
			DataObject p = (DataObject) ancestors.elementAt( i );
		/* System.err.println("p.getName()="+p.getName()); */
			if ( comments )
			output.write(
			  getBeginCommentMark() + " class " + p.getName() + " " +
			  getEndCommentMark() + "\n" );
			if ( null != p ) {
			Vector members = p.getAttributes();
			for ( int j = 0; j < members.size(); j++ ) {
				Attribute attr = (Attribute) members.elementAt( j );
			/* System.err.println("attr.getName()="+attr.getName()); */
				memberPrep( attr );
				writeMemberStuff( attr );
			}
			}
		}
		}

	 if ( comments )
		  output.write(	getBeginCommentMark() +
		" class " + o.getName() + " " + getEndCommentMark() + "\n" );
	}

	public void writeMemberStuff( Attribute  a ) throws Exception {
		if ( a.isConstant() )
		return;

	// FIX should this warning be in SourceGenerator_DO as well?
	// It is not as necessary there because the generated DO code
	// will not compile, but that's an ugly way to alert the user.
		String name = a.getName().toLowerCase();
		if ( name.equals( "oid" ) || name.equals( "version" ) )
		throw new Exception( "ERROR:  the field " + name +
			" is already defined by the class GenericDO." );

		if ( a.isArray() ) {
		// FIX complete implementation?
		// arrays are handled as separate tables
		}
		else {
		defaultWriteMemberStuff( a );
		}
	}

    public String getForeignKeySQL( DataObject o ) {
	Hashtable fkByDO = o.getForeignKeysGroupedByReferencedDO();
	// FOREIGN KEY( e_fn, e_ln ) REFERENCES person( p_fn, p_ln )
	String ret = "";
	for ( Enumeration e = fkByDO.keys(); e.hasMoreElements(); ) {
		 DataObject referencedDO = (DataObject) e.nextElement();
		 Hashtable fkGroups = (Hashtable) fkByDO.get( referencedDO );
		 for ( Enumeration e2 = fkGroups.keys(); e2.hasMoreElements(); ) {
		String fkGroup = (String) e2.nextElement();
//System.err.println("fkGroup ="+fkGroup );
		Vector v = (Vector) fkGroups.get( fkGroup );
		int numColumnsInFK = v.size();
//System.err.println("numColumnsInFK ="+numColumnsInFK );
		int numColumnsInPK = referencedDO.getPrimaryKeys().length;
//System.err.println("numColumnsInPK ="+numColumnsInPK );
		if ( numColumnsInFK != numColumnsInPK ) {
			 if ( numColumnsInFK < numColumnsInPK )
			Builder.print( "ERROR: " +
				 o.getName() + " is missing columns needed to " +
				 "complete a reference to " +referencedDO.getName() +
				 "." );
			 else
			Builder.print( "ERROR: " +
				 o.getName() + " has TOO MANY columns in its " +
				 "reference to " + referencedDO.getName() +
				 ".  If you have 2 foreign keys that reference " +
				 "the same table, you probably need to add " +
				 "foreignKeyGroup values to your <referenceObject> "+
				 "tags in your .doml file, or make sure the "+
				 "grouping is correct." );
			 // FIX abort the build somehow? Add abort() to Builder?
		}
		if ( 0 < numColumnsInFK ) {
			 String fkList = "";
			 String pkList = "";
			 String comma = "";
			 for ( int i = 0; i < numColumnsInFK; i++ ) {
			String[] pk_fk_pair = (String[]) v.elementAt(i);
			pkList += comma + pk_fk_pair[0];
			fkList += comma + pk_fk_pair[1];
			comma = ", ";
			 }
			 ret += ", FOREIGN KEY( " + fkList + " )" +
				 " REFERENCES " + referencedDO.getDBTableName() +
					"( " + pkList + " )\n";
		}
		 }
	}
	return ret;
    }

    public void writeFileFooter(  DataObject o ) throws Exception {
	tagSet.setTagValue( INDEXES, indexes );

	  tagSet.setTagValue( PRIMARY_KEY_CLAUSE, "" );
	Attribute[] pk = o.getPrimaryKeys();
	String pkList = "";
	String comma = "";
	  int i = 0;
	  for ( ; i < pk.length; i++ ) {
		 pkList += comma + pk[i].getName();
		 comma = ", ";
	  }
	if ( 0 != i )
		 tagSet.setTagValue( PRIMARY_KEY_CLAUSE,
			  "PRIMARY KEY( " +pkList + " )" );

	 tagSet.setTagValue( FOREIGN_KEY_REFERENCES_CLAUSE,
				 getForeignKeySQL( o ) );

		defaultWriteFileFooter( o );
	}

	private String getBeginCommentMark() throws Exception {
	  return tagSet.getTagValue( BEGIN_COMMENT );
	}

	private String getEndCommentMark() throws Exception {
	  return tagSet.getTagValue( END_COMMENT );
	}

}
