/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: DOTreeNode.java,v 1.1 2001/09/28 16:01:07 cryd0221 Exp $
 */



package org.enhydra.dods.gui.tree;

import org.enhydra.dods.store.doa.*;
import java.util.Vector;
/**
 * This class provides a wrapper for DataObjects and DOPackages in the treeview.
 *
 * @author  Philip Martin
 * @version $Revision: 1.1 $
 */
public class DOTreeNode {
    /**
     * The DataObject in this node.
     */
    protected DataObject DO;
    /**
     * All of the Objects which descend from this node in the tree.
     */
    protected Vector children;
    /**
     * true when this is a wrapper for a DOPackage.
     */
    protected boolean folder;
    /**
     * This constructor creates a wrapper for the given DataObject which is
     * not a folder.
     */
    public DOTreeNode(DataObject d) {
        this(d, false);
    }
    /**
     * This Constructor creates a wrapper for the given DataObject which is
     * either a Folder or not.
     */
    public DOTreeNode(DataObject d, boolean isFolder) {
        DO = d;
        folder = isFolder;
        children = new Vector();
        
        if(DO instanceof DOPackage) {
//System.err.println("DOTreeNode.ctor");
            DOPackage DOP = (DOPackage)DO;
//System.err.println("DOP.getDataObjects().size()="+DOP.getDataObjects().size());
            for(int i = 0; i < DOP.getDataObjects().size(); i++) {
                DataObject daob = (DataObject)DOP.getDataObjects().elementAt(i);
                if(daob instanceof DOPackage) {
                    children.addElement(new DOTreeNode(daob, true));
                }
                else {
                    children.addElement(new DOTreeNode(daob));
                }
            }
        }
    }
    /**
     * @return the DataObject contained in this wrapper.
     */
    public DataObject getDataObject() {
        return DO;
    }
    /**
     * @return the vector containing the node's children.
     */
    public Vector getChildren() {
        if(folder) {
            return children;
        }
        return null;
    }
    /**
     * The String format for these nodes is just the name of the dataobject
     * it is for.
     * @return the name of the DataObject in this node.
     */
    public String toString() {
        return DO.getName();
    }
    /**
     * @return !folder.
     */
    public boolean isLeaf() {
        return !folder;
    }

    /**
     * This method looks through all of the tree which descends from this
     * node and finds the node containing the given DataObject in its children.
     * @return The node containing the given DataObject as a child or null.
     */
    public DOTreeNode getParentWithChildContaining(DataObject data) {
        for(int i = 0; i < children.size(); i++) {
            DOTreeNode dotn = (DOTreeNode)children.elementAt(i);
            if(dotn.getDataObject().equals(data)) {
                return this;
            }
        }
        for(int i = 0; i < children.size(); i++) {
            DOTreeNode dotn = (DOTreeNode)children.elementAt(i);
            DOTreeNode returnvalue = dotn.getParentWithChildContaining(data);
            if(returnvalue != null) {
                return returnvalue;
            }
        }
        return null;
    }


  public DOTreeNode getNodeContaining(Object dataObject) {
    DOTreeNode node = null;
    if (getDataObject().equals(dataObject)) {
      node = this;
    } else {
      if (getChildren() != null) {
        int count = getChildren().size();
        for (int i = 0 ; i < count ; i++) {
          node = (DOTreeNode) getChildren().elementAt(i);
          if (node.getDataObject().equals(dataObject)) {
            break;
          } else {
            node = node.getNodeContaining(node);
            if (node != null) {
              break;
            }
          }
        }
      }
    }
    return node;
  }


    /**
     * This method updates the order of the children so that all of the leaves
     * are before all of the folders.  It then calls sortChildren to put them
     * in alphabetical order.
     */
    public void updateChildrenOrder() {
        Vector newChildren = new Vector();
        for(int i = 0; i < children.size(); i++) {
            DOTreeNode insert = (DOTreeNode)children.elementAt(i);
            if(insert.isLeaf()){
                newChildren.insertElementAt(insert, 0);
            }
            else {
                newChildren.addElement(insert);
                insert.updateChildrenOrder();
            }
        }
        children = newChildren;
        sortChildren();
    }
    /**
     * This method puts the children in alpha order without disturbing
     * The separation of leaves and Folders.
     */
    private void sortChildren() {
        if(children.size() < 2) {
            return;
        }
        Vector newChildren = new Vector();
        DOTreeNode insert = (DOTreeNode)children.elementAt(0);
        int i = 0;
        while(insert.isLeaf()) {
            if(newChildren.size()==0) {
                newChildren.addElement(insert);
                i++;
                if(i==children.size()) {
                    children = newChildren;
                    return;
                }
                insert = (DOTreeNode)children.elementAt(i);
                continue;
            }
            if(newChildren.size()==1) {
                DataObject setData = ((DOTreeNode)newChildren.elementAt(0)).getDataObject();
                if(insert.getDataObject().getName().compareTo(setData.getName()) < 0) {
                    newChildren.insertElementAt(insert, 0);
                }
                else {
                    newChildren.addElement(insert);
                }
                i++;
                if(i==children.size()) {
                    children = newChildren;
                    return;
                }
                insert = (DOTreeNode)children.elementAt(i);
                continue;
            }
            for(int j = 1; j < newChildren.size(); j++) {
                DataObject setData1 = ((DOTreeNode)newChildren.elementAt(j-1)).getDataObject();
                DataObject setData2 = ((DOTreeNode)newChildren.elementAt(j)).getDataObject();
                if(insert.getDataObject().getName().compareTo(setData1.getName())<0) {
                    newChildren.insertElementAt(insert, 0);
                    i++;
                    if(i==children.size()) {
                        children = newChildren;
                        return;
                    }
                    insert = (DOTreeNode)children.elementAt(i);
                    break;
                }
                else if(insert.getDataObject().getName().compareTo(setData2.getName())<0) {
                    newChildren.insertElementAt(insert, j);
                    i++;
                    if(i==children.size()) {
                        children = newChildren;
                        return;
                    }
                    insert = (DOTreeNode)children.elementAt(i);
                    break;
                }
                else if(j == newChildren.size()-1) {
                    newChildren.addElement(insert);
                    i++;
                    if(i==children.size()) {
                        children = newChildren;
                        return;
                    }
                    insert = (DOTreeNode)children.elementAt(i);
                }
            }
        }
        int leafCount = newChildren.size();
        while(!insert.isLeaf()) {
            if(newChildren.size()==leafCount) {
                newChildren.addElement(insert);
                i++;
                if(i==children.size()) {
                    children = newChildren;
                    return;
                }
                insert = (DOTreeNode)children.elementAt(i);
                continue;
            }
            if(newChildren.size()==leafCount+1) {
                DataObject setData = ((DOTreeNode)newChildren.elementAt(leafCount)).getDataObject();
                if(insert.getDataObject().getName().compareTo(setData.getName())<0) {
                    newChildren.insertElementAt(insert, leafCount);
                }
                else {
                    newChildren.addElement(insert);
                }
                i++;
                if(i==children.size()) {
                    children = newChildren;
                    return;
                }
                insert = (DOTreeNode)children.elementAt(i);
                continue;
            }
            for(int j = leafCount + 1; j < newChildren.size(); j++) {
                DataObject setData1 = ((DOTreeNode)newChildren.elementAt(j-1)).getDataObject();
                DataObject setData2 = ((DOTreeNode)newChildren.elementAt(j)).getDataObject();
                if(insert.getDataObject().getName().compareTo(setData1.getName())<0) {
                    newChildren.insertElementAt(insert, leafCount);
                    i++;
                    if(i==children.size()) {
                        children = newChildren;
                        return;
                    }
                    insert = (DOTreeNode)children.elementAt(i);
                    break;
                }
                else if(insert.getDataObject().getName().compareTo(setData2.getName())<0) {
                    newChildren.insertElementAt(insert, j);
                    i++;
                    if(i==children.size()) {
                        children = newChildren;
                        return;
                    }
                    insert = (DOTreeNode)children.elementAt(i);
                    break;
                }
                else if(j == newChildren.size()-1) {
                    newChildren.addElement(insert);
                    i++;
                    if(i==children.size()) {
                        children = newChildren;
                        return;
                    }
                    insert = (DOTreeNode)children.elementAt(i);
                }
            }
        }
    }
    /**
     * This method returns true if this node or any node that descends
     * from this node has the given node as a child.
     */
    public boolean contains(DOTreeNode node) {
        boolean returnValue = false;
        if(children.contains(node)) {
            return true;
        }
        int i = 0;
        while(returnValue == false && i < children.size()) {
            DOTreeNode node2 = (DOTreeNode)children.elementAt(i);
            returnValue = node2.contains(node);
            i++;
        }
        return returnValue;
    }
    /**
     * This method returns true if this node or any node that descends
     * from this node has the given DataObject in a node as a child.
     */
    public boolean contains(DataObject data) {
        boolean returnValue = false;
        if(DO.equals(data)) {
            returnValue = true;
        }
        int i = 0;
        while(returnValue == false && i < children.size()) {
            DOTreeNode node2 = (DOTreeNode)children.elementAt(i);
            returnValue = node2.contains(data);
            i++;
        }
        return returnValue;
    }
}
