/*
 * Enhydra Java Application Server Project
 *
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific terms governing rights and limitations
 * under the License.
 *
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 * $Id: Attribute.java,v 1.1 2001/09/28 16:01:07 cryd0221 Exp $
 */
package org.enhydra.dods.store.doa;

import org.enhydra.dods.store.doml.DOMLConstants;

/**
 * This class holds all of the information about an Attribute for a DataObject.
 *
 * @author  Philip Martin
 * @version $Revision: 1.1 $
 *
 */
public class Attribute {
    /**
     * String for the initialValue of the Attribute.
     */
    protected String initialValue;
    /**
     * True when there is an error in some part of the Attribute (ie invalid
     * initial value.
     */
    protected boolean error;
    /**
     * The type of the variable in java.
     */
    protected String javaType;
    /**
     * True if this attribute is used as an index in the database.
     */
    protected boolean index;
    /**
     * True if this attribute is a reference to another object instead of
     * its own atttribute.
     */
    protected boolean objectRef;
    /**
     * True if this is an array instead of a single instance.
     */
    protected boolean array;
    /**
     * The DBType that this javatype will map to.
     */
    protected String dbType;
    /**
     * True if this Attribute can be null in the database.
     */
    protected boolean CanBeNull;
    /**
     * True if this Attribute is a referential constraint.
     * The Attribute is a reference to another DO which must exist in the DB.
     */
    protected boolean refConstraint;
    /**
     * True if this Attribute is a primary key of this DO.
     * This means this DO is not using oid/verion.
     */
    protected boolean primaryKey;
    /**
     * Only meaningful if refConstraint==true.
     * Name of PK column in other DO being referenced as a foreign key.
     * Empty string or null indicates standard Enhydra oid foreign key.
     */
    protected String foreignKeyColumnName;
    /**
     * foreignKeyGroup
     *
     * Only meaningful if refConstraint==true and foreignKeyColumn is set.
     * Name of the "group" to which this FK belongs.
     * For example, say you have these tables:
     *   create table Person (
     *      fname CHAR(10),
     *      lname CHAR(10),
     *      PRIMARY KEY(fname,lname));
     *   create table Employee (
     *      emp_fn CHAR(10),
     *      emp_ln CHAR(10),
     *      spouse_fn CHAR(10),
     *      spouse_ln CHAR(10),
     *      FOREIGN KEY(emp_fn,emp_ln)       REFERENCES Person(fname,lname),
     *      FOREIGN KEY(spouse_fn,spouse_ln) REFERENCES Person(fname,lname));
     *
     * Because Employee has 2 FKs referring to the same table,
     * we need a way to group the fn and ln columns.
     *
     * The .doml file would contain:
     *    <table id="pkg.Person" notUsingOId="true">
     *       <column id="fname" isPrimaryKey="true">
     *          <type dbType="CHAR" javaType="String" size="10"/>
     *       </column>
     *       <column id="lname" isPrimaryKey="true">
     *          <type dbType="CHAR" javaType="String" size="10"/>
     *       </column>
     *    </table>
     *    <table id="Employee">
     *       <column id="emp_fn">
     *          <referenceObject reference="pkg.Person"
     *			foreignKeyColumnName="fname" foreignKeyGroup="emp"/>
     *          <type dbType="CHAR" javaType="String" size="10"/>
     *       </column>
     *       <column id="emp_ln">
     *          <referenceObject reference="pkg.Person"
     *			foreignKeyColumnName="lname" foreignKeyGroup="emp"/>
     *          <type dbType="CHAR" javaType="String" size="10"/>
     *       </column>
     *       <column id="spouse_fn">
     *          <referenceObject reference="pkg.Person"
     *			foreignKeyColumnName="fname" foreignKeyGroup="spouse"/>
     *          <type dbType="CHAR" javaType="String" size="10"/>
     *       </column>
     *       <column id="spouse_ln">
     *          <referenceObject reference="pkg.Person"
     *			foreignKeyColumnName="lname" foreignKeyGroup="spouse"/>
     *          <type dbType="CHAR" javaType="String" size="10"/>
     *       </column>
     *    </table>
     */
    protected String foreignKeyGroup = "ForeignKey";

    /**
     * If DO is for a stored procedure (instead of a table)
     * a field can either be an IN/OUT/INOUT or return parameter,
     * or it can be a column in the ResultSet returned by the stored proc.
     * The constants for these values all start with "SP_".
     */
    protected String spParamType  = null;
    protected String spParamIndex = "-1";

    /**
     * True if this Attribute has a constant value.
     */
    protected boolean constant;
    /**
     * True if this Attribute can be used in queries.
     */
    protected boolean usedForQuery;
    /**
     * the name to discribe this Attribute with.
     */
    protected String name;
    /**
     * The size of this attribute in the database.
     */
    protected String size;
    /**
     * The string which will be used for the javadoc tag in the DataObject.
     */
    protected String javadoc;
    /**
     * if this is an array, this is its size.
     */
    protected int arraySize;
    /**
     * The text related to the error if there is one.
     */
    protected String errorText;
    protected DataObject objectRefDO;
    /**
     * Creates an Attribute with the name "NewAttribute"
     */
    public Attribute() {
        this("NewAttribute");
    }
    /**
     * Creates an Attribute with the given name.
     */
    public Attribute(String nam) {
//System.out.println ("Instantiating Attribute:"+nam);
        name = nam;
        initialValue = "";
        error = false;
        javaType = "String";
        index = false;
        objectRef = false;
        array = false;
        dbType = "VARCHAR";
        CanBeNull = false;
        usedForQuery = false;
        size = "";
        javadoc = ("/**\n * \n */");
        arraySize = 1;
        errorText = "";
        objectRefDO = null;
    }
    public void setErrorText(String txt) {
        errorText = txt;
    }
    public String getErrorText() {
        return errorText;
    }
    public void setJavadoc(String jdoc) {
        javadoc = jdoc;
    }
    public String getJavadoc() {
        return javadoc;
    }
    public void setArraySize(int asize) {
        arraySize = asize;
    }
    public int getArraySize() {
        return arraySize;
    }
    public void setError(boolean error) {
        this.error = error;
    }
    public boolean hasError() {
        return this.error;
    }
    public void setObjectRef(boolean objectRef) {
        this.objectRef = objectRef;
    }
    public boolean isObjectRef() {
        return this.objectRef;
    }
    public void setArray(boolean array) {
        this.array = array;
    }
    public boolean isArray() {
        return this.array;
    }
    public void setCanBeNull(boolean CanBeNull) {
        this.CanBeNull = CanBeNull;
    }
    public boolean canBeNull() {
        return this.CanBeNull;
    }
    public void setUsedForQuery(boolean usedForQuery) {
        this.usedForQuery = usedForQuery;
    }
    public boolean isUsedForQuery() {
        return this.usedForQuery;
    }
    public void setRefConstraint(boolean refConstraint) {
        this.refConstraint = refConstraint;
    }
    public boolean isRefConstraint() {
        return this.refConstraint;
    }
    public void setForeignKeyColumnName(String foreignKeyColumnName) {
        this.foreignKeyColumnName = foreignKeyColumnName;
    }
    public boolean isForeignKey() {
        return ( null != this.foreignKeyColumnName
		 && 0 != this.foreignKeyColumnName.length() );
    }
    public String getForeignKeyColumnName() {
        return this.foreignKeyColumnName;
    }
    public void setForeignKeyGroup(String foreignKeyGroup) {
	if ( null != foreignKeyGroup && 0 != foreignKeyGroup.length() )
	    this.foreignKeyGroup = foreignKeyGroup;
    }
    public String getForeignKeyGroup() {
        return this.foreignKeyGroup;
    }
    public void setPrimaryKey(boolean primaryKey) {
        this.primaryKey = primaryKey;
    }
    public boolean isPrimaryKey() {
        return this.primaryKey;
    }

    public void setSpParamType(String s) {
        this.spParamType = s;
    }
    public String getSpParamType() {
        return this.spParamType;
    }

    public void setSpParamIndex(String s) {
        this.spParamIndex = s;
    }
    public String getSpParamIndex() {
	if ( null == this.spParamIndex && isSpReturnValue() )
	    setSpParamIndex( "1" );
        return this.spParamIndex;
    }

    private boolean isStringEqual( String s1, String s2 ) {
    	if ( null == s1 ) s1 = "";
    	if ( null == s2 ) s2 = "";
	return s1.equals( s2 );
    }
    public boolean isResultSetValue() {
	/*
        return	isStringEqual( spParamType, DOMLConstants.SP_RESULTSET )	||
        	isStringEqual( spParamType, "" );
		*/
	return ! ( isSpReturnValue() || isSpInParam() || isSpOutParam() );
    }
    public boolean isSpReturnValue() {
        return	isStringEqual( spParamType, DOMLConstants.SP_RETURN_VALUE );
    }
    public boolean isSpInParam() {
        return	isStringEqual( spParamType, DOMLConstants.SP_IN_PARAM )	||
        	isStringEqual( spParamType, DOMLConstants.SP_INOUT_PARAM );
    }
    public boolean isSpOutParam() {
        return	isStringEqual( spParamType, DOMLConstants.SP_OUT_PARAM )   ||
        	isStringEqual( spParamType, DOMLConstants.SP_INOUT_PARAM ) ||
        	isStringEqual( spParamType, DOMLConstants.SP_RETURN_VALUE );
    }

    public void setConstant(boolean constant) {
        this.constant = constant;
    }
    public boolean isConstant() {
        return this.constant;
    }
    public void setIndex(boolean index) {
        this.index = index;
    }
    public boolean isIndex() {
        return this.index;
    }
    public void setName(String name) {
        this.name = name;
    }
    public String getName() {
        return this.name;
    }
    public void setJavaType(String javaType) {
        this.javaType = javaType;
    }
    public String getJavaType() {
        return this.javaType;
    }
    public void setDBType(java.lang.String dbType) {
        this.dbType = dbType;
    }
    public String getDBType() {
        return this.dbType;
    }
    public void setSize(String size) {
        this.size = size;
    }
    public String getSize() {
        return this.size;
    }
    public void setInitialValue(String initialValue) {
        this.initialValue = initialValue;
    }
    public String getInitialValue() {
        return this.initialValue;
    }
    public DataObject getObjectRefDO() {
        return objectRefDO;
    }
    public void setObjectRefDO(DataObject orDO) {
//System.out.println("Set object ref do for "+name+" = "+orDO);
        this.objectRefDO = orDO;
    }
    /**
     * @return the Attribute as a string of the form "<I>name</I> Attribute"
     */
    public String toString() {
        return (name + " Attribute");
    }

    /**
     * @return the Attribute as a string of the form "<I>name</I> Attribute"
     */
    public String toLongString() {
		StringBuffer sb = new StringBuffer(200);
		String sep = System.getProperty("line.separator");
		String inset = "    ";
		sb.append("Attribute: [name="+name+"]");
		sb.append(sep+inset+"initialValue="+initialValue);
		sb.append(sep+inset+"error="+error);
		sb.append(sep+inset+"javaType="+javaType);
		sb.append(sep+inset+"index="+index);
		sb.append(sep+inset+"objectRef="+objectRef);
		sb.append(sep+inset+"array="+array);
		sb.append(sep+inset+"dbType="+dbType);
		sb.append(sep+inset+"CanBeNull="+CanBeNull);
		sb.append(sep+inset+"refConstraint="+refConstraint);
		sb.append(sep+inset+"primaryKey="+primaryKey);
		sb.append(sep+inset+"foreignKeyColumnName="+foreignKeyColumnName);
		sb.append(sep+inset+"foreignKeyGroup="+foreignKeyGroup);
		sb.append(sep+inset+"spParamType="+spParamType);
		sb.append(sep+inset+"spParamIndex="+spParamIndex);
		sb.append(sep+inset+"constant="+constant);
		sb.append(sep+inset+"usedForQuery="+usedForQuery);
		sb.append(sep+inset+"size="+size);
		sb.append(sep+inset+"javadoc="+javadoc);
		sb.append(sep+inset+"arraySize="+arraySize);
		sb.append(sep+inset+"errorText="+errorText);
		sb.append(sep+inset+"objectRefDO="+objectRefDO);
		sb.append(sep+"/Attribute");
        return sb.toString();
    }

    /**
     * @return true if the foreignKeyColumnName is equals
     */
    private boolean isForeignKeyColumnNameEqual(Attribute att) {
      boolean ret = false;
      if (foreignKeyColumnName == null) {
          if (att.getForeignKeyColumnName() == null)
              ret = true;
      } else {
          ret = foreignKeyColumnName.equals(att.getForeignKeyColumnName());
      }
      return ret;
    }

    /**
     * @return true if all of the things describing this Attribute are equal.
     */
    public boolean equals(Attribute att) {
        if(!name.equals(att.getName()))
            return false;
        if(!javaType.equals(att.getJavaType()))
            return false;
        if(CanBeNull != att.canBeNull())
            return false;
        if(array != att.isArray())
            return false;
        if(arraySize != att.getArraySize())
            return false;
        if(!dbType.equals(att.getDBType()))
            return false;
        if(error != att.hasError())
            return false;
        if(primaryKey != att.isPrimaryKey())
            return false;
	if(!isForeignKeyColumnNameEqual(att))
              return false;
        if(refConstraint != att.isRefConstraint())
            return false;
        if(constant != att.isConstant())
            return false;
        if(usedForQuery != att.isUsedForQuery())
            return false;
        if(!errorText.equals(att.getErrorText()))
            return false;
        if(index != att.isIndex())
            return false;
        if(!initialValue.equals(att.getInitialValue()))
            return false;
        if(!javadoc.equals(att.getJavadoc()))
            return false;
        if(objectRef != att.isObjectRef())
            return false;
        if(!size.equals(att.getSize()))
            return false;
        return true;
    }
}
