/*
 * Enhydra Java Application Server Project
 *
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific terms governing rights and limitations
 * under the License.
 *
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 * $Id: DODSTypes.java,v 1.1 2001/09/28 16:01:07 cryd0221 Exp $
 */
package org.enhydra.dods.store.doa;

import org.enhydra.dods.exceptions.*;
import com.lutris.util.*;
import java.io.*;
import java.util.*;

public class DODSTypes {

  // The one, the only instance...
  private static DODSTypes dodsTypes = null;


  // The config details for the types
  public Config typecfg = null;

  // The current database in use
  private String dbms = null;


/**
 * DODS_HOME is the name of a property that is set with -D
 * when DODS is run.  The value for the property should be
 * the directory where DODS can find its config files.
 */
  private static final String DODS_HOME = "DODS_HOME";


/**
 * The following are the constants for the substitutions
 * that can occur in the types config details.
 */
  public static final String JAVA_VAR = "<JAVA_VAR>";
  public static final String JAVA_TYPE = "<JAVA_TYPE>";
  public static final String JDBC_VAR = "<JDBC_VAR>";
  public static final String JDBC_TYPE = "<JDBC_TYPE>";
  public static final String JAVA_VALUE = "<JAVA_VALUE>";
  public static final String DB_VALUE = "<DB_VALUE>";


/**
 * Constructor.
 */
  protected DODSTypes() {

  }


/**
 * Gets the instance of this class.  As this is a static
 * method, use 'DODSTypes.getInstance()' anywhere to get
 * the initialised instance.
 *
 * @return see comment.
 */
  public static DODSTypes getInstance() {

    // is there already an instance available?
    if (dodsTypes != null) {

      // yes, use it...
      return dodsTypes;

    }

    // create a new instance
    dodsTypes = new DODSTypes();

    // initialise it
    try {

      dodsTypes.init();

    } catch (Exception e) {

      System.err.println("Unable to initialise DODS Types: "+e.getMessage());
      System.exit(1);

    }

    // use this instance
    return dodsTypes;

  }


/**
 * Initialises this class.  Reads the 'types.conf' file in
 * and holds the information.
 */
  private void init() throws Exception {

    // Get the DODS home directory
    String dodsDir = System.getProperty(DODS_HOME);

    // was the property found?
    if (dodsDir == null) {

      // no, indicate this
      throw new Exception("Cannot find property:  " + DODS_HOME);

    }

    // build the config filename for the 'types.conf'
    String confFilename = dodsDir + File.separator + "types.conf";
    File confFile = new File(confFilename);

    // does the file exist?
    if (!confFile.exists()) {

      // no, indicate this
      throw new FileNotFoundException("No conf file " + confFilename);

    }

    // make sure it's not a directory
    if (!confFile.isFile()) {

      // no, indicate this
      throw new Exception("Not a file: " + confFilename);

    }

    // DBG
    // System.err.println("About to read config file...");

    // read the config data from the file
    FileInputStream cfgFIS = new FileInputStream(confFilename);
    ConfigFile cfgF = new ConfigFile(cfgFIS);
    cfgFIS.close();

    // DBG
    // System.err.println("Config file read...");

    // get the config from the file
    typecfg = cfgF.getConfig();

  }


/**
 * Sets the current DBMS in use in DODS.
 *
 * @param dbms  The DBMS to use
 */
  public void setDBMS(String dbms) {

    this.dbms = dbms;

  }


/**
 * Gets the current DBMS in use in DODS.
 *
 * @return see comment.
 */
  public String getDBMS() {

    return dbms;

  }


/**
 * Gets the DODS Types for the current DBMS
 *
 * @return see comment.  Null is returned if errors occur.
 */
  public String[] getDODSTypes() {

    // DBG
    // System.out.println("DBMS is: <"+dbms+">");

    // null pointer checks...
    if (dbms == null)
      return null;

    String[] dodsTypes = null;

    try {

      dodsTypes = typecfg.getStrings(dbms+"."+"TYPE_NAMES");

    } catch (ConfigException ce) {

      System.err.println("Can't get DODS types: "+ce.getMessage());

    }

    return dodsTypes;

  }


/**
 * Gets the SQL_TYPE for the DODS type passed.
 *
 * @param dodsType  The type to find the info for
 * @return see comment.  Null is returned if errors occur.
 */
  public String getSQLType(String dodsType) {

    // null pointer checks...
    if (dbms == null || dodsType == null)
      return null;

    String sqlType = null;

    try {

      sqlType = typecfg.getString(dbms+"."+dodsType+"."+"SQL_TYPE");

    } catch (ConfigException ce) {

    }

    return sqlType;

  }


/**
 * Gets the SQL_TYPE for the DODS type passed.
 *
 * @param dodsType  The type to find the info for
 * @return see comment.  Null is returned if errors occur.
 */
  public String getDBType(String dodsType) {

    // null pointer checks...
    if (dbms == null || dodsType == null)
      return null;

    String sqlType = null;

    try {

      sqlType = typecfg.getString(dbms+"."+dodsType+"."+"SQL_TYPE");

    } catch (ConfigException ce) {

    }

    return sqlType;

  }


/**
 * Gets the JAVA_TYPE for the DODS type passed.
 *
 * @param dodsType  The type to find the info for
 * @return see comment.  Null is returned if errors occur.
 */
  public String getJavaType(String dodsType) {

    // null pointer checks...
    if (dbms == null || dodsType == null)
      return null;

    String javaType = dodsType ;//PM! null;

    try {

      javaType = typecfg.getString(dbms+"."+dodsType+"."+"JAVA_TYPE");

    } catch (ConfigException ce) {

    }

    return javaType;

  }


/**
 * Gets the JDBC_TYPE for the DODS type passed.
 *
 * @param dodsType  The type to find the info for
 * @return see comment.  Null is returned if errors occur.
 */
  public String getJDBCType(String dodsType) {

    // null pointer checks...
    if (dbms == null || dodsType == null)
      return null;

    String jdbcType = null;

    try {

      jdbcType = typecfg.getString(dbms+"."+dodsType+"."+"JDBC_TYPE");

    } catch (ConfigException ce) {

    }

    return jdbcType;

  }


/**
 * Gets the JAVA_TO_JDBC details for the DODS type passed.
 *
 * @param dodsType  The type to find the info for
 * @return see comment.  Null is returned if errors occur.
 */
  public String getJavaToJDBC(String dodsType) {

    // null pointer checks...
    if (dbms == null || dodsType == null)
      return null;

    String java2jdbc= null;

    try {

      java2jdbc= typecfg.getString(dbms+"."+dodsType+"."+"JAVA_TO_JDBC");

    } catch (ConfigException ce) {

    }

    return java2jdbc;

  }


/**
 * Gets the JDBC_TO_JAVA details for the DODS type passed.
 *
 * @param dodsType  The type to find the info for
 * @return see comment.  Null is returned if errors occur.
 */
  public String getJDBCToJava(String dodsType) {

    // null pointer checks...
    if (dbms == null || dodsType == null)
      return null;

    String jdbc2java= null;

    try {

      jdbc2java= typecfg.getString(dbms+"."+dodsType+"."+"JDBC_TO_JAVA");

    } catch (ConfigException ce) {

    }

    return jdbc2java;

  }


/**
 * Gets the JAVA_TYPE_CLONE for the DODS type passed.
 *
 * @param dodsType  The type to find the info for
 * @return see comment.  Null is returned if errors occur.
 */
  public String getJavaTypeClone(String dodsType) {

    // null pointer checks...
    if (dbms == null || dodsType == null)
      return null;

    String javaTypeClone = null;

    try {

      javaTypeClone = typecfg.getString(dbms+"."+dodsType+"."+"JAVA_TYPE_CLONE");

    } catch (ConfigException ce) {

    }

    return javaTypeClone;

  }


/**
 * Is a SIZE_ALLOWED for the DODS type passed.
 *
 * @param dodsType  The type to find the info for
 * @return see comment.  False is returned if errors occur.
 */
  public boolean isSizeAllowed(String dodsType) {

    // null pointer checks...
    if (dbms == null || dodsType == null)
      return false;

    boolean sizeAllowed = false;

    // DBG
    // System.err.println("DBMS is: "+dbms);

    try {

      sizeAllowed = typecfg.getBoolean(dbms+"."+dodsType+"."+"SIZE_ALLOWED");

      // DBG
      // System.err.println("SIZE_ALLOWED for dodsType: "+dodsType+", is: "+sizeAllowed);

    } catch (ConfigException ce) {

      //System.err.println("ConfigException in isSizeAllowed: "+ce.getMessage());

    }

    // DBG
    // System.err.println("SIZE_ALLOWED result for dodsType: "+dodsType+", is: "+sizeAllowed);

    return sizeAllowed;

  }


/**
 * Gets the JAVA_INITIAL for the DODS type passed.
 *
 * @param dodsType  The type to find the info for
 * @return see comment.  Null is returned if errors occur.
 */
  public String getJavaInitial(String dodsType) {

    // null pointer checks...
    if (dbms == null || dodsType == null)
      return null;

    String javaInitial = null;

    try {

      javaInitial = typecfg.getString(dbms+"."+dodsType+"."+"JAVA_INITIAL");

    } catch (ConfigException ce) {

    }

    return javaInitial;

  }


/**
 * Gets the DB_INITIAL for the DODS type passed.
 *
 * @param dodsType  The type to find the info for
 * @return see comment.  Null is returned if errors occur.
 */
  public String getDBInitial(String dodsType) {

    // null pointer checks...
    if (dbms == null || dodsType == null)
      return null;

    String dbInitial = null;

    try {

      dbInitial = typecfg.getString(dbms+"."+dodsType+"."+"DB_INITIAL");

    } catch (ConfigException ce) {

    }

    return dbInitial;

  }


/**
 * Gets the JAVA_EXACT_COMPARE for the DODS type passed.
 *
 * @param dodsType  The type to find the info for
 * @return see comment.  Null is returned if errors occur.
 */
  public String getJavaExactCompare(String dodsType) {

    // null pointer checks...
    if (dbms == null || dodsType == null)
      return null;

    String javaExactCompare = null;

    try {

      javaExactCompare = typecfg.getString(dbms+"."+dodsType+"."+"JAVA_EXACT_COMPARE");

    } catch (ConfigException ce) {

    }

    return javaExactCompare;

  }


/**
 * Gets the JAVA_INEXACT_COMPARE for the DODS type passed.
 *
 * @param dodsType  The type to find the info for
 * @return see comment.  Null is returned if errors occur.
 */
  public String getJavaInexactCompare(String dodsType) {

    // null pointer checks...
    if (dbms == null || dodsType == null)
      return null;

    String javaInexactCompare = null;

    try {

      javaInexactCompare = typecfg.getString(dbms+"."+dodsType+"."+"JAVA_INEXACT_COMPARE");

    } catch (ConfigException ce) {

    }

    return javaInexactCompare;

  }


/**
 * Determines if the type passed is a Java primitive
 *
 * @param javaType  The type to check
 * @return True, if java type is a primitive
 */
  public boolean isJavaPrimitive(String javaType) {

    // null pointer checks...
    if (javaType == null)
      return false;

    return (javaType.equals("int") ||
            javaType.equals("boolean") ||
            javaType.equals("float") ||
            javaType.equals("double") ||
            javaType.equals("char") ||
            javaType.equals("byte") ||
            javaType.equals("short") ||
            javaType.equals("long"));

  }


}
