/*
 * Enhydra Java Application Server Project
 *
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific terms governing rights and limitations
 * under the License.
 *
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 * $Id: DataObject.java,v 1.1 2001/09/28 16:01:07 cryd0221 Exp $
 */
package org.enhydra.dods.store.doa;

import org.enhydra.dods.exceptions.*;
import java.util.*;
/**
 * A holder for all the information describing a Data Object.
 *
 * @author  Philip Martin
 * @version $Revision: 1.1 $
 */
public class DataObject {
    /**
     * Type of mapping this Object will use to go into the Database
     */
    protected String dbTableType;
    /**
     * All of the Attributes that this Data Object has.
     */
    protected Vector attributes;
    /**
     * Name of the Package this class will be in.
     */
    protected String packageName;
    /**
     * All of the DataObjects that this DataObject inherits from.
     */
    protected Vector ancestors;
    /**
     * This is true if the DataObject is abstract.
     */
    protected boolean Abstract;
    /**
     * This is true if the DataObject is for a stored procedure.
     */
    protected boolean isStoredProc = false;
    /**
     * This is true if the DataObject is not using oid/version.
     */
    protected boolean notUsingOId;
    /**
     * All of the DataObjects which extend this one.
     */
    protected Vector extendedBy;
    /**
     * Name of the dbTable that this Object will go into.
     */
    protected String dbTableName;
    /**
     * True when this DataObject is an index for the dataBase.
     */
    protected boolean index;
    /**
     * This is the type of DataBase that the project is being mapped to.
     */
    protected String database = "Standard"; // default db
    /**
     * The name of this DataObject.  This also defines the future class name
     * which will be nameDO.class
     */
    protected String name;
    /**
     * All of the DataObjects which have this DataObject as an ancestor.
     */
    protected Vector descendents;
    /**
     * True if this Class will be final.
     */
    protected boolean Final;
    /**
     * True if this Class will use "lazy loading" where the DO values are
     * retrieved from the DB only when on the first get or set call.
     */
    protected boolean lazyLoading;
    /**
     * True if this Class will fully cache all objects from the database.
     */
    protected boolean fullCaching;
    /**
     * True if this Class will implement a cache.
     */
    protected boolean caching;
    /**
     * The DataObject that this DataObject extends.
     */
    protected DataObject extensionOf;
    /**
     * Valid dbTableType where all concrete classes get a table.
     */
    public static final String CONCRETE = "Concrete";
    /**
     * Valid dbTableType where all classes get a table.
     */
    public static final String ALL_CLASSES = "All Classes";
    /**
     * Valid dbTableType where an entire hierarchy is mapped to one table.
     */
    public static final String ENTIRE_HIERARCHY = "Entire Hierarchy";
    /**
     * Constructor which initializes everything to default values except name.
     * @param s The DataObject's name.
     */
    public DataObject(String s) {
		this();
		name = s;
    }
    /**
     * Constructor which initializes all of the variables to default values.
     */
    public DataObject() {
		Abstract = false;
		index = false;
		name = "NewObject";
		dbTableName = "NewDBTable";
		packageName = "";
		Final = false;
		lazyLoading = false;
		fullCaching = false;
		caching = false;
		dbTableType = CONCRETE;
		extensionOf = null;
		extendedBy = new Vector();
		descendents = new Vector();
		ancestors = new Vector();
		attributes = new Vector();
    }
    /**
     * Sets the name for this DataObject if it is a valid className.
     * @exception DataObjectNameException this is thrown if there is a name
     *  conflict or if the string has spaces ect.
     */
    public void setName(String s) throws DataObjectNameException {
		if(s.equals("")) {
		    throw new DataObjectNameException("Class name not Specified");
		}
		s.trim();
		if(!checkIdentifier(s)) {
		    throw new DataObjectNameException("Invalid class name: Use valid Java Identifier");
		}
		name = s;
    }
    /**
     * @return The name of the DataObject
     */
    public String getName() {
		return name;
    }
    /**
     * Sets up the name of the Table that this DataObject will be in.
     * @exception DataObjectNameException Thrown if the name is invalid.
     */
    public void setDBTableName(String dbtablename) throws DataObjectNameException {
		if(dbtablename.equals("")) {
		    throw new DataObjectNameException("DB Table Name not specified.");
		}
		dbTableName = dbtablename;
    }
    /**
     * @return The dbTableName for this DataObject.
     */
    public String getDBTableName() {
		return dbTableName;
    }
    /**
     * Set's the name of the Package for this DataObject.
     * @exception DataObjectNameException Thrown if the package name is invalid.
     */
    public void setPackageName(String packagename) throws DataObjectNameException {
		packagename.trim();
		if(!checkPackageIdentifier(packagename)) {
		    throw new DataObjectNameException("Invalid PackageName: Use valid Identifier");
		}
		packageName = packagename;
    }
    /**
     * @return The packageName.
     */
    public String getPackageName() {
		return packageName;
    }
    /**
     * @return The packageName with leading "root." stripped off.
     */
    public String getAdjustedPackageName() {
		if ( packageName.startsWith( "root." ) )
		    return packageName.substring( 5 );
		return packageName;
    }
    /**
     * Sets Final for this DataObject
     */
    public void setFinal(boolean Fina) {
		Final = Fina;
    }
    /**
     * @return The type of this DataObject.
     */
    public boolean isFinal() {
		return Final;
    }
    /**
     * Sets lazyLoading for this DataObject
     */
    public void setLazyLoading(boolean lazy) {
		lazyLoading = lazy;
    }
    /**
     * @return The type of this DataObject.
     */
    public boolean isLazyLoading() {
		return lazyLoading;
    }
    /**
     * Sets lazyLoading for this DataObject
     */
    public void setFullCaching(boolean full) {
		fullCaching = full;
    }
    /**
     * @return The type of this DataObject.
     */
    public boolean isFullCaching() {
		return fullCaching;
    }
    /**
     * Sets caching for this DataObject
     */
    public void setCaching(boolean impl_cache) {
		caching = impl_cache;
    }
    /**
     * @return The type of this DataObject.
     */
    public boolean isCaching() {
		return caching;
    }
    /**
     * Sets the DataObject that this extends to a new DataObject
     * @param extensionof The new DataObject.
     */
    public void setExtensionOf(DataObject extensionof) {
		if(extensionOf != null) {
		    extensionOf.getExtendedBy().removeElement(this);
		}
		extensionOf = extensionof;
		if(extensionOf != null) {
		    extensionOf.getExtendedBy().addElement(this);
		}
		updateAncestors();
    }
    /**
     * @return The DataObject that this extends.
     */
    public DataObject getExtensionOf(){
		return extensionOf;
    }
    /**
     * @return The Vector of DataObjects Extended by this DataObject.
     */
    public Vector getExtendedBy() {
		return extendedBy;
    }
    /**
     * @return The Vector of DataObjects that are Descended from this one.
     */
    public Vector getDescendents() {
		return descendents;
    }
    /**
     * @return The Vector containing all of the attributes of this DataObject.
     */
    public Vector getAttributes() {
		return attributes;
    }
    /**
     * Sets the attributes to the new vector of attributes.
     */
    public void setAttributes(Vector atts) {
		attributes.removeAllElements();
		for(int i = 0; i < atts.size(); i++) {
		    attributes.addElement(atts.elementAt(i));
		}
    }
    /**
     * Adds a new Attribute.
     */
    public void addAttribute(Attribute att) {
//System.err.println("DataObject.addAttribute");
		attributes.addElement(att);
    }
    /**
     * Sets if this is an index or not.
     */
    public void setIndex(boolean Index) {
		index = Index;
    }
    /**
     * @return True if this is an index DataObject
     */
    public boolean isIndex() {
		return index;
    }
    /**
     * Sets if this is an abstract Class.
     */
    public void setAbstract(boolean abstract1) {
		Abstract = abstract1;
    }
    /**
     * @return True if this is DO is not using oid/version.
     */
     public boolean notUsingOId() {
		  return notUsingOId;
     }
     /**
     * Sets if this is Class is not using oid/version.
     */
     public void setNotUsingOId(boolean b) {
		  notUsingOId = b;
     }

    /**
     * @return True if this is DO is for a stored procedure.
     */
     public boolean isStoredProc() {
		  return isStoredProc;
     }
     /**
     * Sets if this is Class is not using oid/version.
     */
     public void setIsStoredProc(boolean b) {
		  isStoredProc = b;
     }

    /**
     * @return True if this is a concrete DataObject.
     */
    public boolean isConcrete() {
		return ( isLeaf() && ! isAbstract() );
    }
    /**
     * @return True if this is a leaf DataObject.
     */
    public boolean isLeaf() {
		Vector descendants = getDescendents();
		if ( 0 == descendants.size() )
		    return true;
		return false;
    }
    /**
     * @return True if this is an Abstract DataObject.
     */
    public boolean isAbstract() {
		return Abstract;
    }

    /**
     * @return The type of Database this Project is for.
     */
    public String getDatabase() {
		return database;
    }
    /**
     * Sets the database this is for.
     * FIX: Should throw an exception if an invalid DataBase is chosen.
     */
    public void setDatabase(String db) {
       if (db.trim().toLowerCase().startsWith("standard")) {
		 database = "Standard";
       } else {
		 database = db;
       }
    }

    /**
     * Sets the DBTableType to a new valid value.
     * @exception DataObjectTypeException Thrown if the DBTableType is invalid.
     */
    public void setDBTableType(String dbtabletype) throws DataObjectTypeException {
		if(dbtabletype.equals(CONCRETE) || dbtabletype.equals(ALL_CLASSES) || dbtabletype.equals(ENTIRE_HIERARCHY)) {
		    dbTableType = dbtabletype;
		}
		else {
		    throw new DataObjectTypeException("Invalid Database Mapping type");
		}
    }
    /**
     * @return THe DBTableType for this DataObject.
     */
    public String getDBTableType() {
		return dbTableType;
    }
    /**
     * @return The vector containing all of the Ancestors of this DataObject.
     */
    public Vector getAncestors() {
		return ancestors;
    }
    /**
     * Determines if two DataObjects are equal by comparing each value in them.
     */
    public boolean equals(DataObject doa) {
		if(!doa.getName().equals(this.getName()))
		    return false;
		if(!doa.getPackageName().equals(this.getPackageName()))
		    return false;
		if(index != doa.isIndex())
		    return false;
		if(extensionOf != null) {
		    if(doa.getExtensionOf() == null) {
				return false;
		    }
		}
		else {
		    if(doa.getExtensionOf() != null) {
				return false;
		    }
		}
		if(!dbTableType.equals(doa.getDBTableType()))
		    return false;
		if(!dbTableName.equals(doa.getDBTableName()))
		    return false;
		if(Final != doa.isFinal())
		    return false;
		if(Abstract != doa.isAbstract())
		    return false;
		if(doa.getAttributes().size() != attributes.size())
		    return false;
		if(doa.getAncestors().size() != ancestors.size())
		    return false;
		if(doa.getDescendents().size() != descendents.size())
		    return false;
		if(doa.getExtendedBy().size() != extendedBy.size())
		    return false;
		for(int i = 0; i < attributes.size(); i++) {
		    if(!((Attribute)attributes.elementAt(i)).equals((Attribute)doa.getAttributes().elementAt(i)))
				return false;
		}
		return true;
    }
    /**
     * Returns a String representation of the form:
     * "<I>packageName</I>.<I>name</I>DO"
     */
    public String toString() {
		return new String(packageName + "." + name + "DO");
    }
    /**
     * Returns true if the string is a valid identifier.
     */
    public boolean checkIdentifier(String s) {
		int length = s.length();
		char[] chars = new char[length];
		s.getChars(0, length-1, chars, 0);
		if(!Character.isJavaIdentifierStart(chars[0]))
		    return false;
		for(int i = 1; i < length; i++){
		    if(!Character.isJavaIdentifierPart(chars[i]))
				return false;
		}
		return true;
    }
    /**
     * Similar to checkIdentifier, but this will not give an error when "." are
     * found in the string.
     */
    public boolean checkPackageIdentifier(String s) {
		int length = s.length();
		char[] chars = new char[length];
		s.getChars(0, length-1, chars, 0);
		if(!Character.isJavaIdentifierStart(chars[0]))
		    return false;
		for(int i = 1; i < length; i++){
		    if(!Character.isJavaIdentifierPart(chars[i])) {
				if(chars[i] != '.')
				    return false;
		    }
		}
		return true;
    }
    /**
     * This method reworks the hierarchy of DataObjects when a relationship
     * between them is changed.
     */
    public void updateAncestors() {
		for(int i = 0; i < ancestors.size(); i++) {
		    ((DataObject)ancestors.elementAt(i)).getDescendents().removeElement(this);
		    for(int j = 0; j < descendents.size(); j++) {
				((DataObject)ancestors.elementAt(i)).getDescendents().removeElement(
						((DataObject)descendents.elementAt(j)));
		    }
		}
		ancestors = new Vector();
		if(extensionOf != null) {
		    for(int i = 0; i < extensionOf.getAncestors().size(); i++) {
				ancestors.addElement(extensionOf.getAncestors().elementAt(i));
		    }
		    ancestors.addElement(extensionOf);
		}
		for(int i = 0; i < ancestors.size(); i++) {
		    ((DataObject)ancestors.elementAt(i)).getDescendents().addElement(this);
		    for(int j = 0; j < descendents.size(); j++) {
				((DataObject)ancestors.elementAt(i)).getDescendents().addElement(
						((DataObject)descendents.elementAt(j)));
		    }
		}
		for(int i = 0; i < extendedBy.size(); i++) {
		    ((DataObject)extendedBy.elementAt(i)).updateAncestors();
		}
    }
    /**
     * This method checks to see if any of the Attributes in this DataObject
     * have errors.
     */
    public boolean hasAttributeErrors() {
		for(int i = 0; i < attributes.size(); i++) {
		    if(((Attribute)attributes.elementAt(i)).hasError()) {
				return true;
		    }
		}
		return false;
    }

    /**
     * This method will take in a vector of DataObjects and Strings and
     * stick all of the strings at the front and alphabetize the DataObjects.
     */
    public static Vector alphabetizeVector(Vector DOS) {
		Vector newDOS = new Vector();
		int strings = 0;
		for(int i = 0; i < DOS.size(); i++) {
		    if(DOS.elementAt(i) instanceof String) {
				newDOS.insertElementAt(DOS.elementAt(i), 0);
				strings++;
		    }
		    else{
				DataObject DO = (DataObject)DOS.elementAt(i);
				boolean added = false;
				for(int j = strings; j < newDOS.size(); j++) {
				    DataObject newDO = (DataObject)newDOS.elementAt(j);
				    if(DO.getName().compareTo(newDO.getName()) < 0) {
						newDOS.insertElementAt(DO, j);
						added = true;
						break;
				    }
				}
				if(!added) {
				    newDOS.addElement(DO);
				}
		    }
		}
		return newDOS;
    }

    /**
     * Return the Primary Key attributes for this data object.
     *
     * @return an array (possibly of size 0) of the PKs in this DO.
     * @author Jay Gunter
     */
    public Attribute[] getPrimaryKeys() {
		Vector v = new Vector();
		Vector a = getAttributes();
		for ( int i = 0; i < a.size(); i++ ) {
		    Attribute a1 = (Attribute) a.elementAt(i);
		    if ( a1.isPrimaryKey() )
				v.addElement( a1 );
		}
		Attribute[] ret = new Attribute[ v.size() ];
		v.copyInto( ret );
		return ret;
    }

    /**
     * Return the Foreign Key attributes for this data object.
     *
     * @return an array (possibly of size 0) of the FKs in this DO.
     * @author Jay Gunter
     */
    public Attribute[] getForeignKeys() {
		Vector v = new Vector();
		Vector a = getAttributes();
		for ( int i = 0; i < a.size(); i++ ) {
		    Attribute a1 = (Attribute) a.elementAt(i);
//System.err.println("a1.getName()="+a1.getName());
//System.err.println("a1.isRefConstraint()="+a1.isRefConstraint());
		    if ( a1.isRefConstraint() ) {
				String nameOfPKinReferencedTable = a1.getForeignKeyColumnName();
				if ( 0 != nameOfPKinReferencedTable.length() ) {
				    v.addElement( a1 );
				}
		    }
		}
		Attribute[] ret = new Attribute[ v.size() ];
		v.copyInto( ret );
		return ret;
    }

    public Hashtable getForeignKeysGroupedByReferencedDO() {
		/*
		    Given "referenced DOs" like:

				CREATE TABLE person (
				  p_fn      CHAR(10),
				  p_ln      CHAR(10),
				  PRIMARY KEY( p_fn, p_ln )
				);
				CREATE TABLE manager (
				  m_id      CHAR(10),
				  m_title   CHAR(10),
				  PRIMARY KEY( m_id )
				);

		    and this "referencing" DO:

				CREATE TABLE employee (
				  emp_fn    CHAR(10),
				  emp_ln    CHAR(10),
				  spouse_fn CHAR(10),
				  spouse_ln CHAR(10),
				  emp_mgr   CHAR(10),
				  FOREIGN KEY(emp_fn,emp_ln)       REFERENCES person(p_fn,p_ln),
				  FOREIGN KEY(spouse_fn,spouse_ln) REFERENCES person(p_fn,p_ln),
				  FOREIGN KEY(emp_mgr)		     REFERENCES manager(m_id)
				);

       The .doml file would contain:
		  <table id="pkg.Person" notUsingOId="true">
		     <column id="p_fn" isPrimaryKey="true">
				<type dbType="CHAR" javaType="String" size="10"/>
		     </column>
		     <column id="p_ln" isPrimaryKey="true">
				<type dbType="CHAR" javaType="String" size="10"/>
		     </column>
		  </table>
		  <table id="pkg.Manager" notUsingOId="true">
		     <column id="m_id" isPrimaryKey="true">
				<referenceObject reference="pkg.Person"
						      foreignKeyColumnName="p_fn"/>
				<type dbType="CHAR" javaType="String" size="10"/>
		     </column>
		     <column id="m_title">
				<type dbType="CHAR" javaType="String" size="10"/>
		     </column>
		  </table>
		  <table id="Employee">
		     <column id="emp_fn">
				<referenceObject reference="pkg.Person"
						      foreignKeyColumnName="p_fn" foreignKeyGroup="emp"/>
				<type dbType="CHAR" javaType="String" size="10"/>
		     </column>
		     <column id="emp_ln">
				<referenceObject reference="pkg.Person"
						      foreignKeyColumnName="p_ln" foreignKeyGroup="emp"/>
				<type dbType="CHAR" javaType="String" size="10"/>
		     </column>
		     <column id="spouse_fn">
				<referenceObject reference="pkg.Person"
						      foreignKeyColumnName="p_fn" foreignKeyGroup="spouse"/>
				<type dbType="CHAR" javaType="String" size="10"/>
		     </column>
		     <column id="spouse_ln">
				<referenceObject reference="pkg.Person"
						      foreignKeyColumnName="p_ln" foreignKeyGroup="spouse"/>
				<type dbType="CHAR" javaType="String" size="10"/>
		     </column>
		  </table>


		    This routine returns a Hashtable where the referenced DOs are keys.
		    The value for each hash key (referenced DO) is another hashtable
		    keyed by foreignKeyGroup name.
		    The value for each key (foreignKeyGroup) is a vector {}
		    of primary-key/foreign-key name pairs [].

		    When called for the 'employee' DataObject, this routine returns:

				<PersonDO>  :  "empGroup"     : { [ "p_fn", "e_fn" ],
												  [ "p_ln", "e_ln" ]
												}
						       "spouseGroup"  : { [ "p_fn", "spouse_fn" ],
												  [ "p_ln", "spouse_ln" ]
												}
				<ManagerDO> :  "unnamedGroup" : { [ "m_id", "e_mgr" ]
												}

		    This is the information needed to construct the
		    FOREIGN KEY clauses in "CREATE TABLE employee".

		    It is also useful for building setQuery methods.
		*/
		Hashtable ret = new Hashtable();
		Vector attrs = getAttributes();
		for ( int i = 0; i < attrs.size(); i++ ) {
		    Attribute a = (Attribute) attrs.elementAt(i);

//System.err.println(a.toLongString());

//System.err.println("jj a.getName()="+a.getName());
//System.err.println("a.isRefConstraint()="+a.isRefConstraint());
		    if ( a.isRefConstraint() ) {
				String nameOfPKinReferencedTable = a.getForeignKeyColumnName();
//System.err.println("nameOfPKinReferencedTable ="+nameOfPKinReferencedTable );
				if ( null != nameOfPKinReferencedTable &&
				     0 != nameOfPKinReferencedTable.length() ) {
				    // a is a foreign key Attribute
				    DataObject referencedDO = a.getObjectRefDO();
//System.err.println("referencedDO ="+referencedDO );
				    Hashtable groups = (Hashtable) ret.get( referencedDO );
//System.err.println("groups ="+groups );
				    if ( null == groups ) {
						groups = new Hashtable();
						ret.put( referencedDO, groups );
				    }
				    String foreignKeyGroup = a.getForeignKeyGroup();
//System.err.println("foreignKeyGroup ="+foreignKeyGroup );
				    Vector pairs = (Vector) groups.get( foreignKeyGroup );
//System.err.println("pairs ="+pairs );
				    if ( null == pairs ) {
						pairs = new Vector();
						groups.put( foreignKeyGroup, pairs );
				    }
				    String[] pk_fk_pair = new String[ 2 ];
				    pk_fk_pair[0] = nameOfPKinReferencedTable;
				    pk_fk_pair[1] = a.getName(); // name of FK column
				    pairs.addElement( pk_fk_pair );
				}
		    }
		}
		return ret;
    }

    // used by (move to?) SourceGenerator_Query
    public Vector getPKFKpairsForReferencedDO( DataObject referencedDO ) {
		Hashtable fkByTable = getForeignKeysGroupedByReferencedDO();
		return (Vector) fkByTable.get( referencedDO );
    }

    // not needed?
    public boolean foreignKeysMatchPrimaryKeys( Attribute[] fk ) {
		Hashtable h = new Hashtable();
		Attribute[] pk = getPrimaryKeys();
		for ( int i = 0; i < pk.length; i++ ) {
		    h.put( pk[i].getName(), pk[i] );
		}
		for ( int j = 0; j < fk.length; j++ ) {
		    if ( fk[j].isRefConstraint() )
				h.remove( fk[j].getForeignKeyColumnName() );
		}
		return 0 == h.size();
    }

    public boolean hasResultSet() {
		Vector attrs = getAttributes();
		for ( int i = 0; i < attrs.size(); i++ ) {
		    Attribute a = (Attribute) attrs.elementAt(i);
		    if ( ! a.isSpInParam() && ! a.isSpOutParam() )
				    return true;
		}
		return false;
    }

}
