/*
 * Enhydra Java Application Server Project
 *
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific terms governing rights and limitations
 * under the License.
 *
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 * $Id: DOMLDocumentFactory.java,v 1.1 2001/09/28 16:01:07 cryd0221 Exp $
 */
package org.enhydra.dods.store.doml;

// DODS imports
import org.enhydra.dods.store.doa.Attribute;
import org.enhydra.dods.store.doa.DataObject;
import org.enhydra.dods.store.doa.DOPackage;
import org.enhydra.dods.store.doa.DODSProject;

// Common DOM imports
import org.w3c.dom.Document;
import org.w3c.dom.DocumentType;
import org.w3c.dom.DOMImplementation;
import org.w3c.dom.Element;
import org.w3c.dom.Text;
import org.xml.sax.InputSource;

// Standard imports
import java.util.Vector;
import java.util.Enumeration;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.io.IOException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.BufferedWriter;
import java.io.PrintWriter;

// //
// / This needs to be converted to only write out using new attribute
// / names.

/**
 * This class is used to create DOML files out of DODS Project trees.
 *
 * @author Lutris Technologies
 */
public class DOMLDocumentFactory implements DOMLConstants {
    private String vendorName = new String();

    /**
     * Method declaration
     *
     *
     * @param target
     * @param elementName
     * @param text
     *
     * @see
     */
    private void appendTextElement(Element target, String elementName,
                                   String text) {
        Document doc = target.getOwnerDocument();
        Element  textElement = doc.createElement(elementName);
        Text     textNode = doc.createTextNode(text);

        target.appendChild(textElement);
        textElement.appendChild(textNode);
    }

    /**
     * Method declaration
     *
     *
     * @param doc
     * @param attr
     *
     * @return
     *
     * @see
     */
    private Element generateAttributeElement(Document doc,
                                             Attribute fieldDO) {
        Element fieldElement = doc.createElement(FIELD);
        String  stringValue = null;

        // name
        fieldElement.setAttribute(keyTrim(FIELD_NAME), fieldDO.getName());

        // index
        if (fieldDO.isIndex()) {
            fieldElement.setAttribute(keyTrim(FIELD_INDEX), Boolean.TRUE.toString());
        }

        // primary-key
        if (fieldDO.isPrimaryKey()) {
            fieldElement.setAttribute(keyTrim(FIELD_PRIMARY), Boolean.TRUE.toString());
        }

        // can-query
        if (fieldDO.isUsedForQuery()) {
            fieldElement.setAttribute(keyTrim(FIELD_CANQUERY), Boolean.TRUE.toString());
        }

        // constant
        if (fieldDO.isConstant()) {
            fieldElement.setAttribute(keyTrim(FIELD_CONSTANT), Boolean.TRUE.toString());
        }

        // java-type
        stringValue = fieldDO.getJavaType();
        if (stringValue != null && stringValue.trim().length() > 0) {
            fieldElement.setAttribute(keyTrim(FIELD_JAVATYPE), stringValue);
        }

        // database-type
        stringValue = fieldDO.getDBType();
        if (stringValue != null && stringValue.trim().length() > 0) {
            fieldElement.setAttribute(keyTrim(FIELD_DATABASETYPE), stringValue);
        }

        // size
        stringValue = fieldDO.getSize();
        if (stringValue != null && stringValue.trim().length() > 0) {
            fieldElement.setAttribute(keyTrim(FIELD_SIZE), stringValue);
        }

        // null-allowed
        if (fieldDO.canBeNull()) {
            fieldElement.setAttribute(keyTrim(FIELD_NULLALLOWED), Boolean.TRUE.toString());
        }

        // default
        stringValue = fieldDO.getInitialValue();
        if (stringValue != null && stringValue.length() > 0) {
            appendTextElement(fieldElement, DEFAULT, stringValue);
        }

        // javadoc
        stringValue = fieldDO.getJavadoc();
        if (stringValue != null) {
            String spaced = stringValue.replace('*', ' ').replace('/',
                                                ' ').trim();

            if (spaced.length() > 0) {
                stringValue = expandEndOfLine(stringValue);
                appendTextElement(fieldElement, keyTrim(FIELD_JAVADOC), stringValue);
            }
        }

        // constraint and foreign keys
        DataObject objectRef = (DataObject) fieldDO.getObjectRefDO();
        if (objectRef != null ) {
	    if ( fieldDO.isRefConstraint()) {
	        fieldElement.setAttribute(keyTrim(FIELD_CONSTRAINT),
						Boolean.TRUE.toString());
	    } else {
		fieldElement.setAttribute(keyTrim(FIELD_CONSTRAINT),
						Boolean.FALSE.toString());
	    }
	    if ( fieldDO.isForeignKey() ) {
		fieldElement.setAttribute(keyTrim(FIELD_FK_COLUMN),
			fieldDO.getForeignKeyColumnName() );
		fieldElement.setAttribute(keyTrim(FIELD_FK_GROUP),
			fieldDO.getForeignKeyGroup() );

	    }
        }

	// stored procedure parameters / return values
	String spParamType = fieldDO.getSpParamType();
	if ( null == spParamType )
	    spParamType = SP_RESULTSET;
        fieldElement.setAttribute(keyTrim(FIELD_SP_PARAM_TYPE), spParamType);
	//
	String spParamIndex = "" + fieldDO.getSpParamIndex();
        fieldElement.setAttribute(keyTrim(FIELD_SP_PARAM_INDEX), spParamIndex);

        return fieldElement;
    }

    /**
     * Method declaration
     *
     *
     * @param dataObject
     * @param parentElement
     *
     * @see
     */
    private void placeDataObjectInDocument(DataObject dataObject,
                                           Element parentElement) {
        Document doc = parentElement.getOwnerDocument();
        String   nodeType = CLASS;
        boolean  printTable = true;
        String   name = dataObject.getName();

        if (dataObject instanceof DOPackage) {
            printTable = false;
            nodeType = PACKAGE;
        }
        Element element = null;

        if (dataObject instanceof DOPackage) {
            element = createPackageElement(doc, name);
        } else {
            element = createClassElement(doc, name, dataObject);
        }
        parentElement.appendChild(element);
        if (printTable) {
            if (name.equals(dataObject.getDBTableName()) == false) {
                element.setAttribute(keyTrim(CLASS_TABLE), dataObject.getDBTableName());
            }
        }
        Vector attributes = dataObject.getAttributes();

        if (attributes != null) {
            Enumeration e = attributes.elements();

            while (e.hasMoreElements()) {
                Attribute attribute = (Attribute) e.nextElement();

                element.appendChild(generateAttributeElement(doc, attribute));
            }
        }
        if (nodeType.equalsIgnoreCase(PACKAGE)) {
            DOPackage dopackage = (DOPackage) dataObject;
            Vector    vector = dopackage.getDataObjects();

            if (vector != null) {
                Enumeration e = vector.elements();

                while (e.hasMoreElements()) {
                    DataObject childDataObject = (DataObject) e.nextElement();

                    placeDataObjectInDocument(childDataObject, element);
                }
            }
        }
    }

    /**
     * Method declaration
     *
     *
     * @param doc
     * @param name
     *
     * @return
     *
     * @see
     */
    private Element createPackageElement(Document doc, String name) {
        Element element = null;

        element = doc.createElement(PACKAGE);
        element.setAttribute(keyTrim(PACKAGE_NAME), name);
        return element;
    }

    /**
     * Method declaration
     *
     *
     * @param doc
     * @param name
     * @param classDO
     *
     * @return
     *
     * @see
     */
    private Element createClassElement(Document doc, String name,
                                       DataObject classDO) {
        Element classElement = null;

        classElement = doc.createElement(CLASS);
        classElement.setAttribute(keyTrim(PACKAGE_NAME), name);

        // lazy-loading
        if (classDO.isLazyLoading()) {
            classElement.setAttribute(keyTrim(CLASS_LAZYLOADING), Boolean.TRUE.toString());
        }

        // use-oid
        if (classDO.notUsingOId()) {
            classElement.setAttribute(keyTrim(CLASS_OID), Boolean.FALSE.toString());
        } else {
            classElement.setAttribute(keyTrim(CLASS_OID), Boolean.TRUE.toString());
        }

        // is-stored-proc
        if (classDO.isStoredProc()) {
            classElement.setAttribute(keyTrim(CLASS_IS_STORED_PROC), Boolean.TRUE.toString());
        } else {
            classElement.setAttribute(keyTrim(CLASS_IS_STORED_PROC), Boolean.FALSE.toString());
        }

        // caching
        if (classDO.isFullCaching()) {
            classElement.setAttribute(keyTrim(CLASS_CACHING), CACHE_FULL);
        } else if (classDO.isCaching()) {
            classElement.setAttribute(keyTrim(CLASS_CACHING), CACHE_PARTIAL);
        }

        // modifier
        if (classDO.isFinal()) {
            classElement.setAttribute(keyTrim(CLASS_MODIFIER), FINAL);
        } else if (classDO.isAbstract()) {
            classElement.setAttribute(keyTrim(CLASS_MODIFIER), ABSTRACT);
        }


        // extends
        DataObject extension = classDO.getExtensionOf();

        if (extension != null) {
            classElement.setAttribute(keyTrim(CLASS_EXTENDS),
                                      extension.getPackageName() + "."
                                      + extension.getName());
        }

        // vendor
        String classVendor = classDO.getDatabase();
        if ((classVendor != null) &&
            (vendorName  != null) &&
            (classVendor.trim().length() > 0) &&
            (vendorName.trim().length()  > 0)) {
            classVendor = classVendor.trim();
           if ((! classVendor.equalsIgnoreCase(DEFAULT_CLASS_VENDOR)) &&
               (! classVendor.equalsIgnoreCase(vendorName))) {
             classElement.setAttribute(keyTrim(CLASS_VENDOR),
                                       classVendor);
          }
       }


        return classElement;
    }

    /**
     * Constructor. Each factory object should be able to create many documents
     * out of many given projects without having to call for new factory
     * objects (although there is no real saving in doing so).
     */
    public DOMLDocumentFactory() {}

    /**
     * Create a DOM Document out of an existing project data structure.
     *
     * @param project
     * The DODSProject tree structure that will be read.
     * @return
     * A Document which is a DOML style DOM object.
     */
    public Document createDocument(DODSProject project) throws DOMLException {
        Document doc = createEmptyDOML();
        Element  dbElement = doc.createElement(VENDOR);

        vendorName = project.getDatabase();

        doc.getDocumentElement().appendChild(dbElement);
        placeDataObjectInDocument(project, dbElement);
        if (vendorName != null && vendorName.trim().length() > 0) {
            dbElement.setAttribute(keyTrim(VENDOR_NAME), vendorName);
        }
        return doc;
    }

    /**
     * Method declaration
     *
     *
     * @return
     *
     * @see
     */
    private Document createEmptyDOML() {
        File     tempFile = null;
        Document doc = null;

        try {
            tempFile = createTempDOMLFile();
        } catch (IOException e) {
            e.printStackTrace();
            tempFile = null;
        }
        if (tempFile != null) {
            doc = parseFile(tempFile);
        }
        return doc;
    }

    // Create a document using the current parser: Xerces or IBM.

    /**
     * Method declaration
     *
     *
     * @return
     *
     * @exception IOException
     *
     * @see
     */
    private File createTempDOMLFile() throws IOException {
        File           tempFile = null;
        FileWriter     fileWriter = null;
        BufferedWriter bufWriter = null;
        PrintWriter    printWriter = null;

        tempFile = File.createTempFile("tmp", ".tmp");
        tempFile.deleteOnExit();
        fileWriter = new FileWriter(tempFile);
        bufWriter = new BufferedWriter(fileWriter);
        printWriter = new PrintWriter(bufWriter);
        printWriter.println();
        printWriter.println("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
        printWriter.println("<doml/>");
        printWriter.flush();
        bufWriter.flush();
        printWriter.close();
        printWriter = null;
        bufWriter = null;
        fileWriter = null;
        return tempFile;
    }

    /**
     * Method declaration
     *
     *
     * @param in
     *
     * @return
     *
     * @see
     */
    private Document parseFile(File inFile) {
        final String    XERCES_PARSER = "org.apache.xerces.parsers.DOMParser";
        final String    IBM_PARSER = "com.ibm.xml.parsers.DOMParser";
        Document        doc = null;
        Object          parser = null;
        Class           parserClass = null;
        FileInputStream inStream = null;
        InputSource     inputSource = null;

        try {
            inStream = new FileInputStream(inFile);
            inputSource = new InputSource(inStream);
            Constructor[] cons = new Constructor[0];
            Constructor   constructor = null;
            Class[]       params = new Class[0];
            Object[]      initArgs = new Object[0];

            try {
                parserClass = Class.forName(IBM_PARSER);
            } catch (ClassNotFoundException e) {
                parserClass = Class.forName(XERCES_PARSER);
            }
            cons = parserClass.getConstructors();
            for (int i = 0; i < cons.length; i++) {
                params = cons[i].getParameterTypes();

                // stop on non-default constructors.
                if (params.length == 0) {
                    constructor = cons[i];
                    break;
                }
            }
            parser = constructor.newInstance(initArgs);
        } catch (Exception e) {
            parser = null;
            e.printStackTrace();
        }

        //
        //
        if (parser != null) {

            // parser.parse(inStream);
            Method[] meths = parserClass.getMethods();
            Method   parse = null;
            Method   getDocument = null;
            Class[]  paramTypes = null;

            for (int i = 0; i < meths.length; i++) {
                if (meths[i].getName().equals("parse")) {
                    Class[] pt = meths[i].getParameterTypes();

                    if (pt.length == 1) {
                        if (pt[0].getName().equals("org.xml.sax.InputSource")) {
                            parse = meths[i];
                        }
                    }
                } else if (meths[i].getName().equals("getDocument")) {
                    getDocument = meths[i];
                }
            }
            if (parse != null) {
                Object paramValues[] = new Object[1];

                paramValues[0] = inputSource;
                try {
                    parse.invoke(parser, paramValues);
                    paramValues = new Object[0];
                    doc = (Document) getDocument.invoke(parser, paramValues);
                } catch (Exception e) {
                    e.printStackTrace();
                }
            }
        }
        return doc;
    }

    /**
     * Method declaration
     *
     *
     * @param in
     *
     * @return
     *
     * @see
     */
    private String expandEndOfLine(String in) {
        String out = in;

        if (out != null) {
            StringBuffer buf = new StringBuffer();

            buf.append(out);
            int index = -1;
            int prev = -1;

            index = buf.toString().indexOf(10);
            if (index > 0) {
                prev = buf.charAt(index - 1);
            }
            while (index > -1 && prev != 13) {
                buf.insert(index, (char) 13);
                index = buf.toString().indexOf(10, index + 2);
                if (index > 0) {
                    prev = buf.charAt(index - 1);
                } else {
                    break;
                }
            }
            out = buf.toString();
        }
        return out;
    }

    public String keyTrim(String in) {
       String out = in;
       int index = -1;
       if (out != null) {
         index = out.lastIndexOf('.');
         if ((index > -1) && (out.length() > (index + 1))) {
           out = out.substring(index + 1);
         }
       }
       return out;
    }

}
