/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: DOMLFactory.java,v 1.1 2001/09/28 16:01:07 cryd0221 Exp $
 */

package org.enhydra.dods.store.doml;

import org.enhydra.dods.store.doa.DODSProject;
import java.io.InvalidClassException;
import java.io.File;
import java.io.FileInputStream;
import java.io.ObjectInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import org.w3c.dom.*;

/**
 * This class is mostly an example use of the
 * <b>org.enhydra.dods.project.doml</b> package. It
 * contains just a main method, which can be used for converting an existing
 * DOA file into a DOML file, or a DOML file of any format into a nicely
 * humanly-readable, indented, but possibly incorrect DOML file.
 *
 * This file refers to both the old style dods package
 * <code>dods.data</code> and the new package <code>
 * org.enhydra.dods.store.doa</code>
 *
 * @author Lutris Technologies
 */
public class DOMLFactory {

  private static void printWarnings(DOMLWarning warnings[]) {
    for (int i = 0; i<warnings.length; i++) {
      System.err.println("Warning: " + warnings[i].getMessage());
      Throwable chainedException = warnings[i].getCause();
      if (chainedException != null)
      System.err.println("         " + chainedException.toString());
    }
  }

  public static org.enhydra.dods.store.doa.DODSProject readDOML(File file) {
    DODSProject project = null;
    DOMLProjectFactory projFactory = new DOMLProjectFactory();

    try {
      project = projFactory.createProject(file);
      if (projFactory.hasWarnings())
      printWarnings(projFactory.getWarnings());
    } catch (Exception e) {
      System.err.println("ERROR: Failed to read DOML file " +
                 file.getAbsolutePath() + ", " + e.toString());
      e.printStackTrace();
    }
    if (project == null) {
      System.err.println("ERROR: DOML file " +
                 file.getAbsolutePath() +
                 " creates null project");
    }

    return project;
  }


  public static void writeDOML(DODSProject project,
                               File        outFile,
                               boolean     indent) {
    DOMLDocumentFactory docFactory = new DOMLDocumentFactory();
    Document             doc = null;

    try {
      doc = docFactory.createDocument(project);
    } catch (DOMLException e) {
      System.err.println(
         "ERROR: Failed to create document, " + e.toString());
      e.printStackTrace();
    }

    try {
      FileOutputStream out = new FileOutputStream(outFile);
      DOMLWriter      writer = new DOMLWriter(out, false, indent, false);
      writer.print(doc);
    } catch (Exception e) {
      System.err.println(
        "ERROR: Failed to write file "
        + outFile.getAbsolutePath()
        + ", " +   e.toString());
      e.printStackTrace();
    }
  }

  /**
   * This method is mostly an example use of the
   * <b>org.enhydra.dods.project.doml</b> package,
   * which shows how to convert an existing DOA file into a DOML file, or
   * a DOML file of any format into a nicely humanly-readable, indented,
   * but possibly incorrect DOML file.
   * <p>
   * Usage:
   * <dl>
   *  <dt><code>java org.enhydra.dods.project.doml.DOMLFactory
   *      file.doa</code>
   *  <dd> converts a DOA file into a DOML file of the same name
   *  <dt><code>java org.enhydra.dods.project.doml.DOMLFactory -i
   *      file.doml file2.doml</code>
   *  <dd> converts a DOML file into a humanly readable DOML file, with
   *       possibly illegal DOML.DTD structure.
   *  <dt><code>java org.enhydra.dods.project.doml.DOMLFactory
   *      file.doml</code>
   *  <dd> converts a DOML file into a DOA file.
   * </dl>
   *
   * @param args
   *  This String array should contain one and only one element, the name of
   *  the file to parse.
   */
  public static void main(String[] args) {
    boolean indent = false;
    int     currentArg = 0;
    DODSProject project = null;

    if (args == null || args.length == 0) {
      System.err.println("Usage: DOMLFactory [-i] <readFile> [<writeFile>]");
      System.exit(1);
    }


    if (args[currentArg].equals("-i")) {
      currentArg++;
      indent = true;
    }

    File fromFile = new File(args[currentArg]);
    currentArg++;

    File   toFile = null;
    String toType = "doml";

    if (args.length > currentArg) {
      toFile = new File(args[currentArg]);
      toType = org.enhydra.dods.util.Util.getExtension(toFile);
      if (toType == null) {
        System.err.println("ERROR: Can't handle file " +
                   toFile.getAbsolutePath() +
                   ", no \".\" in the file name");
        System.exit(1);
      }
      currentArg++;
    } else {
      String fileBase = org.enhydra.dods.util.Util.getBase(fromFile);
      if (fromFile.getAbsolutePath().toLowerCase().endsWith(".doml")) {
        toType = "doa";
      }
      toFile = new File(fromFile.getParent(), fileBase + "." + toType);
    }

    if (fromFile.getAbsolutePath().toLowerCase().endsWith(".doa")) {
      System.err.println(
          "ERROR: Use the dods.doml.Upgrade class to convert DOA files to DOML files");
      System.exit(1);
    } else if (fromFile.getAbsolutePath().toLowerCase().endsWith(".doml")) {
      project = readDOML(fromFile);
    } else {
      System.err.println("ERROR: File " + fromFile.getAbsolutePath() +
                 " format unknown");
      System.exit(1);
    }

    if (toType.equals("doml")) {
      writeDOML(project, toFile, indent);
    } else {
      System.err.println("ERROR: File " + toFile.getAbsolutePath() +
                 " format " + toType + " unknown");
      System.exit(1);
    }
    System.out.println("Done");
  }
}
