/*
 * Enhydra Java Application Server Project
 *
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific terms governing rights and limitations
 * under the License.
 *
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 * $Id: DOMLProjectFactory.java,v 1.1 2001/09/28 16:01:07 cryd0221 Exp $
 */
package org.enhydra.dods.store.doml;

// DODS imports
import org.enhydra.dods.store.doa.*;
import org.enhydra.dods.exceptions.*;

// Xerces imports
import org.apache.xerces.parsers.DOMParser;

// Standard DOM imports
import org.w3c.dom.*;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXParseException;

// Standard imports
import java.io.*;
import java.util.*;

/**
 *
 * This class is used to create DODS Project trees out of DOML files.
 *
 * @author Lutris Technologies
 *
 */
public class DOMLProjectFactory implements DOMLConstants {
	private String	 vendorName = new String();
	private Hashtable  dataObjectHash = null;
	private Hashtable  dataObjectExtensionPendingHash = null;
	private Hashtable  dataObjectReferencePendingHash = null;
	private Vector	 warningVector = null;
	private Properties keyMap = null;
	private Properties valueMap = null;
	private boolean	current = true;

	/**
	 * Create a DOML Project producting factory. A single DOML Project
	 * Factory can be used repeatedly for multiple calls to createProject.
	 */
	public DOMLProjectFactory() {
		initUpgradeMaps();
	}

	/**
	 * Create a DODS DODSProject tree from the given DOM Document (which should
	 * conform to the DOML.DTD)
	 *
	 * @param file
	 * The File that contains a DOML DOM Document.
	 *
	 * @return
	 * A DODSProject data structure.
	 *
	 * @exception FileNotFoundException
	 * Thrown if the given file could not be found.
	 *
	 * @exception IOException
	 * Thrown if there is a problem reading the file.
	 *
	 * @exception SAXException
	 * Thrown if there is a problem parsing the file as a DOM object
	 *
	 * @exception DOMLException
	 * Thrown if there is a problem parsing the file as a DOML object
	 */
	public DODSProject createProject(File file)
			throws FileNotFoundException, IOException, SAXException,
				   DOMLException {
		DOMLUpdate update = new DOMLUpdate(file);

		current = update.isCurrent();
		if (current) {
			file = update.removeDocType();
		}
		FileInputStream fis = new FileInputStream(file);

		// Create an input source for the document we want to parse.
		// First line provides the actual input stream, the second
		// line specifies the system identifier for error reporting.
		InputSource	 input = new InputSource(fis);

		input.setSystemId(file.getCanonicalPath());

		// Create a new DOM parser with a standard error handler.
		DOMParser parser = new DOMParser();

		// Parse the document.
		parser.parse(input);

		// Retrieve the document directly from the parser.
		Document doc = parser.getDocument();

		return createProject(doc);
	}

	/**
	 * Create a DODS DODSProject tree from the given DOM Document (which should
	 * conform to the DOML.DTD)
	 *
	 * @param doc
	 *
	 * The Document that contains a DOML DOM Document.
	 *
	 * @return
	 *
	 * A DODSProject data structure.
	 * @exception DOMLException
	 *
	 * Thrown if there is a problem parsing the file as a DOML object
	 */
	public DODSProject createProject(Document doc) throws DOMLException {
		DODSProject  project = null;
		NodeList	 elements = null;
		Node		 vendorNode = null;
		Node		 rootPackage = null;
		Node		 rootPackageNameNode = null;
		NamedNodeMap map = null;

		dataObjectExtensionPendingHash = null;
		dataObjectReferencePendingHash = null;
		warningVector = null;
		dataObjectHash = new Hashtable();
		elements = doc.getElementsByTagName(key(VENDOR));
		if (elements.getLength() == 0) {
			throw new DOMLException("No vendor elements in file");
		}
		if (elements.getLength() > 1) {
			throw new DOMLException("Only one vendor allowed per file");
		}
		vendorNode = elements.item(0);
		if (vendorNode == null) {
			throw new DOMLException("Vendor element is corrupt.");
		}
		map = vendorNode.getAttributes();
		vendorName = getAttributeValue(map, key(VENDOR_NAME));
		if (vendorName != null) {
			vendorName = value(VENDOR_NAME, vendorName);
		}
		rootPackage = vendorNode.getFirstChild();
		while (rootPackage != null
			   && rootPackage.getNodeName().equals(PACKAGE) == false) {
			rootPackage = rootPackage.getNextSibling();
		}
		if (rootPackage == null) {
			throw new DOMLException("Root package not found.");
		}
		map = rootPackage.getAttributes();
		if (map == null) {
			throw new DOMLException("Root package is empty.");
		}
		rootPackageNameNode = map.getNamedItem(key(PACKAGE_NAME));
		if (rootPackageNameNode == null) {
			throw new DOMLException("Name of root package not found.");
		}
		project = new DODSProject(rootPackageNameNode.getNodeValue());
		project.setDatabase(vendorName);
		placeNodeIntoPackageDataObject(rootPackage, project);
		if (dataObjectReferencePendingHash != null) {
			Enumeration e = dataObjectReferencePendingHash.keys();

			while (e.hasMoreElements()) {
				String	  name = (String) e.nextElement();
				Vector	  vector =
					(Vector) dataObjectReferencePendingHash.get(name);
				Enumeration v = vector.elements();

				while (v.hasMoreElements()) {
					Attribute attr = (Attribute) v.nextElement();

					registerWarning(new DOMLWarning("Reference for class "
													+ name + " by attribute "
													+ attr.getName()
													+ " unresolved."));
				}
			}
		}
		if (dataObjectExtensionPendingHash != null) {
			Enumeration e = dataObjectExtensionPendingHash.keys();

			while (e.hasMoreElements()) {
				String	  name = (String) e.nextElement();
				Vector	  vector =
					(Vector) dataObjectExtensionPendingHash.get(name);
				Enumeration v = vector.elements();

				while (v.hasMoreElements()) {
					DataObject classDO = (DataObject) v.nextElement();

					registerWarning(new DOMLWarning("Extension for class "
													+ name + " by class "
													+ classDO.getPackageName()
													+ "." + classDO.getName()
													+ " unresolved."));
				}
			}
		}
		return project;
	}

	/**
	 * Determine if the recent call to <b>createProject</b> produced parsing
	 * warnings relative to the current implementation of DODs and the DOML
	 * parser.
	 *
	 * @return
	 * the boolean true if parsing the document caused warnings.
	 */
	public boolean hasWarnings() {
		return (warningVector != null);
	}

	/**
	 * Fetch the array of warnings produced by the recent call to
	 * createProject.
	 *
	 * @return
	 * the DOMLWarning array, which is guarenteed to be non-null, but might
	 * be length zero.
	 */
	public DOMLWarning[] getWarnings() {
		int		 arraySize = (warningVector == null) ? 0
								: warningVector.size();
		DOMLWarning retVal[] = new DOMLWarning[arraySize];

		if (warningVector != null) {
			warningVector.copyInto(retVal);
		}
		return retVal;
	}

	// /
	// / PRIVATE
	// /

	/**
	 * Method declaration
	 *
	 *
	 * @param warning
	 *
	 * @see
	 */
	private void registerWarning(DOMLWarning warning) {
		if (warningVector == null) {
			warningVector = new Vector();
		}
		warningVector.addElement(warning);
	}

	/**
	 * Method declaration
	 *
	 *
	 * @param classDO
	 * @param map
	 *
	 * @see
	 */
	private void placeAttributesIntoClassDataObject(DataObject classDO,
			NamedNodeMap map) {
//System.out.println ("Loading Attributes for ClassDataObject:"+classDO);
		Node	classAttr = null;
		String  stringValue = new String();
		boolean booleanValue = false;

		// table
		classAttr = map.getNamedItem(key(CLASS_TABLE));
		try {
			if (classAttr == null) {
				classDO.setDBTableName(classDO.getName());
			} else {
				classDO.setDBTableName(classAttr.getNodeValue());
			}
		} catch (DataObjectNameException done) {
			registerWarning(new DOMLWarning(done,
											"ERROR: Illegal table name "
											+ classDO.getName()));
		}

		// lazy-loading
		booleanValue = isAttributeValue(map, key(CLASS_LAZYLOADING));
		classDO.setLazyLoading(booleanValue);

		// use-oid
		if (current) {
			booleanValue = isAttributeValue(map, key(CLASS_OID), true);
//System.err.println("4 booleanValue ="+booleanValue );
			classDO.setNotUsingOId(!booleanValue);
		} else {
			booleanValue = isAttributeValue(map, V3_OID, false);
//System.err.println("3 booleanValue ="+booleanValue );
			classDO.setNotUsingOId(booleanValue);
		}

		// is-stored-proc
		booleanValue = isAttributeValue(map, key(CLASS_IS_STORED_PROC), false);
		classDO.setIsStoredProc(booleanValue);

		// modifier - combines final and abstract elements from v3.
		if (current) {
			stringValue = getAttributeValue(map, key(CLASS_MODIFIER));
			booleanValue = stringValue.equalsIgnoreCase(FINAL);
		} else {
			booleanValue = isAttributeValue(map, V3_FINAL);
		}
		classDO.setFinal(booleanValue);
		if (booleanValue) {
			booleanValue = false;
		} else {
			if (current) {
				booleanValue = stringValue.equalsIgnoreCase(ABSTRACT);
			} else {
				booleanValue = isAttributeValue(map, V3_ABSTRACT);
			}
		}
		classDO.setAbstract(booleanValue);

		// caching
		stringValue = getAttributeValue(map, key(CLASS_CACHING), CACHE_NONE);
		if (stringValue.equalsIgnoreCase(CACHE_NONE)) {
			classDO.setFullCaching(false);
			classDO.setCaching(false);
		} else if (stringValue.equalsIgnoreCase(CACHE_PARTIAL)) {
			classDO.setFullCaching(false);
			classDO.setCaching(true);
		} else {
			classDO.setFullCaching(true);
			classDO.setCaching(true);
		}

		// extends
		stringValue = getAttributeValue(map, key(CLASS_EXTENDS));
		if (stringValue.length() > 0) {
			DataObject extendsDO = null;

			extendsDO = (DataObject) dataObjectHash.get(stringValue);
			if (extendsDO == null) {
				addPendingExtensionWatch(stringValue, classDO);
			} else {
				classDO.setExtensionOf(extendsDO);
			}
		}

		// index
		classDO.setIndex(false);	// unused and no longer in DTD

		// vendor
		stringValue = getAttributeValue(map, key(CLASS_VENDOR));
		if (stringValue == null || stringValue.trim().length() == 0
				|| stringValue.trim().equalsIgnoreCase(DEFAULT_CLASS_VENDOR)) {
			stringValue = vendorName;
		}
		classDO.setDatabase(stringValue);
	}

	/**
	 * Method declaration
	 *
	 *
	 * @param node
	 * @param classDO
	 * @param attributes
	 *
	 * @see
	 */
	private void placeNodeIntoClassDataObject(Node node, DataObject classDO,
											  NamedNodeMap attributes) {

		/* We store the classDO for future lookup by ref objects/extensions */
		String hashindex = classDO.getPackageName() + "." + classDO.getName();

		if (classDO.getPackageName() == null
				|| classDO.getPackageName().length() == 0) {
			hashindex = classDO.getName();
		}
		watchClass(hashindex, classDO);
		if (attributes != null) {
			placeAttributesIntoClassDataObject(classDO, attributes);
		}
		NodeList fields = node.getChildNodes();

		if (fields != null) {
			for (int i = 0; i < fields.getLength(); i++) {
				Node fieldNode = fields.item(i);

				placeNodeIntoFieldDataObject(fieldNode, classDO);
			}
		}
	}

	/**
	 * Method declaration
	 *
	 *
	 * @param name
	 * @param classDO
	 *
	 * @see
	 */
	private void watchClass(String name, DataObject classDO) {
		dataObjectHash.put(name, classDO);
		if (dataObjectExtensionPendingHash != null
				&& dataObjectExtensionPendingHash.containsKey(name)) {
			Vector	  vector =
				(Vector) dataObjectExtensionPendingHash.get(name);
			Enumeration e = vector.elements();

			while (e.hasMoreElements()) {
				DataObject extendedTable = (DataObject) e.nextElement();

				extendedTable.setExtensionOf(classDO);
			}
			dataObjectExtensionPendingHash.remove(name);
		}
		if (dataObjectReferencePendingHash != null
				&& dataObjectReferencePendingHash.containsKey(name)) {
			Vector	  vector =
				(Vector) dataObjectReferencePendingHash.get(name);
			Enumeration e = vector.elements();

			while (e.hasMoreElements()) {
				Attribute attribute = (Attribute) e.nextElement();

				attribute.setObjectRef(true);
				attribute.setObjectRefDO(classDO);
			}
			dataObjectReferencePendingHash.remove(name);
		}
	}

	/**
	 * Method declaration
	 *
	 *
	 * @param name
	 * @param classDO
	 *
	 * @see
	 */
	private void addPendingExtensionWatch(String name, DataObject classDO) {
		Vector vector = null;

		if (dataObjectExtensionPendingHash == null) {
			dataObjectExtensionPendingHash = new Hashtable();
		}
		if (dataObjectExtensionPendingHash.containsKey(name)) {
			vector = (Vector) dataObjectExtensionPendingHash.get(name);
		} else {
			vector = new Vector();
		}
		vector.addElement(classDO);
		dataObjectExtensionPendingHash.put(name, vector);
	}

	/**
	 * Method declaration
	 *
	 *
	 * @param name
	 * @param attr
	 *
	 * @see
	 */
	private void addPendingReferenceWatch(String name, Attribute attr) {
		Vector vector = null;

		if (dataObjectReferencePendingHash == null) {
			dataObjectReferencePendingHash = new Hashtable();
		}
		if (dataObjectReferencePendingHash.containsKey(name)) {
			vector = (Vector) dataObjectReferencePendingHash.get(name);
		} else {
			vector = new Vector();
		}
		vector.addElement(attr);
		dataObjectReferencePendingHash.put(name, vector);
	}

	/**
	 * Method declaration
	 *
	 *
	 * @param fieldNode
	 * @param classDO
	 *
	 * @see
	 */
	private void placeNodeIntoFieldDataObject(Node fieldNode,
											  DataObject classDO) {
		if (fieldNode.getNodeType() == Node.TEXT_NODE) {

			/* This isn't a data node, this is just whitespace padding */
			return;
		}

		Attribute	fieldDO = null;
		NamedNodeMap map = fieldNode.getAttributes();
		Node		 idFieldAttr = null;
		Node		 childNode = null;
		String	   fieldName = null;
		String	   stringValue = new String();
		boolean	  booleanValue = false;

		if (map != null) {
			idFieldAttr = map.getNamedItem(key(FIELD_NAME));
		}
		if (idFieldAttr != null) {
			fieldName = idFieldAttr.getNodeValue();
			if (fieldName.length() == 0) {
				fieldName = null;
			}
		}
//System.err.println("fieldName ="+fieldName );


		if (fieldName == null) {
			registerWarning(new DOMLWarning("Unnamed field in class "
											+ classDO.getPackageName() + "."
											+ classDO.getName()));
			return;
		}
		fieldDO = new Attribute(fieldName);

		String spParamIndex = "-1";

		// stored procedure parameters / return value
		Node spTypeAttr = map.getNamedItem(key(FIELD_SP_PARAM_TYPE));
		if ( null != spTypeAttr ) {
			String spParamType = spTypeAttr.getNodeValue();
			if ( null == spParamType || 0 == spParamType.length() ) {
				spParamType = SP_RESULTSET;
				spParamIndex = "1";
			}
			spParamType = spParamType.trim();
			if (! (		spParamType.equals( SP_RETURN_VALUE )		||
					spParamType.equals( SP_IN_PARAM )		||
					spParamType.equals( SP_OUT_PARAM )		||
					spParamType.equals( SP_INOUT_PARAM )		||
					spParamType.equals( SP_RESULTSET )		) ) {
				registerWarning(new DOMLWarning(
					"Class " + classDO.getPackageName() + "." +
					classDO.getName() +
					" field " + fieldName +
					" attribute " + FIELD_SP_PARAM_TYPE +
					" has bad value: " + spParamType + "." ) );
				spParamType = SP_RESULTSET;
			}
			fieldDO.setSpParamType( spParamType );

			// stored procedure parameters have indexes to indicate
			// their order in the stored procedure call.
			if ( "-1".equals( spParamIndex ) ) {
				spParamIndex = null;
				Node spIndexAttr = map.getNamedItem(key(FIELD_SP_PARAM_INDEX));
				if ( null != spIndexAttr )
					spParamIndex = spIndexAttr.getNodeValue();
				if ( null == spParamIndex || 0 == spParamIndex.length() ) {
						registerWarning(new DOMLWarning(
							"Class " + classDO.getPackageName() +
							"." + classDO.getName() +
							" field " + fieldName +
							" attribute " + FIELD_SP_PARAM_INDEX +
							" is missing." ) );
				}
			}
		}
		fieldDO.setSpParamIndex( spParamIndex );

		// no unique
		// no-more error
		fieldDO.setError(false);
		fieldDO.setErrorText(new String());
		classDO.addAttribute(fieldDO);
		Node fieldAttr = map.getNamedItem(key(FIELD_SIZE));

		if (fieldAttr != null) {
			fieldDO.setSize(fieldAttr.getNodeValue());
		}

		// index
		booleanValue = isAttributeValue(map, key(FIELD_INDEX));
		fieldDO.setIndex(booleanValue);

		// primary-key
		booleanValue = isAttributeValue(map, key(FIELD_PRIMARY));
		fieldDO.setPrimaryKey(booleanValue);

		// can-query
		booleanValue = isAttributeValue(map, key(FIELD_CANQUERY));
		fieldDO.setUsedForQuery(booleanValue);

		// constant
		booleanValue = isAttributeValue(map, key(FIELD_CONSTANT));
		fieldDO.setConstant(booleanValue);

		// javadoc
		childNode = lookupChildNode(fieldNode, key(FIELD_JAVADOC));
		stringValue = getTextChildValue(childNode);
		if (stringValue != null) {
			fieldDO.setJavadoc(stringValue);
		}

		// default
		childNode = lookupChildNode(fieldNode, key(DEFAULT));
		stringValue = getTextChildValue(childNode);
		if (stringValue != null) {
			fieldDO.setInitialValue(stringValue);
		}

		// type info
		if (!current) {
			childNode = lookupChildNode(fieldNode, V3_TYPE);
			map = childNode.getAttributes();
		}

		// type: null-allowed
		booleanValue = isAttributeValue(map, key(FIELD_NULLALLOWED));
		fieldDO.setCanBeNull(booleanValue);

		// type: size
		stringValue = getAttributeValue(map, key(FIELD_SIZE));
		if (stringValue.length() > 0) {
			fieldDO.setSize(stringValue);
		}

		// type: database-type
		stringValue = getAttributeValue(map, key(FIELD_DATABASETYPE),
										DEFAULT_DATABASE_TYPE);
		fieldDO.setDBType(stringValue);

		// type: java-type
		stringValue = getAttributeValue(map, key(FIELD_JAVATYPE),
										DEFAULT_JAVA_TYPE);
		//csc_052101_start - This block of code effectivelychecks to see if
		//we are dealing with a 4.0 doml; if so, see if we're dealing with a
		//standard Java type. If not, then we are dealing with a DO, and we
		//need to append the suffix onto the name (provided it doesn't
		//already have the suffix).
		if (current && !stringValue.endsWith(REF_SUFFIX)){
			boolean stdType = false;
			for (int i = 0; i < JAVA_TYPE_LIST.length; i++) {
				if (JAVA_TYPE_LIST[i].equals(stringValue)) {
					stdType = true;
					break;
				}
			}
			if (!stdType) {
				stringValue = stringValue + REF_SUFFIX;
			}
		}
		//csc_052101_end
		fieldDO.setJavaType(stringValue);

		String fkColumn = null;
		String fkGroup = null;

		// Create reference from java-type
		// reference: contraint
		booleanValue = false;
//System.err.println("stringValue="+stringValue);
		if (isJavaTypeReference(stringValue)) {
//System.err.println("here 0");
			String	 refName = null;
			DataObject refObj = null;

			refName = stringValue.substring(0, stringValue.length() - 2);
			refObj = (DataObject) dataObjectHash.get(refName);
			if (refObj != null) {
				fieldDO.setObjectRef(true);
				fieldDO.setObjectRefDO(refObj);
			} else {
				addPendingReferenceWatch(refName, fieldDO);
			}
//System.err.println("current="+current);
			if (current) {
				booleanValue = isAttributeValue(map, key(FIELD_CONSTRAINT));
				fkColumn = getAttributeValue( map, key(FIELD_FK_COLUMN) );
				fieldDO.setForeignKeyColumnName( fkColumn );
				fkGroup = getAttributeValue( map, key(FIELD_FK_GROUP) );
				fieldDO.setForeignKeyGroup( fkGroup );
			} else {
				childNode = lookupChildNode(fieldNode, V3_REFERENCE_OBJ );
				if (childNode != null) {
					map = childNode.getAttributes();
					booleanValue = isAttributeValue(map, key(FIELD_CONSTRAINT));
					fkColumn = getAttributeValue( map, key(FIELD_FK_COLUMN) );
					fieldDO.setForeignKeyColumnName( fkColumn );
					fkGroup = getAttributeValue( map, key(FIELD_FK_GROUP) );
					fieldDO.setForeignKeyGroup( fkGroup );
				}
			}
//System.err.println("fkColumn ="+fkColumn );
//System.err.println("fkGroup ="+fkGroup );
		} else {
//System.err.println("here 1");
//System.err.println("x current="+current);
			if (current) {
				booleanValue = isAttributeValue(map, key(FIELD_CONSTRAINT));
				fkColumn = getAttributeValue( map, key(FIELD_FK_COLUMN) );
				fieldDO.setForeignKeyColumnName( fkColumn );
				fkGroup = getAttributeValue( map, key(FIELD_FK_GROUP) );
				fieldDO.setForeignKeyGroup( fkGroup );
			} else {

//this is where the problem is arising: with the new style doml, we never find a 'reference'
//attribute...see bug 12111 for description.

				childNode = lookupChildNode(fieldNode, V3_REFERENCE_OBJ);
				if (childNode != null) {

					map = childNode.getAttributes();
					booleanValue = isAttributeValue(map, key(FIELD_CONSTRAINT));

					String refName = getAttributeValue( map, V3_REFERENCE );
//System.err.println("refName ="+refName );
					if ( null != refName ) {
							if ( refName.endsWith( "DO" ) )
							refName = refName.substring(0, refName.length()-2);
						DataObject refObj = null;
						refObj = (DataObject) dataObjectHash.get(refName);
//System.err.println("refObj ="+refObj );
						if (refObj != null) {
							fieldDO.setObjectRef(true);
							fieldDO.setObjectRefDO(refObj);
						} else {
							addPendingReferenceWatch(refName, fieldDO);
						}
					}
					fkColumn = getAttributeValue( map, key(FIELD_FK_COLUMN) );
					fieldDO.setForeignKeyColumnName( fkColumn );
					fkGroup = getAttributeValue( map, key(FIELD_FK_GROUP) );
					fieldDO.setForeignKeyGroup( fkGroup );
				}
			}
		}
//System.err.println("fkColumn ="+fkColumn );
//System.err.println("fkGroup ="+fkGroup );
//System.err.println("booleanValue="+booleanValue);
		fieldDO.setRefConstraint(booleanValue);

	}

	/**
	 * Method declaration
	 *
	 *
	 * @param node
	 * @param packageDO
	 *
	 * @exception DOMLException
	 *
	 * @see
	 */
	private void placeNodeIntoPackageDataObject(Node node, DOPackage packageDO)
			throws DOMLException {
		NodeList elements = node.getChildNodes();
		String   childPackageName = packageDO.getPackageName();

		if (childPackageName == null || childPackageName.length() == 0) {
			childPackageName = packageDO.getName();
		} else {
			childPackageName = childPackageName + "." + packageDO.getName();
		}
		if (elements != null) {
			for (int i = 0; i < elements.getLength(); i++) {
				Node element = elements.item(i);

				if (element.getNodeType() == Node.TEXT_NODE) {

					/* we ignore text at a package level, it's whitespace */
					continue;
				}
				NamedNodeMap elementAttrs = element.getAttributes();
				String	   name = null;

				if (elementAttrs != null) {
					Node idAttr =
						elementAttrs.getNamedItem(key(PACKAGE_NAME));

					if (idAttr != null) {
						name = idAttr.getNodeValue();
					}
					if (name != null && childPackageName != null) {
						if (name.startsWith(childPackageName)) {
							name = name.substring(childPackageName.length()
												  + 1);
						}
					}
				}
				if (name == null) {
					registerWarning(new DOMLWarning("null name for "
													+ element.getNodeName()
													+ " node in "
													+ childPackageName));
					continue;
				}
				if (element.getNodeName().equals(PACKAGE)) {
					DOPackage subPackage = null;

					try {
						subPackage = new DOPackage(name);
						subPackage.setPackageName(childPackageName);
					} catch (DataObjectNameException done) {
						throw new DOMLException(done,
												"DOML package " + name
												+ " has invalid package name");
					}
					placeNodeIntoPackageDataObject(element, subPackage);
					packageDO.addDOPackage(subPackage);
				} else {
					DataObject classDO = null;

					try {
						classDO = new DataObject(name);
						classDO.setPackageName(childPackageName);
					} catch (DataObjectNameException done) {
						throw new DOMLException(done,
												"DOML non-package " + name
												+ " has invalid package name");
					}
					placeNodeIntoClassDataObject(element, classDO,
												 elementAttrs);
					packageDO.addDataObject(classDO);
				}
			}
		}
	}

	/**
	 * Method declaration
	 *
	 *
	 * @param map
	 * @param name
	 *
	 * @return
	 *
	 * @see
	 */
	private boolean isAttributeValue(NamedNodeMap map, String name,
									 boolean def) {
		String  s = getAttributeValue(map, name);
		boolean b = def;

		if (s != null) {
			b = Boolean.valueOf(s.toLowerCase()).booleanValue();
		}
		return b;
	}

	private boolean isAttributeValue(NamedNodeMap map, String name) {
		return isAttributeValue(map, name, false);
	}

	/**
	 * Method declaration
	 *
	 *
	 * @param map
	 * @param name
	 *
	 * @return
	 *
	 * @see
	 */
	private String getAttributeValue(NamedNodeMap map, String name) {
		return getAttributeValue(map, name, new String());
	}

	/**
	 * Method declaration
	 *
	 *
	 * @param map
	 * @param name
	 * @param defValue
	 *
	 * @return
	 *
	 * @see
	 */
	private String getAttributeValue(NamedNodeMap map, String name,
									 String defValue) {
		Node   node = null;
		String value = null;

		if (map != null) {
			node = map.getNamedItem(name);
		}
		if (node != null) {
			value = node.getNodeValue();
		}
		if ((value == null) || (value.trim().length() == 0)) {
			value = defValue.trim();
		} else {
			value = value.trim();
		}
		return value;
	}

	/**
	 * Method declaration
	 *
	 *
	 * @param keyIn
	 *
	 * @return
	 *
	 * @see
	 */
	private String key(String keyIn) {
		String keyOut = keyIn;

		if (!current) {
			keyOut = keyMap.getProperty(keyIn, keyIn);
		}
		keyOut = keyTrim(keyOut);
//System.err.println("keyOut ="+keyOut );
		return keyOut;
	}

	public String keyTrim(String in) {
		String out = in;
		int	index = -1;

		if (out != null) {
			index = out.lastIndexOf('.');
			if ((index > -1) && (out.length() > (index + 1))) {
				out = out.substring(index + 1);
			}
		}
		return out;
	}

	/**
	 * Method declaration
	 *
	 *
	 * @param elementName
	 * @param attributeName
	 * @param valueIn
	 *
	 * @return
	 *
	 * @see
	 */
	private String value(String key, String valueIn) {
		String	   valueOut = valueIn;
		String	   lookup = null;
		StringBuffer buf = new StringBuffer();

		buf.append(key);
		buf.append(".");
		buf.append(valueIn);
		lookup = valueMap.getProperty(buf.toString());
		if (lookup != null) {
			valueOut = lookup;
		}
		return valueOut;
	}

	/**
	 * Method declaration
	 *
	 *
	 * @see
	 */
	private void initUpgradeMaps() {
		keyMap = new Properties();
		valueMap = new Properties();

		//
		keyMap.setProperty(VENDOR, V3_DATABASE);
		keyMap.setProperty(VENDOR_NAME, V3_DATABASE);
		keyMap.setProperty(PACKAGE_NAME, V3_ID);
		keyMap.setProperty(CLASS_TABLE, V3_TABLE);
		keyMap.setProperty(CLASS_LAZYLOADING, V3_LAZY_LOADING);
		keyMap.setProperty(CLASS_OID, V3_OID);
		keyMap.setProperty(CLASS_EXTENDS, V3_EXTENDS);
		keyMap.setProperty(FIELD_NAME, V3_ID);
		keyMap.setProperty(FIELD_INDEX, V3_INDEX);
		keyMap.setProperty(FIELD_PRIMARY, V3_PRIMARY);
		keyMap.setProperty(FIELD_CANQUERY, V3_CAN_QUERY);
		keyMap.setProperty(FIELD_CONSTANT, V3_CONSTANT);
		keyMap.setProperty(FIELD_NULLALLOWED, V3_NULL_ALLOWED);
		keyMap.setProperty(FIELD_DATABASETYPE, V3_DATABASE_TYPE);
		keyMap.setProperty(FIELD_JAVATYPE, V3_JAVA_TYPE);
		keyMap.setProperty(DEFAULT, V3_DEFAULT);

		//
		valueMap.setProperty(VENDOR_NAME + "." + V3_STANDARD, STANDARD);
		valueMap.setProperty(VENDOR_NAME + "." + V3_MICROSOFT, MICROSOFT);
	}

	/**
	 * Method declaration
	 *
	 *
	 * @param parent
	 * @param lookFor
	 *
	 * @return
	 *
	 * @see
	 */
	private Node lookupChildNode(Node parent, String lookFor) {
		Node   cursor = null;
		String nodeName = new String();

		cursor = parent.getFirstChild();
		while (cursor != null) {
			nodeName = cursor.getNodeName().trim();
			if (nodeName.equalsIgnoreCase(lookFor)) {
				break;
			} else {
				cursor = cursor.getNextSibling();
			}
		}
		return cursor;
	}

	/**
	 * Method declaration
	 *
	 *
	 * @param parent
	 *
	 * @return
	 *
	 * @see
	 */
	private String getTextChildValue(Node parent) {
		Node   textNode = null;
		String stringValue = null;

		if (parent == null) {

			// skip
		} else if (parent.getNodeType() == Node.TEXT_NODE) {

			// skip
		} else {
			textNode = parent.getFirstChild();
		}
		if (textNode != null) {
			stringValue = textNode.getNodeValue();
		}
		return stringValue;
	}

	/**
	 * Method declaration
	 *
	 *
	 * @param type
	 *
	 * @return
	 *
	 * @see
	 */
	private boolean isJavaTypeReference(String type) {
		boolean ref = true;

		for (int i = 0; i < JAVA_TYPE_LIST.length; i++) {
//csc_052101 - looks like a really simple mistake: notice that it was
//always comparing against the first element in the list
//			if (JAVA_TYPE_LIST.equals(type)) {
			if (JAVA_TYPE_LIST[i].equals(type)) {	//csc_052101
				ref = false;
				break;
			}
		}
		if (ref) {
			ref = type.endsWith(REF_SUFFIX);
		}
		return ref;
	}

}
