/*
 * Enhydra Java Application Server Project
 *
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific terms governing rights and limitations
 * under the License.
 *
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 *
 * Contributor(s):
 *  @author     christianc@lutris.com
 *
 * $Id: DodsGenerator.java,v 1.1 2001/09/28 16:01:07 cryd0221 Exp $
 */
package org.enhydra.dods.taskdefs;

import java.io.*;
import java.lang.reflect.*;
import java.util.*;
import org.apache.tools.ant.*;
import org.apache.tools.ant.taskdefs.*;
import org.apache.tools.ant.types.*;

/**
 * <p>This Ant taskdef invokes DODS to generate a set of java
 * classes from a doml file and then compiles those classes into the
 * destination directory. The files will only be regenerated/compiled
 * if the date on the doml file is NEWER than at least one of the generated
 * files, unless of course force=true.
 *
 * <p>This taskdef extends the Javac taskdef, so you should refer to
 * that documentation for paramters that affect compilation.
 *
 * <p>DODS specific tags are as follows:</p>
 *
 *  <table border="1" cellspacing="3" cellpadding="3" width="100%">
 *    <tr>
 *      <th>tag</th>
 *      <th>required</th>
 *      <th>default value</th>
 *      <th>description</th>
 *    </tr>
 *    <tr>
 *      <td>doml</td>
 *      <td>yes</td>
 *      <td>&nbsp;</td>
 *      <td>The doml input file describing data object mapping<td>&nbsp;</td>
 *    </tr>
 *    <tr>
 *      <td>sourceout</td>
 *      <td>yes</td>
 *      <td>&nbsp;</td>
 *      <td>Target for generated classes, expressed as a directory path</td>
 *    </tr>
 *    <tr>
 *      <td>enhydraVersion</td>
 *      <td>no</td>
 *      <td>-enhydra1</td>
 *      <td>-enhydra3 or -enhydra4 (see the DODS docs for details)</td>
 *    </tr>
 *    <tr>
 *      <td>jdbcVersion</td>
 *      <td>no</td>
 *      <td>-jdbc1</td>
 *      <td>-jdbc1 or -jdbc2 (see the DODS docs for details)</td>
 *    </tr>
 *    <tr>
 *      <td>ejb</td>
 *      <td>no</td>
 *      <td>&nbsp;</td>
 *      <td>-ejb indicates we want to generate EJB classes (see the DODS docs 
 *          for details)</td>
 *    </tr>
 *    <tr>
 *      <td>lazy</td>
 *      <td>no</td>
 *      <td>&nbsp;</td>
 *      <td>-lazy indicates we wish to use lazy loading (see the DODS docs for details)</td>
 *    </tr>
 *    <tr>
 *      <td>vendor</td>
 *      <td>no</td>
 *      <td>&nbsp;</td>
 *      <td>-vendor options include: Standard, Oracle, Msql, InstantDB, Informix, 
 *          Sybase, PostgreSQL (see the DODS docs for details)</td>
 *    </tr>
 *    <tr>
 *      <td>force</td>
 *      <td>no</td>
 *      <td>false</td>
 *      <td>force=true will force the DODS files to be regenerated</td>
 *    </tr>
 *    <tr>
 *      <td>showdetail</td>
 *      <td>no</td>
 *      <td>false</td>
 *      <td>showdetail=true will print all the DODS debugging messages</td>
 *    </tr>
 *
 *  </table>
 * <p>
 *
 * <p>Look at the Dods build.xml file for further usage examples.
 *
 * @author Christian Cryder <a href="mailto:christianc@enhydra.org">christianc@enhydra.org</a>
 */
public class DodsGenerator extends Javac {

    //private vars
    protected File domlFile = null;
    protected File domlTempFile = null;
    protected File sourceoutDir = null;
    protected String enhydraVersion = "-enhydra3";
    protected String jdbcVersion = "-jdbc1";
    protected String generateEjb = null;
    protected String lazy = null;
    protected String vendor = null;
    protected boolean force = false;
    protected String verbose = null;
    protected static String logHdr1 = " [dods_gen] ";


    /**
     * Force a rebuild
     */
    public void setForce(boolean force) {
        this.force = force;
    }

    /**
     * Set verbosity
     */
    public void setShowdetail(boolean val) {
        if (val) {
            verbose = "-verbose";
        } else {
            verbose = null;
        }
    }

    /**
     * Sets the doml descriptor file.
     */
    public void setDoml(File domlFile) {
        this.domlFile = domlFile;
        domlTempFile = new File(domlFile+".tmp");
    }

    /**
     * Sets the target directory
     */
    public void setSourceout(File sourceout) {
        this.sourceoutDir = sourceout;
    }
    
    /**
     * Set the enhydra version for the generated code
     *
     * @param enhydraVersionName Either enhydra3 to generate code compatible 
     *      with EAF (uses DBTransactions), or enhydra4 to generate code that 
     *      uses Connections or DataSources
     */
    public void setEnhydraversion(Enhydra3or4 attr) {
        String option = attr.getValue();
        if (option.equals("enhydra3")) {
            enhydraVersion = "-enhydra3";
        } else if (option.equals("enhydra4")) {
            enhydraVersion = "-enhydra4";
        } else {
            enhydraVersion = "-enhydra4";
        }
    }

    /**
     * Set the jdbc version for the generated code
     *
     * @param jdbcVersionName Either jdbc1 to generate code that accepts Connections 
     *      as parameters (can be used with either enhydra3 or enhydra4), or jdbc2 to 
     *      generate code that uses JNDI to lookup DataSources to obtain Connections 
     *      (implies enhydra4)
     */
    public void setJdbcversion(Jdbc1or2 attr) {
        String option = attr.getValue();
        if (option.equals("jdbc1")) {
            jdbcVersion = "-jdbc1";
        } else if (option.equals("jdbc2")) {
            jdbcVersion = "-jdbc2";
        } else {
            jdbcVersion = "-jdbc2";
        }
    }

    /**
     * Set whether we generate ejb wrappers
     *
     * @param attr True to generate EJB wrappers (implies -enhydra4)
     */
    public void setEjb(boolean attr) {
        if (attr) {
            generateEjb = "-ejb";
        } else {
            generateEjb = null;
        }
    }

    /**
     * Set lazy-loading for all data objects.  Referenced objecs will be 
     * fetched only when needed
     *
     * @param attr True to set lazy loading
     */
    public void setLazy(boolean attr) {
        if (attr) {
            lazy = "-lazy";
        } else {
            lazy = null;
        }
    }

    /**
     * Set the database vendor if any
     *
     * @param vendorName Database vendor name: Standard, Oracle, Msql, InstantDB, 
     *      Informix, Sybase, PostgreSQL.  This is handy for testing against 
     *      various database vendors; you don't need to change the line in your 
     *      .doml file
     */
    public void setVendor(DBvendor attr) {
        String option = attr.getValue();
        if (option!=null) {
            vendor = "-vendor=" + option;
        } else {
            vendor = null;
        }
    }


    /**
     * Simply invoke DODS Generator, and then compile the generated files
     */
    public void execute() throws BuildException {
    
        // first off, make sure that we've got the required attributes.
        if (domlFile==null) throw new BuildException("doml attribute must be set!");
        if (sourceoutDir==null) throw new BuildException("sourceout attribute must be set!");
        if (!sourceoutDir.isDirectory()) throw new BuildException("sourceout must be a valid directory!");

        //find the tmp doml file
        File domlTmp = null;
        if (domlTmp==null) {
            String s = domlFile.getAbsolutePath().toLowerCase();
            int i = s.indexOf(".doml");
            if (i>-1) domlTmp = new File(domlFile.getAbsolutePath().substring(0,i)+".uptodate");
            else domlTmp = new File(domlFile.getAbsolutePath()+".uptodate");
        }
        
        //next see if we need to rebuild (ie. if the sourceoutDir does not exist, OR
        //if any of the files in sourceoutDir are _older_ than the doml file)
        boolean regen = (force || 
                         !sourceoutDir.exists() ||
                         !domlTmp.exists() || 
                         domlTmp.lastModified() < domlFile.lastModified());

/*
//We can't do this because if there were multiple domls in the same directory
//it would cause problems. Instead, we'll just have to rely on the .uptodate 
//method (above)        
        if (!regen) {
            long domlLastMod = domlFile.lastModified();
            DirectoryScanner ds = new DirectoryScanner();
            String[] includes = {"**\\*.java"};
            ds.setIncludes(includes);
            ds.setBasedir(sourceoutDir);
            ds.scan();
            String[] files = ds.getIncludedFiles();
            for (int i=0; i<files.length;i++) {
                File f = new File(sourceoutDir+File.separator+files[i]);
                if (f.lastModified()<domlLastMod) {
                    regen = true;
                    break;
                }
            }
        }
*/        
        if (regen) project.log(logHdr1+"Regenerating DODS files from "+domlFile+"... "+(force ? "(force=true)" : ""), Project.MSG_INFO);
        else project.log(logHdr1+domlFile+" up to date.", Project.MSG_INFO);

        //if we need to regen...
        if (regen) {
            //Save the system properties, since DODS will jam them.  Restore afterwards.
            java.util.Properties oldProps = System.getProperties();
            System.setProperties((Properties) oldProps.clone());

    /*
            //Save a backup copy of the doml file (DODS has a tendency to stomp on these too)
            Copy copy = new Copy();
            copy.setProject(project);
            copy.setFile(domlFile);
            copy.setTofile(domlTempFile);
            copy.execute();
    */
            //build an args list
            ArrayList argsList = new ArrayList();
            argsList.add(domlFile.getAbsolutePath());
            argsList.add(sourceoutDir.getAbsolutePath());
            argsList.add("regen");
            argsList.add(enhydraVersion);
            argsList.add(jdbcVersion);
            if (generateEjb!=null) argsList.add(generateEjb);
            if (vendor!=null) argsList.add(vendor);
            if (lazy!=null) argsList.add(lazy);
            if (verbose!=null) argsList.add(verbose);

            //now generate...
            try {
                // Execute DODS directly.  This requires that enhydra.jar, or an equivalent
                // path containing the needed DODS code is on the CLASSPATH of ANT.
                String dodsClassName = "org.enhydra.dods.builder.engine.Builder";
                Method m = null;
                Class c = null;
                try {
                    c = Class.forName(dodsClassName);
                    m = c.getMethod("main", new Class[] {String[].class});
                } catch(Exception e) {
                    // Reflection errors. Class or method does not exist!
                    e.printStackTrace();
                    throw new BuildException(e);
                }
                String [] args = (String[]) argsList.toArray(new String[argsList.size()]);
                if (m!=null) {
                    m.invoke(null, new Object[] {args});
                }

            } catch(InvocationTargetException ite) {
                ite.printStackTrace();
                throw new BuildException(ite);

            } catch (Exception e) {
                e.printStackTrace();
                throw new BuildException(e);
        
            } finally {
                //make sure we restore the system props
                System.setProperties(oldProps);

    /*
                //delete the temp file too
                Delete delete = new Delete();
                delete.setProject(project);
                delete.setFile(domlTempFile);
                delete.execute();
    */            
    
            }


            //finally, compile the newly created files
            super.execute();
            
            //assuming we make it past the compile ok, go ahead and touch the 
            //domlTmp file
            Touch touch = new Touch();
            touch.setProject(project);
            touch.setFile(domlTmp);
            touch.execute();
        }
    }


    /**
     * Enumerated attribute with the values "enhydra3" or "enhydra4"
     */
    public static class Enhydra3or4 extends EnumeratedAttribute {
        public String[] getValues() {
            return new String[] {"enhydra3", "enhydra4"};
        }
    }

    /**
     * Enumerated attribute with the values "jdbc1" or "jdbc2"
     */
    public static class Jdbc1or2 extends EnumeratedAttribute {
        public String[] getValues() {
            return new String[] {"jdbc1", "jdbc2"};
        }
    }

    /**
     * Enumerated attribute with the supported database vendors
     */
    public static class DBvendor extends EnumeratedAttribute {
        public String[] getValues() {
            return new String[] {"Standard", "Oracle", "Msql", "InstantDB", "Informix", "Sybase", "PostgreSQL"};
        }
    }
}
