/*
 * Enhydra Java Application Server Project
 *
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 *
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 *
 * Contributor(s):
 *   - This class Copyright 1999 Plugged In Communications Pty Ltd 
 *     All rights reserved. Released under Enhydra Licence 10 August 1999
 *
 * $Id: Util.java,v 1.1 2001/09/28 16:01:07 cryd0221 Exp $
 */

package org.enhydra.dods.util;

import java.io.File;
import java.text.*;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.Toolkit;
import javax.swing.UIManager;


/*
 * Author      : Chris Ryan (cryan@pisoftware.com)
 *
 * Description : This class provides various static convenience methods
 *               for often used processes.
 *
 */
public final class Util {


  /**
   * Constructor.  Don't want anyone to invoke one of these.
   */
  protected Util() {
  }


  /**
   * This method replaces all instances of one string to
   * another string within the main string.  For example,
   * if the main string is "Foo bar baz", and the string
   * to replace is "ba" with "cu", then the returned
   * result will be "Foo cur cuz".
   *
   * @param main_value  The value to do the replacing on
   * @param search_value  The value to search for
   * @param sub_value  The value to replace any found values
   * @return see comment.
   */
  public static String replace(String main_value, String search_value, String sub_value) {

    int pos = 0;
    int old_pos = 0;
    StringBuffer result = new StringBuffer();

    while ((old_pos < main_value.length()) &&
           ((pos = main_value.indexOf(search_value, old_pos)) != -1)) {

      // add everything from the old position to here
      if (pos != old_pos)
        result.append(main_value.substring(old_pos, pos));

      // add the replacement value
      result.append(sub_value);

      // move the old position along
      old_pos = pos + search_value.length();

    }

    // did we get to the end of the string?
    if (old_pos <= main_value.length()) {

      // yes, add the remainder to the result
      result.append(main_value.substring(old_pos));

    }

    return result.toString();
  }


  /**
   * This method attempts to convert the string passed into
   * a date.
   *
   * @param datestr  The string to convert
   * @return The converted string, or null if any errors occur
   */
  public static java.util.Date stringToDate(String datestr) {

    // basic checks...
    if (datestr == null || datestr.length() == 0) {    
      return null;
    }

    // These are the date formats we can accept
    String[] formats = {
      "yyyy/MM/dd H:mm:ss a",    "yyyy/MM/dd K:mm:ss",
      "d/M/yyyy H:mm:ss a",    "d/M/yyyy K:mm:ss",
      "MM/dd/yyyy H:mm:ss a",    "MM/dd/yyyy K:mm:ss",
      "yyyy-MM-dd H:mm:ss a",    "yyyy MM dd H:mm:ss a",
      "yyyy-MM-dd H:mm:ss",    "yyyy MM dd H:mm:ss",
      "yyyy-MM-dd H:mm a",    "yyyy MM dd H:mm a",
      "yyyy-MM-dd H:mm",    "yyyy MM dd H:mm",
      "yyyy/MM/dd",      "dd/MM/yyyy",
      "yyyy-MM-dd",      "yyyy MM dd",
      "MMM dd yyyy h:mm:ss a",  "MMM dd, yyyy h:mm:ss a",
      "MMM dd yyyy h:mm:ss",    "MMM dd, yyyy h:mm:ss",
      "MMM dd yyyy h:mm a",    "MMM dd, yyyy h:mm a",
      "MMM dd yyyy h:mm",    "MMM dd, yyyy h:mm" };

    java.util.Date result = null;

    // try each format...
    for ( int i = 0; i < formats.length; i++ ) {

      // create the date formatter with this format string
      SimpleDateFormat df = new SimpleDateFormat( formats[ i ] );

      try {

        // parse the string with this date formatter
        result = df.parse( datestr );

        // jump out if we get here, since we've got a valid result
        break;
      } catch ( ParseException e ) {

        // ignore this - if the string can't be parsed we'll get this exception
      }

    }

    // return the parsed date... hopefully
    return result;

  }

  public static String getExtension(File file) {
    String ext  = null;
    String name = null;
    int dotAt = 0;
    if (file != null) {
      name = file.getName();
      dotAt = name.lastIndexOf('.');
      if (dotAt > 0) {
        ext = name.substring(dotAt + 1).toLowerCase().trim();
      }
    }
    return ext;
  }

  public static String getBase(File file) {
    String base  = null;
    String name = null;
    int dotAt = 0;
    if (file != null) {
      name = file.getName();
      dotAt = name.lastIndexOf('.');
      if (dotAt > 0) {
        base = name.substring(0, dotAt).trim();
      }
    }
    return base;
  }

  public static Point getCenteringPoint(Dimension size) {
    Point point = new Point();
    Dimension screen;
    screen = Toolkit.getDefaultToolkit().getScreenSize();
    if (size.height > screen.height) {
      size.height = screen.height;
    }
    if (size.width > screen.width) {
      size.width = screen.width;
    }
    point.x = (screen.width  - size.width)  / 2;
    point.y = (screen.height - size.height) / 2;
    return point;
  }

  public static void setLookForWindowsOnly() {
    try {
      String systemLook = UIManager.getSystemLookAndFeelClassName();
      if (systemLook.trim().endsWith("WindowsLookAndFeel")) {
        UIManager.setLookAndFeel(systemLook);
      }
    }
    catch (javax.swing.UnsupportedLookAndFeelException e) {}
    catch (IllegalAccessException e) {}
    catch (InstantiationException e) {}
    catch (ClassNotFoundException e) {};
  }

}
